"""
HierarchicalLayout.py

This algorithm examines the semantics of the graph structure to come up with an
aesthetic layout for most graph types. 
It starts by finding roots to the graph and removing cycles by reversing back
arcs. Then it assigns each node in the graph to a layer, creating a proper
layered hiearchy. Between each layer, it performs crossing minimization on the
edges. Finally, it tries to align nodes with children/parents in other
layers before finally drawing everything. 

This implementation can deal with edges traversing multiple layers, self-loops,
and to some degree, with hyperedges (edges that have multiple children/parents).

Based on the Sugiyama layout algorithm

See Automatic Layout of Diagrams in Rational Rose by Elias Andersson for
ideas to improve this algorithm.

Notes on Sugiyama algorithm: 
  http://plg.uwaterloo.ca/~itbowman/CS746G/ASug.html
  Notes on: Methods for Visual Understanding of Hierarchical System Structures
  Ivan Bowman, March 12, 1998  
  
Info on Sugiyama algorithm:
  On Automated Graph Layout
  AN ALGORITHMIC AND A DECLARATIVE APPROACH DESIGNED AND IMPLEMENTED IN
    THE FUNCTIONAL SYSTEM CLARITY
  ROGIER VAN DALEN AND MIKE SPAANS, JUNE 2001
  http://www.tech.port.ac.uk/staffweb/addist/Layout.pdf
  
Survey of various ways of doing crossing minimization 2 layers:
  http://portal.acm.org/citation.cfm?id=844266
  Heuristics and meta-heuristics for 2-layer straight line crossing minimization
  Rafael Marti , Manuel Laguna
  
Description of how to create a proper layered graph (step 2)
  http://www.michael-forster.de/publications/gd02.pdf
  Applying Crossing Reduction Strategies to Layered Compound Graphs
  Michael Forster
  
Description of the entire Sugiyama algorithm in some detail:
  http://www.ics.uci.edu/~ses/papers/grafdraw.pdf
  Automatic Graph Drawing Algorithms
  Susan Sim
  
Thesis implementation of Sugiyama algorithm:
  http://etd.lib.fsu.edu/theses/available/etd-05062004-232310/unrestricted/Pitch_Patarasuk_Thesis.pdf
  CROSSING REDUCTION FOR LAYERED HIERARCHICAL GRAPH DRAWING
  By PITCH PATARASUK
  
AT&T implementation report of Sugiyama:
    http://citeseer.ist.psu.edu/gansner93technique.html
    A Technique for Drawing Directed Graphs
    Emden R. Gansner, Eleftherios Koutsofios, Stephen C. North, Kiem-Phong Vo
    AT&T Bell Laboratories Murray Hill, New Jersey 07974
    
Comprehensive overview of algorithms involved in Sugiyama:
  http://citeseer.ist.psu.edu/bastert00layered.html
  Layered Drawings of Digraphs
  Oliver Bastert and Christian Matuszewski
  2001
  
#todo: list
# 2) Greedy crossing heuristic
# 3) Better rank assignment alg
# 4) StateCharts + Hierarchy layout
  
Created Summer 2005, Denis Dube
"""
__revision__ = 0.1
__author__ = 'Denis Dube'
__date__ = 'Summer 2005'

import sys
import time
import random
import math
#from tkMessageBox import showinfo

from OptionDatabase       import OptionDatabase
from OptionDialog         import OptionDialog
from Utilities            import optimizeLinks, optimizerHyperLink
from ModelSpecificCode    import isConnectionLink


def applyLayout( atom3i = None, settings = False, selection = None ):
  
   # Instantiate the layout algorithm, if not already done  
  if( HierarchicalLayout.instance == None ):
    if( atom3i == None ):
      raise Exception, "You forgot to initilize "+__name__+" before using it!"
    HierarchicalLayout.instance = HierarchicalLayout(atom3i)    
  
  if( atom3i ):
    HierarchicalLayout.instance.updateATOM3instance( atom3i )
  
  if( settings ):
    HierarchicalLayout.instance.settings( selection ) 
  else:    
    HierarchicalLayout.instance.main( selection )   
    
class HierarchicalLayout:

  instance = None
        
    
  def __init__(self, atom3i ):
     
    self.cb = atom3i.cb
    self.atom3i = atom3i
    
    self.gridSize = 0

    # Instantiate the Option Database module
    self.__optionsDatabase = OptionDatabase( atom3i.parent,
          'Options_HiearchicalLayout.py', 'Hieararchical Layout Configuration')
    
    # Local methods/variables with short names to make things more readable :D
    newOp = self.__optionsDatabase.createNewOption
    IE = OptionDialog.INT_ENTRY
    BE = OptionDialog.BOOLEAN_ENTRY
      
    # Create New Options
    # Format: OptionKey, defaultValue, optionTuple, promptString, helpString
    
    optionList = [OptionDialog.LABEL, "Times 12", "blue", "left" ]
    
    newOp( 'label0001', None, optionList, 'Node spacing', '' )
    newOp( 'xOffset', 30, IE, "Minimum X Distance", 
        "Minimum horizontal distance between any 2 tree nodes (negative" 
        + " values work too) (Default 30)" )   
    newOp( 'yOffset', 30, IE, "Minimum Y Distance", 
        "Minimum vertical distance between any 2 tree nodes (Default 30)" )  
    newOp( 'addEdgeObjHeight', True, BE, "Add edge object height", 
        "Increment spacing between node layers with edge object drawing of"\
        + " maximum height between 2 given layers" ) 
    
    newOp('sep0000', 'Ignored', OptionDialog.SEPERATOR, "Ignored?", "Ignored!") 
    newOp( 'label0002', None, optionList, 'Miscellaneous options', '' )
    newOp( 'Origin', False, BE, "Start tree at origin?", 
        "If false, the current position of the selected nodes is used" )  
#    newOp( 'Manual Cycles', False, BE, "Manual Cycle Breaking", 
#        "Forces the user to break cycles by manually clicking on nodes" )      
    newOp( 'uncrossPhase1', 5, IE, "Maximum uncrossing iterations", 
        "Maximum number of passes to try to reduce edge crossings" \
        + "\nNote: these only count when no progress is being made" )  
    newOp( 'uncrossPhase2', 15, IE, "Maximum uncrossing random restarts", 
        "These can significantly improve quality, but they restart the " \
        + "uncrossing phase to the beginning..." \
        + "\nNote: these only count when no progress is being made" ) 
    newOp( 'baryPlaceMax', 10, IE, "Maximum gridpositioning iterations", 
        "Number of times a barycenter placement heuristic is run to " \
        + "ensure everything is centered" ) 
    
    newOp('sep0001', 'Ignored', OptionDialog.SEPERATOR, "Ignored?", "Ignored!") 
    newOp('label0003', None, optionList, 'Arrow post-processing options', '')
        
    newOp( 'Spline optimization' , False, BE, "Spline optimization", 
        "Sets the arrow to smooth mode and adds 2 extra control points" )
    newOp( 'Arrow curvature', 0, IE, "Arrow curvature", 
        "Adds a curve of magnitude X to the arrows, "
        +"set to 0 for a straight arrow." )    
        
     
    # Load the options from the file, on failure the defaults above are used.
    self.__optionsDatabase.loadOptionsDatabase()
  
  
  
  def updateATOM3instance( self, atom3i ):
    """ Possible to have multiple instances of atom3 """
    self.cb = atom3i.cb 
    self.atom3i = atom3i   
   
   
   
  def settings( self, selection ):
    """
    Dialog to interactively change the spring's behavior
    Automatically applies layout if not canceled
    """
    if( self.__optionsDatabase.showOptionsDatabase() ):
      self.main( selection )
    
    
    
  def main( self, selection ): 
    """
    Main algorithm, does all the high-level steps, delegates details to other
    methods.
    """
    t = time.time()

    # Step 1: Get all entity nodes (semantic objects) and wrap them
    entityNodeList, linkNodeDict = self.__getEntityLinkTuple(selection)    
    if(len(entityNodeList) == 0):
      return
      
    # Initilize the node wrapper class
    NodeWrapper.Node2WrapperDict = dict()    
    NodeWrapper.ID2LayerEdgeDict = dict() 
    NodeWrapper.SelfLoopList = []
      
    # Wrap the AToM3 semantic nodes to make applying the algorithms easier     
    wrappedNodeList = []
    for node in entityNodeList:
      wrappedNodeList.append(NodeWrapper(node, NodeWrapper.REGULAR_NODE))
          
    # Build a connection map (rapid access maps to children and parents)
    NodeWrapper.LinkMap = dict()
    for wrappedNode in wrappedNodeList:
      wrappedNode.buildConnectivityMaps(linkNodeDict)
      
    # Step 2: Build a proper layered hieararchy
    levelDictionary = self.__createLayers(wrappedNodeList)
          
    # Step 3: Minimize crossings    
    levelDictionary = self.__barycentricOrdering(levelDictionary)
    
    # Step 4: Barycenter node placement in a grid
    self.__prettyNodePlacer(levelDictionary)
    
    # Step 5: Draw nodes and edges on the canvas
    if(len(selection) != 0):
      topLeft = self.__getMaxUpperLeftCoordinate(entityNodeList)
    else:
      topLeft = [0, 0]
    self.__drawNodes(levelDictionary, linkNodeDict, topLeft)
            
    self.debugLevelDict(levelDictionary)
    
    print '\nHierarchical layout took', time.time() - t, 'seconds to compute'
    
  def debugLevelDict(self, levelDictionary, level=None):
    if(level != None):
      levelNames = []
      for node in level:
        levelNames.append([node.getName(), node.getOrder()])
      print 'LEVEL', levelNames
      return      
    for levelInt in levelDictionary.keys():
      levelNames = []
      for node in levelDictionary[levelInt]:
        #levelNames.append([node.getName(), node.getOrder()])
        levelNames.append([node.getName(), node.getGridPosition()])
      print 'LEVEL', levelInt, levelNames
    
    
    
  def debugCrossing(self, levelDictionary):
    def updateOrder(orderedLayer):
      i = 0
      for node in orderedLayer:
        node.setOrder(i)
        i += 1
    for i in range(0, len(levelDictionary) ):
      updateOrder(levelDictionary[i])
    
    total = 0
    for i in range(0, len(levelDictionary) - 1):      
      temp = self.__countCrossings(levelDictionary[i], levelDictionary[i + 1])
      total += temp
      print '   Crossings between levels', i, i + 1, '=', temp
    print 'Total crossings', total      
               
     
    
    
  def __barycentricOrdering(self, levelDictionary):
    """
    Given a dictionary mapping of integer levels to nodes
    Re-orders the nodes on each level to have fewer edge crossings
    Implementation draws from:
      CROSSING REDUCTION FOR LAYERED HIERARCHICAL GRAPH DRAWING
      By PITCH PATARASUK
    However, I don't perform same barycenter order reversals; for some reason
    this yielded MORE crossings rather than less. I DO use random order 
    restarts however, which has given me good to optimal results in tests.
    """
    def updateOrder(orderedLayer):
      """ 
      The ordering is implicit in the node sequence in the list
      However to do a node sort, it's handy to have each node know its order
      This order is used only within __barycentricOrdering() except for debug
      """
      i = 0
      for node in orderedLayer:
        node.setOrder(i)
        i += 1
        
    def copyDict(levelDictionary):
      """ Handy method to make a real copy of the levelDictionary """
      newDict = dict()
      for i in levelDictionary.keys():
        newDict[i] = levelDictionary[i][:]
      return newDict
      
    # Init variables
    totalLevelsInt = len(levelDictionary.keys())
    bestGlobalCrossings = sys.maxint   # Minimum # of crossings found so far
    bestLevelDictionary = None         # Copy of the level dict with min cross
    phaseOneIterations = 0
    phaseOneMax = self.__optionsDatabase.get('uncrossPhase1')  
    phaseTwoIterations = 0
    phaseTwoMax = self.__optionsDatabase.get('uncrossPhase2')  
    
    # Initilize layer orderings 
    for level in levelDictionary.values():
      updateOrder(level)
    
    # Main crossing reduction loop
    while(phaseTwoIterations < phaseTwoMax):
      
      # Iterate over all the levels in the level dictionary
      # If totalLevelsInt = 3, then this range yields [0, 1]
      for j in range(0, totalLevelsInt - 1):
        
        ## Down-Barycenter (Parent to child)
        for node in levelDictionary[j]:
          node.computeBarycenter(True)
        # Sort the list in place according to barycenter values
        levelDictionary[j].sort(lambda a, b: cmp(a.getBarycenter(),
                                                 b.getBarycenter()))
        # Save the node's sorted position in it's internal parameter
        updateOrder(levelDictionary[j])
                
        ## Up-Barycenter (Child to parent)
        for node in levelDictionary[j + 1]:
          node.computeBarycenter(False)
        # Sort the list in place according to barycenter values
        levelDictionary[j + 1].sort(lambda a, b: cmp(a.getBarycenter(),
                                                     b.getBarycenter()))
        # Save the node's sorted position in it's internal parameter
        updateOrder(levelDictionary[j + 1])
        

      # Count crossings
      globalCrossings = 0
      for j in range(0, totalLevelsInt - 1):        
        globalCrossings += self.__countCrossings(levelDictionary[j], 
                                                 levelDictionary[j + 1])
        
      # Did the last iteration of the algorithm reduce crossings?
      if(globalCrossings < bestGlobalCrossings):
        # Crossings reduced, reset counter since more reductions are possible
        phaseOneIterations = 0        
        # Store a copy of levelDictionary, best ordering yet
        bestLevelDictionary = copyDict(levelDictionary)            
        bestGlobalCrossings = globalCrossings    
      else:
        # Crossings not reduced, keep trying, might reduce later
        phaseOneIterations += 1
        if(phaseOneIterations > phaseOneMax):
          # Reductions have ceased, try a random restart...
          phaseOneIterations = 0
          phaseTwoIterations += 1
          
          # Randomize layer orderings
          for currentLevel in levelDictionary.values():
            indexList = range(0, len(currentLevel))
            shuffledList = indexList[:]
            random.shuffle(shuffledList)
            for k in shuffledList:
              currentLevel[indexList.pop()].setOrder(k)
        
    # Return the levelDictionary with the least crossings
    return bestLevelDictionary
           

           
  def __countCrossings(self, layerA, layerB):
    """
    Inputs: layerA and layerB are lists of NodeWrapper objects
    Output: # of crossings between two node layers in O(|E| log |Vsmall|)
    
    NOTE: Most other algorithms for this are O(|E| + Number of crossings)
    Implementation of:
      Simple and Efficient Bilayer Cross Counting
      Wilhelm Barth, Michael Junger, and Petra Mutzel
      GD 2002, LNCS 2528, pp. 130-141, 2002
    """
    # Assumed that layerA is above layerB, so children of A are in B
    # Now figure out which layer is smaller to improve running time a bit
    if(len(layerA) < len(layerB)):
      smallLayer = layerA
      largeLayer = layerB
      isParent2Child = False
    else:
      smallLayer = layerB
      largeLayer = layerA
      isParent2Child = True
      
    # Sort the edges and come up with a sequence of edges (integer indices)
    edgeSequence = []
    for node in largeLayer:
      if(isParent2Child):
        nodeList = node.children.keys()
      else:
        nodeList = node.parents.keys()
      tempList = []
      for targetNode in nodeList:
        tempList.append(targetNode.getOrder())
      tempList.sort()
      edgeSequence.extend(tempList)
          
    # Build the accumulator tree
    firstindex = 1    
    while(firstindex < len(smallLayer)):
      firstindex *= 2
    treesize = 2 * firstindex - 1
    firstindex -= 1
    tree = dict() # Heh, python dictionaries are so abused :)
    for i in range(0, treesize):
      tree[i] = 0
    
    # Count the crossings
    crosscount = 0
    for k in range(0, len(edgeSequence)):
      index = edgeSequence[k] + firstindex
      tree[index] += 1
      while(index > 0):
        if(index % 2):
          crosscount += tree[index + 1]
        index = (index - 1) / 2
        tree[index] += 1
  
    return crosscount
  
    
    

  def __getMaxUpperLeftCoordinate(self, entityNodeList):
    """ 
    Returns the maximum upper left coordinate of all the nodes the layout is
    being applied to
    This corresponds to the minumum x and y coords of all the nodes
    """
    minX = sys.maxint
    minY = sys.maxint
    for node in entityNodeList:
      if(node.graphObject_.y < minY):
        minY = node.graphObject_.y
      if(node.graphObject_.x < minX):
        minX = node.graphObject_.x 
    return (minX, minY)
        
        
    
  def __getEntityLinkTuple(self, selection):
    """
    If selection is empty, get all nodes & links on the canvas
    Else returns the entities and links in the selection
    Returns a tuple containing:
      entityList = List of entity ASG nodes
      linkNodeDict = Mapping of link ASG nodes to VisualObj graph objects
    """
    entityNodeList = []    # Non-edge entities
    linkNodeDict = dict()  # Regular and self-looping edges
    
    # Selection may contain a mixed bag of nodes and links
    if(selection):
      for node in selection:
        semObj = node.semanticObject  
        if(isConnectionLink(node)):   
          #linkNodeList.append(semObj)
          linkNodeDict[semObj] = node
        else:           
          entityNodeList.append(semObj)
          
    # No selection? Grab all nodes in diagram
    else:
      if(not self.atom3i.ASGroot): 
        return ([], [])
      for nodetype in self.atom3i.ASGroot.nodeTypes:  
        for node in self.atom3i.ASGroot.listNodes[nodetype]:      
          if(isConnectionLink(node.graphObject_)):  
            #linkNodeList.append(node)
            linkNodeDict[node] = node.graphObject_
          else:            
            entityNodeList.append(node)

   
    if(selection):
      return (entityNodeList, linkNodeDict)    
    return (entityNodeList, linkNodeDict)    
     
     

  def __createLayers(self, wrappedNodeList):
    """
    Place all nodes in a proper layered hierarchy
    Returns a dictionary indexed by integer level of lists of nodes
    """
    
    wrappedNodeList = self.__cycleEliminator(wrappedNodeList)  
              
    # Debug
    print 'Topological sorting'
    for wrappedNode in wrappedNodeList:
      print '   ', wrappedNode.getName()
      
    # Get roots in the acyclic graph (wrappedNodeList is topologically sorted)
    rootNodes = []
    for wrappedNode in wrappedNodeList:
      if(len(wrappedNode.parents.keys()) == 0):
        rootNodes.append(wrappedNode)
                   
    print "rootNodes"
    for node in rootNodes:
      print '   ', node.getName()


    # Place each node on a layer 
    # NOTE: This may set a single node's layer more than once but since
    #       a topological sort is in effect, the last one is the "right" one...
    queuedNodes = rootNodes[:]
    currentLevelInt = 0
    while(len(queuedNodes) > 0):
      tempQueue = list()
      for node in queuedNodes:
        node.setLayer(currentLevelInt) # Make each node aware of its layer
        tempQueue += node.children.keys() # Children of this node
      queuedNodes = tempQueue
      currentLevelInt += 1
      
    # Build the level dictionary for easy access to node layers
    levelDictionary = dict()
    for node in wrappedNodeList:
      currentLevelInt = node.getLayer()
      if(not levelDictionary.has_key(currentLevelInt)):
        levelDictionary[currentLevelInt] = [node]
      else:
        levelDictionary[currentLevelInt].append(node)
      
    # Add dummy nodes as neccessary to finish the levelDictionary...
    return self.__dumbAndDummy(levelDictionary)


        

  def __cycleEliminator(self, wrappedNodeList):
    """
    Uses topological sort then reverses backward edges to eliminate cycles
    Returns nodes in topological sort order
    """

#    TODO: implement this idea
#    From: A Technique for Drawing Directed Graphs
#    Emden R. Gansner, Eleftherios Koutsofios, Stephen C. North, Kiem-Phong Vo
#    AT&T Bell Laboratories Murray Hill, New Jersey 07974
#It seems reasonable to try to reverse a smaller or even minimal set of edges. 
#One difficulty is that finding a minimal set (the feedback arc set problem) 
#is NP-complete [EMW] [GJ]. More important, this would probably not improve the 
#drawings. We implemented a heuristic to reverse edges that participate in many 
#cycles. The heuristic takes one non-trivial strongly connected component at a 
#time, in an arbitrary order. Within each component, it counts the number of 
#times each edge forms a cycle in a depth-first traversal. An edge with a 
#maximal count is reversed. This is repeated until there are no more 
#non-trivial strongly connected components.

    
    topSort = self.__getTopologicalSort(wrappedNodeList)
    
    # Reverse backward arcs to eliminate cycles 
    for nodeIndex in range(0, len(topSort)):
      wrappedNode = topSort[nodeIndex]         
      wrappedChildrenNodes = wrappedNode.children.keys()
      
      for wrappedChildrenNode in wrappedChildrenNodes:
        # If the child has a smaller index than the current node it is left
        # of the current node. That means it forms a cycle!
        if(nodeIndex > topSort.index(wrappedChildrenNode)):
                    
          # Get the link (or maybe more than one link) between parent
          # wrappedNode and child wrappedChildrenNode
          linkFlagList = wrappedNode.children[wrappedChildrenNode][:]
          
          # Remove it
          del wrappedNode.children[wrappedChildrenNode]
          del wrappedChildrenNode.parents[wrappedNode]
          
          # Reverse it
          tempList = []
          for linkFlag in linkFlagList:
            tempList.append([linkFlag[0], True]) # Reverse flag set to True
          wrappedChildrenNode.children[wrappedNode] = tempList[:]
          wrappedNode.parents[wrappedChildrenNode] = tempList[:]
        
        # Self-loop situation
        elif(nodeIndex == topSort.index(wrappedChildrenNode)):
          print 'SELFLOOP', wrappedChildrenNode.getName()
          
    return topSort
    
    
    
  def __getTopologicalSort(self, wrappedNodeList):
    """
    Returns the node list as topologically sorted with forward arcs running 
    from left to right (if cycles exist, there will be backward arcs too)
    """
    def DFSlookup(node, sortedNodeList):
      """ Sub-method for topological sort, does depth first search """
      for childNode in node.children.keys(): #node.getChildrenWrappers():
        if(not childNode.isVisited()):
          childNode.setVisited()
          DFSlookup(childNode, sortedNodeList)
      sortedNodeList.append(node)
    
    sortedNodeList = []
    for node in wrappedNodeList:
      if(not node.isVisited()):    
        node.setVisited()    
        DFSlookup(node, sortedNodeList)
    sortedNodeList.reverse()
    return sortedNodeList
    
    

  def __dumbAndDummy(self, levelDictionary):
    """
    If an edge crosses more than 1 layer, a dummy node is added so the edge
    can be bent around other nodes (thus avoiding overlap, min crossing)
    """
    uniqueID = 0  
    # Add dummy nodes if an edge traverses > 1 layer
    for currentLevelInt in levelDictionary.keys():
      for node in levelDictionary[currentLevelInt]:       
      
        children = node.children.keys() # Children of this node
        for child in children:
          
          # Child should be exactly 1 layer below the parent at current level
          if(currentLevelInt < child.getLayer() - 1):
            # Insert a dummy node here! One for each layer crossed
            
            dummyParent = node            
            linkFlagList = node.children[child][:]
            del node.children[child] 
            uniqueID += 1
                        
            for i in range(currentLevelInt+1, child.getLayer() ):
              #print 'I need a dummy on level', i, 'for child', child.getName()
              # Create the dummy node, add it to level dict
              dummyNode = NodeWrapper((uniqueID, linkFlagList), 
                                       NodeWrapper.MULTI_LAYER_EDGE, i)
              levelDictionary[i].append(dummyNode)
              
              # I need to be able to trace links back and forth
              # Including the dummy nodes...
              dummyParent.children[dummyNode] = linkFlagList              
              dummyNode.parents[dummyParent] = linkFlagList
              dummyParent = dummyNode
            dummyNode.children[child] = linkFlagList
              
    return levelDictionary

      
      
  def __prettyNodePlacer(self, levelDictionary):
    """
    Given a layering & ordering of nodes, this will set them on a grid such
    that they respect the layering and ordering but such that nodes are as
    close as possible (horizontally) to their children and parents.
    This is done by sweeping up/down using barycenters.
    """    
    def updateOrder(orderedLayer):
      """ 
      The ordering is implicit in the node sequence in the list
      However to do a node sort, it's handy to have each node know its order
      This order is used only within __barycentricOrdering() except for debug
      """
      i = 0
      for node in orderedLayer:
        node.setOrder(i)
        i += 1
    
    # Re-order each level so that self-loops are close to their node 
    # NOTE: self-loops are drawn on the level below their node
    # This actually decreases the # of crossings although they're undetectable
    # by the crossing counting algorithm
    for levelInt in levelDictionary.keys():
      updateOrder(levelDictionary[levelInt])
      for node in levelDictionary[levelInt]:
        if(node in NodeWrapper.SelfLoopList):
          childParent = node.parents.keys()[0]
          
          # Get the orders of all the children of the self-loop's parent
          childOrders = []
          for child in childParent.children.keys():
            childOrders.append(child.getOrder())
          childOrders.sort()
          
          # Use the median order of all the children
          newOrder = childOrders[int(math.ceil(len(childOrders) / 2))]          
          levelDictionary[levelInt].remove(node)
          levelDictionary[levelInt].insert(newOrder, node)
          
          
    # Set priorities for node placements according to fan in, fan out, edges
    for level in levelDictionary.values():
      i = 0
      for node in level:
        # Set initial grid position and calculate node priority
        node.setGridPosition(i)
        node.calculatePriority() 

        # Get maximum horizontal grid size
        i += 1
        self.gridSize = max(self.gridSize , i)
            
    maxIterations = self.__optionsDatabase.get('baryPlaceMax')  
    movements = 1
    iterations = 0
    # Iterate up/down, preserver order, but move to take advantage of barycenter
    while(movements > 0 and iterations < maxIterations): 
      movements = 0
      # Sweep down/up
      for i in range(0, len(levelDictionary) - 1):
        movements += self.__prettyNodeBarycenter(levelDictionary[i], True)
        movements += self.__prettyNodeBarycenter(levelDictionary[i + 1], False)
      iterations += 1
      
    # Make sure we are globally flushed to the left
    minGridX = sys.maxint
    for level in levelDictionary.values():
      minGridX = min(minGridX, level[0].getGridPosition())
    if(minGridX != 0):
      for level in levelDictionary.values():
        for node in level:
          node.setGridPosition(node.getGridPosition() - minGridX)
      

  def __prettyNodeBarycenter(self, levelList, isGoingDown):
    """
    Submethod of the node placer, works on one level, in given direction
    """
    def move(levelList, nodeInLevelIndex, moveNode, isMovingRight):
      """
      Sub-submethod... this will recusively try to move a node horizontally
      to the right or left. The recursion occurs if a node moving to a new
      location must displace a node already occupying the spot. Move will
      fail if a node being nudged away has priority or grid boundary reached
      """
      newGridPosition = moveNode.getGridPosition() + (-1, 1)[isMovingRight]
      
      # Can we move there? Target in bounds?
      if(newGridPosition < 0 or newGridPosition > self.gridSize):
        return 0
            
      neighborIndex = nodeInLevelIndex + (-1, 1)[isMovingRight] # +/- 1 index
      # No neighbor to the right! We can move there
      if(isMovingRight and neighborIndex > len(levelList) - 1):
        isMoving = True
        
      # No neighbor to the left! We can move there
      elif(not isMovingRight and neighborIndex < 0):
        isMoving = True
        
      # Have to shove the neighbor out of his spot...
      else:   
        isMoving = False   
        neighborNode = levelList[neighborIndex]
        neighborGridPosition = neighborNode.getGridPosition()      
        # Neighbor definately in our spot...
        if(neighborGridPosition == newGridPosition):
          movePriority = moveNode.getPriority() 
          neighborPriority = neighborNode.getPriority()
          # Do we out-prioritize the neighbor?
          if(movePriority > neighborPriority):
            isMoving = move(levelList, neighborIndex, 
                            neighborNode, isMovingRight)
        else:
          isMoving = True
        
      # We can move, set new grid position
      if(isMoving):
        moveNode.setGridPosition(newGridPosition)
        return 1  
      return 0
      # End of move method
        
          
    movements = 0
    nodeInLevelIndex = 0
    # Iterate over all nodes, get their bary center, try to move them to it
    for node in levelList:
      baryCenterFloat = node.getGridBarycenter(isGoingDown)
      
      # No children/parent
      if(baryCenterFloat == None):
        continue
        
      desiredGridPosition = int(round(baryCenterFloat))      
      currentGridPosition = node.getGridPosition()
      
      # If not at desired spot, try to move there
      if(currentGridPosition != desiredGridPosition):
        isMovingRight = desiredGridPosition > currentGridPosition
        movements += move(levelList, nodeInLevelIndex, node, isMovingRight)
        
      nodeInLevelIndex += 1   
    return movements
    
    
    
    
  def __drawNodes(self, levelDictionary, linkNodeDict, topLeft):
    """ 
    Takes size of nodes into account to translate grid positions into actual
    canvas coordinates
    """
    setSmooth    = self.__optionsDatabase.get('Spline optimization')   
    setCurvature = self.__optionsDatabase.get('Arrow curvature') 
    minOffsetY = self.__optionsDatabase.get('yOffset') 
    minOffsetX = self.__optionsDatabase.get('xOffset') 
    giveExtraSpaceForLinks = self.__optionsDatabase.get('addEdgeObjHeight') 

    # Caclulate x, y offsets
    offsetX = 0
    levelInt2offsetY = dict()
    for levelInt in levelDictionary.keys():
      currentLevel = levelDictionary[levelInt]
      levelInt2offsetY[levelInt] = 0
      
      # Calculate maximum node size on a per level basis (X is for all levels)
      # Then add minimum seperation distance between nodes
      for node in currentLevel:
        # getSize returns node width, and height of the node & child link icon
        x, y = node.getSize(giveExtraSpaceForLinks)
        offsetX = max(offsetX, x)
        levelInt2offsetY[levelInt] = max(levelInt2offsetY[levelInt], y) 
                                     
        
    maxOffsetX = offsetX + minOffsetX
    halfOffsetX = offsetX / 2
        
    # Send nodes to their final destination, assign final pos to dummy edges
    x, y = topLeft
    for levelInt in levelDictionary.keys():
      currentLevel = levelDictionary[levelInt]      
      longEdgeOffset = [halfOffsetX, levelInt2offsetY[levelInt] / 3]
                    
      # Move each node in the level (Dummy edges save the pos but don't move)     
      for node in currentLevel:
        node.moveTo(x + node.getGridPosition() * maxOffsetX, y, longEdgeOffset)
        
      # Increment y for the next iteration
      y += levelInt2offsetY[levelInt] + minOffsetY
      
    # Self-looping edges (Must move these manually into position)
    for selfLoopedEdge in NodeWrapper.SelfLoopList: 
      x, y = selfLoopedEdge.getEdgePosition()
      obj = selfLoopedEdge.getASGNode().graphObject_
      obj.moveTo(x, y)

    # Re-doing links can take a while, lets show something in meanwhile...
    self.atom3i.parent.update()
     
    # Re-wire the links to take into account the new node positions
    selectedLinks = []
    for obj in linkNodeDict.values():
      selectedLinks.append(obj)
    optimizeLinks(self.cb, setSmooth, setCurvature, 
                  selectedLinks=selectedLinks)
    
    # Re-doing links can take a while, lets show something in meanwhile...
    self.atom3i.parent.update()
    
    # Route multi-layer edges
    self.__edgeRouter()
    
    

    
  def __edgeRouter(self):
    """
    Previously, edges traversing multiple layers were represented as a chain
    of dummy nodes. Now these nodes are used as points on a continuous spline.
    """
    def getEndpoint(nodeTuple, pointList, direction, isReversedEdge):
      """ Gets the nearest arrow endpoint. Handles edge reversal """
      if((direction == 'start' and not isReversedEdge)
         or (direction == 'end' and isReversedEdge)):        
        endNode = nodeTuple[0]
        if(isReversedEdge):
          ix = -2
          iy = -1
        else:
          ix = 0
          iy = 1
      else:        
        endNode = nodeTuple[1]
        if(isReversedEdge):
          ix = 0
          iy = 1
        else:
          ix = -2 
          iy = -1        
          
      # Is it connected to a named port!?!
      if(endNode.isConnectedByNamedPort(edgeObject)):
        handler = endNode.getConnectedByNamedPortHandler(nodeTuple[2]) 
        return dc.coords(handler)[:2]
          
      # Not a named port...
      return list(endNode.getClosestConnector2Point( endNode, pointList[ix], 
                                                             pointList[iy]))   
    
    
    
    #todo: improve method for spline arrows + add comments + optimize?
    print '----------------Dummy Edge Routing-----------------'
    for dummyEdge in NodeWrapper.ID2LayerEdgeDict.keys():
      
      dummyList = NodeWrapper.ID2LayerEdgeDict[dummyEdge]
      dummyNode = dummyList[0]
      dummyChild = dummyNode.children.keys()[0]
      linkFlagList = dummyNode.children[dummyChild]
      
      # Real nodes at start/end of the edge
      edgeSourceNode = dummyNode.parents.keys()[0]
      edgeSourceNode = edgeSourceNode.getASGNode().graphObject_
      dummyNode = dummyList[-1]
      edgeTargetNode = dummyNode.children.keys()[0]
      #print 'Dummy edge number', dummyEdge,
      #print dummyList[0].parents.keys()[0].getName(),  edgeTargetNode.getName()
      edgeTargetNode = edgeTargetNode.getASGNode().graphObject_
      nodeTuple = [edgeSourceNode, edgeTargetNode, None]
      
      # Some edges are internally reversed to break cycles, when drawing
      # this must be taken into account
      isReversedEdge = False
      edgesToRoute = []
      for linkNode, isReversed in linkFlagList:
        edgesToRoute.append(linkNode)
        if(isReversed):
          isReversedEdge = True
        
      # Get all the points the edge must pass through (sorted by layer order)
      dummyList.sort(lambda a, b: cmp(a.getLayer(), b.getLayer()))
      if(isReversedEdge):
        dummyList.reverse()
      sortedDummyRouteList = []
      for node in dummyList:
        sortedDummyRouteList += node.getEdgePosition()
      
      # Set the coordinates of the edge directly 
      # This is complicated by the fact that AToM3 treats edges as two
      # segments that join poorly (for spline arrows)
      for edgeObject in edgesToRoute:        
        dc = edgeObject.graphObject_.dc
        linkObj = edgeObject.graphObject_        
        tag = linkObj.tag
        
        if(isReversedEdge):
          inPoint = dc.coords( tag + "2ndSeg0" )[:2]
          outPoint = dc.coords( tag + "1stSeg0" )[:2]
        else:
          inPoint = dc.coords( tag + "1stSeg0" )[:2]
          outPoint = dc.coords( tag + "2ndSeg0" )[:2]
        
        #print 'Dummy route', sortedDummyRouteList
        numPoints = len(sortedDummyRouteList) / 2
        # Add 2 extra control points for odd case (to make splines nice)
        if(numPoints % 2 == 1):
          if(numPoints == 1):
            center = sortedDummyRouteList
          else:
            start = sortedDummyRouteList[:numPoints - 1]
            end = sortedDummyRouteList[numPoints + 1:]
            center = sortedDummyRouteList[numPoints - 1:numPoints + 1]
          
          if(not isReversedEdge):
            newMid1 = [center[0], center[1] - 20]
            newMid2 = [center[0], center[1] + 20]
          else:
            newMid2 = [center[0], center[1] - 20]
            newMid1 = [center[0], center[1] + 20]
            
                              
          if(numPoints == 1):
            sortedDummyRouteList = newMid1 + center + newMid2 
          else:
            sortedDummyRouteList = start + newMid1 + center + newMid2 + end
          centerIndex = numPoints - 1 + 2
          
        # Add 1 extra control point for even case (to make splines nice)
        else:
          start = sortedDummyRouteList[:numPoints]
          end = sortedDummyRouteList[numPoints:]
          center = [start[-2] + (end[0] - start[-2]) / 2, 
                    start[-1] + (end[1] - start[-1]) / 2]
          sortedDummyRouteList = start + center + end          
          centerIndex = numPoints
          
        # Now I know where the center is... so lets move the center object
        # Is the edge object a hyperlink?
        if(len(edgeObject.in_connections_ + edgeObject.out_connections_) > 2):
          fromObjs = []
          for semObj in edgeObject.in_connections_:
            fromObjs.append(semObj.graphObject_)
          toObjs = []
          for semObj in edgeObject.out_connections_:
            toObjs.append(semObj.graphObject_)
          optimizerHyperLink(dc, linkObj, fromObjs, toObjs, 0, 0, 0, center )
          continue
          
        else:
          linkObj.moveTo(* center)
        
        # Go through the 2 segments in the link
        nodeTuple[2] = edgeObject
        for connTuple in linkObj.connections:
          itemHandler = connTuple[0]
          direction = connTuple[1]
          
          if( direction ):     
            inPoint = getEndpoint(nodeTuple, sortedDummyRouteList,
                                  'start', isReversedEdge)

            segCoords = inPoint + sortedDummyRouteList[:centerIndex+2]
          else:           
            outPoint = getEndpoint(nodeTuple, sortedDummyRouteList,
                                   'end', isReversedEdge) 
            segCoords = sortedDummyRouteList[centerIndex:] + outPoint
            segCoords = self.__reverseCoordList(segCoords)
      
          # Applies the changed coords to the canvas
          dc.coords( * [itemHandler] + segCoords )    
          
          # This may change the associated link drawings: 
          # move them to the new point 
          if( direction ):
            linkObj.updateDrawingsTo(inPoint[0], inPoint[1], itemHandler, 
                                                          segmentNumber=1)
          else:
            linkObj.updateDrawingsTo(outPoint[0], outPoint[1], itemHandler, 
                                                            segmentNumber=2)
    
    

  def __reverseCoordList(self, segCoords):
    """ 
    Input: list of coordinates [x0, y0, x1, y1, ..., xn, yn]
    Output: list of coordinates reversed [xn, yn, ..., x1, y1, x0, y0]
    """    
    reversedCoords = []
    for i in range(len(segCoords) - 1, 0, -2):
      reversedCoords += [segCoords[i - 1], segCoords[i]]
    return reversedCoords
      
      
      
    

class NodeWrapper:
  """
  The Hierarchical layout algorithm works exclusively on objects instantiated
  from this class. For the most part, this is a wrapper around the real entity
  nodes in the original diagram. However, special edges are also treated as
  nodes to improve the final layout.
  """
  
  # These need to be reset for each new run of HierarchicalLayout
  Node2WrapperDict = dict() # Map of ASGNode to NodeWrapper
  ID2LayerEdgeDict = dict() # Map of uniqueID's to NodeWrapper
  SelfLoopList = []
  
  REGULAR_NODE = 0
  MULTI_LAYER_EDGE = 1
  SELF_LOOP_EDGE = 2

  def __init__(self, node, nodeType, layer = 0):  
    self.__visitedFlag = False # Used for DFS 
    self.__layer = layer       # Layer node belongs to
    self.__order = -1          # Node order on the layer
    self.__gridPosition = -1   # Actual node pos on grid (cannot violate order)
    self.__priority = 0        # Priority to push other nodes for new grid pos
    self.__barycenterValue = 0 # Barycenter used for crossing minimization
        
    # Dict format: self.children[childNode] = [[linkNode, isReversed],...]
    self.children = dict() 
    self.parents = dict() 
    
    self.__edgePosition = [0, 0]  # Edge control point position (Dummy edges)
    self.__nodeType = nodeType    # Regular node or some type of edge node?
    self.__node = node            # AToM3 node, subclass of ASGnode
        
    # Regular node tracker
    if(nodeType == NodeWrapper.REGULAR_NODE):
      # node is an AToM3 entity, subclass of ASGNode
      NodeWrapper.Node2WrapperDict[node] = self
      
    # Self loop tracker (Special kind of dummy edge)
    elif(nodeType == NodeWrapper.SELF_LOOP_EDGE):
      NodeWrapper.SelfLoopList.append(self) 
    
    # Multi-layer edge node tracker (Dummy edge)
    else:
      self.__node = []        # self.__node = [linkNode,...]
      linkFlagList = node[1]
      for i in range(0, len(linkFlagList)):
        self.__node.append(linkFlagList[i][0]) # The link node object/s
      node = node[0]        # A unique ID number for the dummy tracker
      
      # Add/append self to the tracker with key "node" which is an ID here
      if(NodeWrapper.ID2LayerEdgeDict.has_key(node)):
        NodeWrapper.ID2LayerEdgeDict[node].append(self)
      else:
        NodeWrapper.ID2LayerEdgeDict[node] = [self]
    
    
    
  
  def buildConnectivityMaps(self, linkNodeDict):  
    """
    Creates dictionaries that provide rapid access to the wrapped children
    and parent nodes that are connected with this node.
    Also provides access to the actual link entity between the nodes
    A flag is provided to indicate if the algorithm needed to reverse the link
    Self-links are discarded (Hiearchical layout don't care none about those)
    """
    # Children map
    for linkNode in self.__node.out_connections_:
      # Make sure the link is in the selection
      if(not linkNodeDict.has_key(linkNode)):
        continue
        
      for childNode in linkNode.out_connections_:
        wrappedChildNode = NodeWrapper.Node2WrapperDict[childNode]
        
        # Is the child ourselves in an evil self-loop situation?
        # If so, then create a "dummy" node for the loop itself
        if(self == wrappedChildNode):
          newNode = NodeWrapper(linkNode, NodeWrapper.SELF_LOOP_EDGE) 
          self.children[newNode] = [[linkNode, False]]
          self.parents[newNode] = [[linkNode, False]]
          newNode.children[self] = [[linkNode, False]]
          newNode.parents[self] = [[linkNode, False]]
          continue
          
        # Could be more than one link between source and target
        if(self.children.has_key(wrappedChildNode)):
          self.children[wrappedChildNode].append([linkNode, False])
        else:
          self.children[wrappedChildNode] = [[linkNode, False]]
          
    # Parent map
    for linkNode in self.__node.in_connections_:
      # Make sure the link is in the selection
      if(not linkNodeDict.has_key(linkNode)):
        continue
        
      for parentNode in linkNode.in_connections_:
        wrappedParentNode = NodeWrapper.Node2WrapperDict[parentNode]
        
        # Is the parent ourselves in an evil self-loop situation?
        # If so skip this, we handled this case already in the child map
        if(self == wrappedParentNode):
          continue
        
        # Could be more than one link between source and target
        if(self.parents.has_key(wrappedParentNode)):
          self.parents[wrappedParentNode].append([linkNode, False])
        else:
          self.parents[wrappedParentNode] = [[linkNode, False]]
  
   
  def setVisited(self):
    """ Depth first search has visited here """
    self.__visitedFlag = True
                  
  def isVisited(self):
    """ Return the visited flag """
    return self.__visitedFlag
            
#  def getChildrenWrappers(self):
#    children = []
#    for link in self.__node.out_connections_:
#      for childNode in link.out_connections_:
#        children.append(NodeWrapper.Node2WrapperDict[childNode])
#    return children
    
  def getName(self):
    # DEBUG -- will fail horribly in general formalisms
    if(self.__nodeType == NodeWrapper.MULTI_LAYER_EDGE):
      return 'DUMMY'
    if(self.__node and self.__node.__dict__.has_key('name')):
      return self.__node.name.toString()
    else:
      return 'NO_NAME'
    
  def moveTo(self, x, y, longEdgeOffset):
    if(self.__nodeType == NodeWrapper.REGULAR_NODE):
      self.__node.graphObject_.moveTo(x, y)
    elif(self.__nodeType == NodeWrapper.SELF_LOOP_EDGE):
      #self.__node.graphObject_.moveTo(x, y)
      #print 'Moving loop to', self.getName(), x, y
      centerObj = self.__node.graphObject_.getCenterObject()
      if(centerObj):
        offset = centerObj.getSize()
        self.__edgePosition = [x + offset[0] / 2, y]
        return 
      self.__edgePosition = [x, y]
    else:
      self.__edgePosition = [x + longEdgeOffset[0], y + longEdgeOffset[1]]
      
  def getEdgePosition(self):
    return self.__edgePosition 
        
  def setLayer(self, layer):
    self.__layer = layer
  def getLayer(self):
    return self.__layer
    
  def setOrder(self, order):
    self.__order = order
  def getOrder(self):
    return self.__order
    
  def computeBarycenter(self, isGoingDown):
    """
    Implements the barycenter heuristic. Basic idea: if a node in layer A is
    connected to node1, node2, and node3 in layer B, then the node in layer
    A should be placed at the average of the positions (order integers) of the
    three nodes in layer B. This must be done from layer A to layer B and then
    from layer B to layer A many times to converge on a global solution. 
    The isGoingDown parameter determines if we are moving from root layer to
    leaf layer (True), or from leaf layer to root layer (False)
    
    Implementation according to:
      http://etd.lib.fsu.edu/theses/available/etd-05062004-232310/
      unrestricted/Pitch_Patarasuk_Thesis.pdf
      CROSSING REDUCTION FOR LAYERED HIERARCHICAL GRAPH DRAWING
      By PITCH PATARASUK
    """
    if(isGoingDown):
      nodeList = self.children.keys()
    else:
      nodeList = self.parents.keys()
      
    numberOfNodes = len(nodeList)
    
    # No nodes in next layer? Just use the arbitrary original order value
    if(numberOfNodes == 0):
      self.__barycenterValue = self.__order
      return 

    # Calculate the barycenter
    orderSum = 0
    for node in nodeList:
      orderSum += node.__order
    self.__barycenterValue = orderSum / float(numberOfNodes)
    
  def getBarycenter(self):
    """ Retrieve the barycenter computed by computeBarycenter() """
    return self.__barycenterValue
    

  def calculatePriority(self):
    """ Priority = Fan in + Fan out, and max level priority for dummies """
    if(self.__nodeType == NodeWrapper.MULTI_LAYER_EDGE
       or self.__nodeType == NodeWrapper.SELF_LOOP_EDGE):
      self.__priority = sys.maxint
      return self.__priority
    else:
      self.__priority = len(self.children) + len(self.parents)
      return self.__priority

  def setPriority(self, priority):
    self.__priority = priority
  def getPriority(self):
    return self.__priority
    
  def getASGNode(self):
    return self.__node
    
  def setGridPosition(self, pos):
    self.__gridPosition = pos
  def getGridPosition(self):  
    return self.__gridPosition
    
  def getGridBarycenter(self, isGoingDown):
    if(isGoingDown):
      nodeList = self.children.keys()
    else:
      nodeList = self.parents.keys()
    numberOfNodes = len(nodeList)
    
    if(numberOfNodes == 0):
      return self.__gridPosition

    orderSum = 0
    for node in nodeList:
      orderSum += node.__gridPosition
    return orderSum / float(numberOfNodes)

  def getSize(self, giveExtraSpaceForLinks=True):
    # Long edge traversing multiple layers
    if(self.__nodeType == NodeWrapper.MULTI_LAYER_EDGE):
      maxSize = [0, 0]
      for node in self.__node:
        centerObj = node.graphObject_.getCenterObject()
        if(centerObj):
          tempSize = centerObj.getSize()
          maxSize = [max(maxSize[0], tempSize[0]), 
                     max(maxSize[1], tempSize[1])]
        return maxSize
      return [0, 0]
      
    # Self looping edge
    elif(self.__nodeType == NodeWrapper.SELF_LOOP_EDGE):
      centerObj = self.__node.graphObject_.getCenterObject()
      if(centerObj):
        return centerObj.getSize()
      return [100, 100]
      
    # Check if single layer links have drawings that require additional space
    elif(giveExtraSpaceForLinks):
      maxSingleLayerLinkHeight = 0
      
      def getMaxLinkHeight(linkFlagList, maxSingleLayerLinkHeight):
        for linkFlag in linkFlagList:
          centerObj = linkFlag[0].graphObject_.getCenterObject()
          if(centerObj):
            maxSingleLayerLinkHeight = max(maxSingleLayerLinkHeight, 
                                           centerObj.getSize()[1])
        return maxSingleLayerLinkHeight
        
    
      # Children
      for wrappedChildNode in self.children.keys():
        if(wrappedChildNode.__layer == self.__layer + 1):
          maxSingleLayerLinkHeight = getMaxLinkHeight(
                                            self.children[wrappedChildNode], 
                                            maxSingleLayerLinkHeight)
                    
      # Parents (for links going in reverse direction)
      for wrappedParentNode in self.parents.keys():
        if(wrappedParentNode.__layer == self.__layer - 1):
          maxSingleLayerLinkHeight = getMaxLinkHeight(
                                            self.parents[wrappedParentNode], 
                                            maxSingleLayerLinkHeight)
     
      size = self.__node.graphObject_.getSize()
      return [size[0], size[1] + maxSingleLayerLinkHeight]
    return self.__node.graphObject_.getSize()
      
  