package qoca;
// $Id: QcCassSolver.java.cpp,v 1.7 2000/11/21 04:26:13 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//----------------------------------------------------------------------------//
// This interface needs to be read in conjuction with that of Solver.         //
//============================================================================//

import java.util.*;
import java.io.*;

public class QcCassSolver extends QcDelLinInEqSystem
{
	protected final static int fDefaultInitCapacity = 256;
	protected final static int fDefaultIncrement = 64;
	protected final static float fDefaultLoadFactor = 1.0f;

	protected Vector fEditMap;
	protected Hashtable fStayMap;
	protected Vector fStaySeqMap;
		// Corresponding data structure to fStayMap.  This is used to speed
		// things up when there is a need to iterate through all the stay
		// constraints.  Should not use fStayMap for iteration as the
		// HashtableEnumerator class is not very efficient for iteration.

	//-----------------------------------------------------------------------//
	// Constructors.                                                         //
	//-----------------------------------------------------------------------//
	public QcCassSolver()
	{
		super();
		fStayMap = new Hashtable(fDefaultInitCapacity * 2, fDefaultLoadFactor);
		fStaySeqMap = new Vector(fDefaultInitCapacity, fDefaultIncrement);
		fEditMap = new Vector(fDefaultInitCapacity, fDefaultIncrement);
	}

	public QcCassSolver(int hintNumConstraints, int hintNumVariables)
	{
		super(hintNumConstraints, hintNumVariables);
		fStayMap = new Hashtable(hintNumConstraints * 2, fDefaultLoadFactor);
		fStaySeqMap = new Vector(hintNumConstraints, fDefaultIncrement);
		fEditMap = new Vector(hintNumConstraints, fDefaultIncrement);
	}

	//-----------------------------------------------------------------------//
	// Variable management methods                                           //
	//-----------------------------------------------------------------------//
	public void addVar(QcFloat v)
	{
		if (isRegistered(v))
			return;

		super.addVar(v);
		QcFloat dvp = new QcFloat("", 0.0f, v.stayWeight(), v.editWeight(), true);
		QcFloat dvm = new QcFloat("", 0.0f, v.stayWeight(), v.editWeight(), true);
		fVBiMap.update(dvp, fTableau.addError());
		fVBiMap.update(dvm, fTableau.addError());

		QcLinPoly p = new QcLinPoly();
		p.addUniqueTerm(new QcLinPolyTerm(-1.0f, v));
		p.addUniqueTerm(new QcLinPolyTerm(1.0f, dvp));
		p.addUniqueTerm(new QcLinPolyTerm(-1.0f, dvm));
		QcConstraint stay = new QcConstraint();
		stay.makeEq(p, -v.fDesireValue);

		if (super.addConstraint(stay)) {
			QcCassConstraint cass = new QcCassConstraint(v, stay, dvp, dvm);
			fStayMap.put(v, cass);
			fStaySeqMap.addElement(cass);
		}
	}

	public boolean removeVar(QcFloat v)
	{
		QcCassConstraint stay = (QcCassConstraint)fStayMap.get(v);

		// Remove the stay constraint permanently since the associated variables
		// is removed. Note that the error variables are not removed here, but
		// through the QcDelLinInEqTableau class.
		super.removeConstraint(stay.fConstraint);
 		super.removeVar(stay.fVariable);

 		// Remove the Cassowary constraint from the stay sequential map.
		fStaySeqMap.removeElement(stay);

		// Remove the Cassowary constraint from stay map.
		fStayMap.remove(v);

		return true;
	}

	//-----------------------------------------------------------------------//
	// Constraint management methods                                         //
	//-----------------------------------------------------------------------//
	public boolean addConstraint(QcConstraint c)
	{
		QcUnsortedListSet terms = c.fLinPoly.fTerms;
		int nsize = terms.size();

		// For each QcFloat, check if it has already been added as stay
		// constraint.
		for (int i = 0; i < nsize; i++) {
			QcFloat v = ((QcLinPolyTerm)terms.elementAt(i)).fVariable;
			if (!isRegistered(v))
				addVar(v);
		}

		boolean success = super.addConstraint(c);
		if (success) {
			if (fCheckPost) { //#ifdef qcCheckPost
				addCheckedConstraint(c);
			} //#endif
			if (fAutoSolve)
				solve();
		}
		return success;
	}

	public boolean addConstraint(QcConstraint c, QcFloat hint)
	{
		QcUnsortedListSet terms = c.fLinPoly.fTerms;
		int nsize = terms.size();

		// For each QcFloat, check if it has already been added as stay
		// constraint.
		for (int i = 0; i < nsize; i++) {
			QcFloat v = ((QcLinPolyTerm)terms.elementAt(i)).fVariable;
			if (!isRegistered(v))
				addVar(v);
		}

		boolean success = super.addConstraint(c, hint);
		if (success) {
			if (fCheckPost) { //#ifdef qcCheckPost
				addCheckedConstraint(c);
			} //#endif
			if (fAutoSolve)
				solve();
		}

		return success;
	}

	public void beginAddConstraint()
	{
		System.out.println(
			"Error: QcSolver::beginAddConstraint() not implemented yet");
	}

	public boolean endAddConstraint()
	{
		System.out.println(
			"Error: QcSolver::endAddConstraint() not implemented yet");
		return false;
	}

	protected void initObjective()
	{
		QcVariableIndexIterator vIt = new QcVariableIndexIterator(fTableau);

		// Changed to include all the variables in the objective function.
		while (!vIt.atEnd()) {
			if (fTableau.isError(vIt.fIndex))
				fTableau.setObjValue(vIt.fIndex, fVBiMap.identifier(vIt.fIndex).fWeight);
			else
				fTableau.setObjValue(vIt.fIndex, 0.0f);

			vIt.increment();
		}

		fTableau.eliminateObjective();
	}

	public boolean changeConstraint(QcConstraint oldc, double rhs)
	{
		boolean success = super.changeConstraint(oldc, rhs);
		if (fCheckPost) { //#ifdef qcCheckPost
			if (success)
				changeCheckedConstraint(oldc, rhs);
		} //#endif
		return success;
	}

		
	public boolean removeConstraint(QcConstraint c)
	{
		boolean success = super.removeConstraint(c);
		if (success) {
#ifdef qcCheckPost
			removeCheckedConstraint(c);
#endif
#if 0
			/* [pjm] TODO: The following addEditVar stuff is a bit of a hack. */
			for (int i = c.getNVars(); --i >= 0;) {
				QcFloat v = c.getVar (i);
				if (/*!QcUtility.isZero (v.fValue - v.fDesireValue)
				      &&*/ (v instanceof QcFixedFloat))
					addEditVar (v);
			}
#endif
			if (fAutoSolve)
				solve();
		}
		return success;
	}

	public boolean reset()
	{
		System.out.println
			("Error: QcCassSolver.reset() not implemented yet");
		return false;
	}

	//-----------------------------------------------------------------------//
	// High Level Edit Variable Interface for use by solver clients.         //
	//-----------------------------------------------------------------------//
	public void addEditVar(QcFloat v)
	{
		super.addEditVar(v);
		QcCassConstraint stay = (QcCassConstraint)fStayMap.get(v);
		stay.fDvp.setToEditWeight();
		stay.fDvm.setToEditWeight();
		fEditMap.addElement(stay);
		fEditVarsSetup = false;
	}

	public void endEdit()
	{
		Enumeration eIt = fEditMap.elements();

		while (eIt.hasMoreElements()) {
			QcCassConstraint edit = (QcCassConstraint)eIt.nextElement();
			edit.fDvp.setToStayWeight();
			edit.fDvm.setToStayWeight();
		}

		fEditMap.removeAllElements();
		Enumeration sIt = fStaySeqMap.elements();

		while (sIt.hasMoreElements()) {
			QcCassConstraint stay = (QcCassConstraint)sIt.nextElement();
			// Set desired value to actual value.
			// FIXME: Is this right for QcFixedFloat?
			fTableau.changeRHS(fOCBiMap.index(stay.fConstraint),
					   -stay.fVariable.fValue);
			stay.fVariable.restVariable();
		}

		super.endEdit();
	}

	public void beginEdit()
	{
		if (!fEditVarsSetup) {
			if (fTableau.getRestrictedRows() > 0) {
				initObjective();
				simplexII();
			}
			fEditVarsSetup = true;
		}
	}

	//-----------------------------------------------------------------------//
	// Constraint Solving methods                                            //
	//-----------------------------------------------------------------------//
	public void solve()
	{
		initObjective();
		simplexII();
		rawSolve();

		Enumeration sIt = fStaySeqMap.elements();

		while (sIt.hasMoreElements()) {
			QcCassConstraint stay = (QcCassConstraint)sIt.nextElement();
			fTableau.changeRHS(fOCBiMap.index(stay.fConstraint),
            		-stay.fVariable.fValue);
		}

		if (fCheckPost) { //#ifdef qcCheckPost
			checkSatisfied();
		} //#endif
	}

	public void resolve()
	{
		if (!fEditVarsSetup)
			beginEdit();

		Enumeration eIt = fEditMap.elements();

		while (eIt.hasMoreElements()) {
			QcCassConstraint edit = (QcCassConstraint)eIt.nextElement();
			fTableau.changeRHS(fOCBiMap.index(edit.fConstraint),
            		-edit.fVariable.fDesireValue);
		}

		dualSimplexII();
		rawSolve();

		if (fCheckPost) { //#ifdef qcCheckPost
			checkSatisfied();
		} //#endif
	}

	protected void loadDesValCache()
	{
		QcStructVarIndexIterator vIt = new QcStructVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			if (!fTableau.isSlack(vIt.fIndex)) {
				fTableau.setDesireValue(vIt.fIndex,
                 		fVBiMap.identifier(vIt.fIndex).fDesireValue);
			}
			vIt.increment();
		}
	}

	protected void rawSolve()
	{
		loadDesValCache();
		QcStructVarIndexIterator vIt = new QcStructVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
         		if (!fTableau.isSlack(vIt.fIndex)) {
				QcFloat v = fVBiMap.identifier(vIt.fIndex);

				if (fTableau.isBasic(vIt.fIndex)) {
					int ci = fTableau.isBasicIn(vIt.fIndex);
					v.setValue(fTableau.getRHS(ci));
				} else
					v.setVariable();
			}
			vIt.increment();
		}
	}

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//

	public void restart()
	{
		super.restart();
		fStayMap.clear();
		fStaySeqMap.removeAllElements();
		fEditMap.removeAllElements();
	}

	public void print()
	{
		super.print();

		System.out.println("Stay constraints: ");
		Enumeration sIt = fStaySeqMap.elements();

		while (sIt.hasMoreElements()) {
			((QcCassConstraint)sIt.nextElement()).print();
			System.out.println();
		}

		System.out.println();
		System.out.println("Edit constraints: ");
		Enumeration eIt = fEditMap.elements();

		while (eIt.hasMoreElements()) {
			((QcCassConstraint)eIt.nextElement()).print();
			System.out.println();
		}
	}

#if qcEnableLoadSave
	public void loadFromBinary(DataInputStream s, QcVariableSet vars)
	{
		super.loadFromBinary(s, vars);
		int size = QcUtility.readBinaryInt(s);

		// This will create the underlying data structures with the
		// enough capacity.
		fStayMap = new Hashtable(size * 2, fDefaultLoadFactor);

		for (int i = 0; i < size; i++) {
			QcFloat v = fVBiMap.identifier(QcUtility.readBinaryString(s));
			QcCassConstraint cass = new QcCassConstraint(s, v, fNotifier);
			fStayMap.put(v, cass);
			fStaySeqMap.addElement(cass);
		}
	}

	public void loadFromString(StringTokenizer token, QcVariableSet vars)
	{
		super.loadFromString(token, vars);
		int size = QcUtility.integerValue(token.nextToken());

		// This will create the underlying data structures with the
		// enough capacity.
		fStayMap = new Hashtable(size * 2, fDefaultLoadFactor);

		for (int i = 0; i < size; i++) {
			QcFloat v = fVBiMap.identifier(token.nextToken());
			QcCassConstraint cass = new QcCassConstraint(token, v, fNotifier);
			fStayMap.put(v, cass);
			fStaySeqMap.addElement(cass);
		}
	}

	public void saveAsBinary(DataOutputStream s, QcVariableSet vars)
	{
		super.saveAsBinary(s, vars);
		QcUtility.writeBinaryInt(s, fStayMap.size());
		Enumeration sIt = fStaySeqMap.elements();

		while (sIt.hasMoreElements())
			((QcCassConstraint)sIt.nextElement()).saveAsBinary(s);
	}

	public void saveAsText(OutputStream s, QcVariableSet vars)
	{
		super.saveAsText(s, vars);
		String str = "" + fStayMap.size() + "\n";
		QcUtility.writeTextString(s, str);

		Enumeration sIt = fStaySeqMap.elements();

		while (sIt.hasMoreElements()) {
			((QcCassConstraint)sIt.nextElement()).saveAsText(s);
			QcUtility.writeTextString(s, "\n");
		}
	}

	public void saveToString(StringBuffer s, QcVariableSet vars)
	{
		super.saveToString(s, vars);
		s.append(fStayMap.size());
		s.append("\n");
		Enumeration sIt = fStaySeqMap.elements();

		while (sIt.hasMoreElements()) {
			((QcCassConstraint)sIt.nextElement()).saveToString(s);
			s.append("\n");
		}
	}
#endif /* qcEnableLoadSave */
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
