//autogenwarning

package qoca;
// $Id: QcConstraint.java.cpp,v 1.10 2000/11/21 05:35:16 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-2000 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//============================================================================//

import java.util.*;
import java.io.*;

public class QcConstraint    
	implements
#ifdef qcEnableDuplicate
		QcDupable, 
#endif
		Serializable
{
	private final static boolean fCheckPre = QcConstants.fCheckPre;
	private final static boolean fCheckPost = QcConstants.fCheckPost;

	final static int fOpInvalid = -1;
	public final static int fOpEqual = 0;
	public final static int fOpLesserEqual = 1;
	public final static int fOpGreaterEqual = 2;
	/* N.B. We currently don't allow strict {less|greater} than.  The
           constants are defined only so that we can leave existing switch
           statements unchanged. */
	/** @deprecated Strict inequalities aren't implemented; treated as `&le;'. */
	public final static int fOpLesserThan = 1;
	/** @deprecated Strict inequalities aren't implemented; treated as `&ge;'. */
	public final static int fOpGreaterThan = 2;
	/** The highest valid operator number.
	    @see #assertValidOp **/
	private final static int fLastOp = 2;
	public static int EQ = fOpEqual;
	public static int LE = fOpLesserEqual;
	public static int GE = fOpGreaterEqual;

	private static int fNextValidId = 1;
		// fInvalidId is currently set to 0. All negative values are reserved
		// for future use.

	/* TODO: What happens if user calls makeEq etc. or assigns to these
	   public vars after adding this constraint to a solver?  I think we
	   should remove these methods, leaving just the constructors.

	   OTOH, perhaps it's better for the solvers to generate deep copies for
	   addConstraint; only problem is to make sure the hashing on fID works
	   (certainly would need doing in C++ version, haven't looked at
	   Java). */

	int fId;		// used as index key
	public QcLinPoly fLinPoly;	// The constraint variable coeffs
	public int fOp;		// The relation
	public double fRHS;		// The RHS value
	String fName;

	//-----------------------------------------------------------------------//
	// Constructors.                                                         //
	//-----------------------------------------------------------------------//
	public QcConstraint()
		{ initialize("", new QcLinPoly(), fOpEqual, 0.0f); }

	public QcConstraint(String name)
		{ initialize(name, new QcLinPoly(), fOpEqual, 0.0f); }

	public QcConstraint(String name, QcLinPoly p, int op, double rhs)
		{ initialize(name, p, op, rhs); }

	public QcConstraint(DataInputStream s, QcVariableBiMap bimap)
		{ loadFromBinary(s, bimap); }

	public QcConstraint(StringTokenizer token, QcVariableBiMap bimap)
		{ loadFromString(token, bimap); }

#if qcCheckInternalInvar
void assertInvar() const
{
       qcAssertInvar (fId > 0);
       fLinPoly.assertInvar();
       assertValidOp (fOp);
}
#endif


	//-----------------------------------------------------------------------//
	// Query functions.                                                      //
	//-----------------------------------------------------------------------//
	public boolean equal(QcConstraint other)
		{ return (this == other); }

#ifdef qcCheckPost
	/** Returns true iff this constraint is satisfied or
            <tt>QcUtility.isZero(LHS - RHS)</tt>.  The current implementation
            ignores the possibility of NaN or infinity problems. */
	public boolean isSatisfied()
	{
		double x = fLinPoly.getValue() - fRHS;

		if (QcUtility.isZero (x))
			return true;

		switch (fOp) {
		case fOpEqual:
			return false;

		//case fOpLesserThan:
		case fOpLesserEqual:
			return x < 0;

		//case fOpGreaterThan:
		case fOpGreaterEqual:
			return x > 0;

		default:
			throw new RuntimeException ("Durchfall on fOp " + fOp);
		}
	}
#endif

	//-----------------------------------------------------------------------//
	// Manipulation functions.                                               //
	//-----------------------------------------------------------------------//
	public void initialize(String name, QcLinPoly lp, int op, double rhs)
	{
		fId = fNextValidId++;
		fLinPoly = lp;
		if (fCheckPre)
			assertValidOp(op);
		fOp = op;
		fRHS = rhs;
		fName = name.trim();
	}

	private static void assertValidOp(int op)
	{
		if (op < 0 || op > fLastOp)
			throw new QcInvalidOperator("invalid operator code " + op + " in constraint");
	}

	/** Make this an equality constraint, expressing the relation that lp = rhs. */
	public void makeEq(QcLinPoly lp, double rhs)
	{
		fOp = fOpEqual;
		fLinPoly = lp;
		fRHS = rhs;
	}

	public void makeGe(QcLinPoly lp, double rhs)
	{
		fOp = fOpGreaterEqual;
		fLinPoly = lp;
		fRHS = rhs;
	}

	public void makeLe(QcLinPoly lp, double rhs)
	{
		fOp = fOpLesserEqual;
		fLinPoly = lp;
		fRHS = rhs;
	}

#if 0
	public void makeGt(QcLinPoly lp, double rhs)
	{
		fOp = fOpGreaterThan;
		fLinPoly = lp;
		fRHS = rhs;
	}

	public void makeLt(QcLinPoly lp, double rhs)
	{
		fOp = fOpLesserThan;
		fLinPoly = lp;
		fRHS = rhs;
	}
#endif

	public void reset()
		// Set implementation variables to defaults
		{ fId = fNextValidId++; }

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//

#ifdef qcEnableDuplicate
	public QcConstraint duplicate(int depth)
	{
		String dupName = ((depth == DEEP)
				  ? new String(fName)
				  : fName);
		return new QcConstraint(dupName, fLinPoly.duplicate(depth), fOp, fRHS);
	}

	/** Create a new constraint by deep copying this one.  [FIXME: Actually,
	    only moderately-deep copying is done.  I don't know whether or not
	    fully-deep copying was intended.  It is recommended that you specify
	    a depth argument instead of using this function, as we might later
	    change the implementation to do use a different depth.] */
	public QcConstraint duplicate()
		{ return duplicate(NOT_QCFLOAT); }
#endif

	/** @deprecated Let the qoca maintainer know if you use this. */
	public void loadFromBinary(DataInputStream s, QcVariableBiMap bimap)
	{
		fName = QcUtility.readBinaryString(s);
		fLinPoly = new QcLinPoly(s, bimap);
		fOp = QcUtility.readBinaryInt(s);
		if (fCheckPre)
			assertValidOp(fOp);
		fRHS = QcUtility.readBinaryDouble(s);;
		fId = fNextValidId++;
	}

	public void loadFromBinary(DataInputStream s, QcVariableSet vars)
	{
		fName = QcUtility.readBinaryString(s);
		fLinPoly = new QcLinPoly(s, vars);
		fOp = QcUtility.readBinaryInt(s);
		if (fCheckPre)
			assertValidOp(fOp);
		fRHS = QcUtility.readBinaryDouble(s);
		fId = fNextValidId++;
	}

	public void loadFromString(StringTokenizer token, QcVariableBiMap bimap)
	{
		fName = token.nextToken();
		fLinPoly = new QcLinPoly(token, bimap);
		fOp = QcUtility.integerValue(token.nextToken());
		if (fCheckPre)
			assertValidOp(fOp);
		fRHS = QcUtility.doubleValue(token.nextToken());
		fId = fNextValidId++;
	}

	public String toString()
	{
		String opString;
		switch (fOp) {
		case fOpEqual: 		opString = "=="; break;
		//case fOpLesserThan: 	opString = "<"; break;
		case fOpLesserEqual: 	opString = "<="; break;
		//case fOpGreaterThan: 	opString = ">"; break;
	       	case fOpGreaterEqual: 	opString = ">="; break;
		default:
			throw new RuntimeException("Durchfall on fOp=" + fOp);
		}
		return fName + '#' + fId + '(' + fLinPoly + ", " + opString + ", " + fRHS + ')';
	}

	public void print()
	{
		System.out.print(toString());
	}

	public void saveAsBinary(DataOutputStream s)
	{
		if (fName.equals(""))
			QcUtility.writeBinaryString(s, "$" + fId);
		else
			QcUtility.writeBinaryString(s, fName);

		fLinPoly.saveAsBinary(s);
		QcUtility.writeBinaryInt(s, fOp);
		QcUtility.writeBinaryDouble(s, fRHS);
	}

	public void saveAsText(OutputStream s)
	{
		String str = "";

		if (fName.equals(""))
			str += "$" + fId;
		else
			str += fName;

		str += " ";
		QcUtility.writeTextString(s, str);
		fLinPoly.saveAsText(s);
		str = " " + fOp + " " + QcUtility.toShortestString(fRHS);
		QcUtility.writeTextString(s, str);
	}

	public void saveToString(StringBuffer s)
	{
		if (fName.equals("")) {
			s.append("$");
			s.append(fId);
		} else
			s.append(fName);

		s.append(" ");
		fLinPoly.saveToString(s);
		s.append(" ");
		s.append(fOp);
		s.append(" ");
		s.append(QcUtility.toShortestString(fRHS));
	}

	//-----------------------------------------------------------------------
	// Access to variables.
	// (Currently no write access is provided; if you need write access,
	// please mail the maintainer.)
	//-----------------------------------------------------------------------
	public String getName()
	{
		return fName;
	}

	public QcLinPoly getLinPoly()
	{
		return fLinPoly; // .duplicate(QcDupable.SAFE);
	}

	public int getNVars()
	{
		return fLinPoly.getNTerms();
	}

	public QcFloat getVar(int ix)
	{
		return fLinPoly.getVar(ix);
	}

	public boolean hasVar (QcFloat v)
	{
		return fLinPoly.hasVar (v);
	}

	public int getOp()
	{
		if (fCheckPost)
			assertValidOp(fOp);
		return fOp;
	}

	public double getRHS()
	{
		return fRHS;
	}
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
