package qoca;
// $Id: QcCoreTableau.java.cpp,v 1.6 2000/11/21 05:33:52 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//----------------------------------------------------------------------------//
// MRowIndex and ARowIndex record associations between rows of fA and fM      //
// which are used when a constraint is deleted or not yet in the basis (means //
// the fM[*,Index] is a unit vector, element fM[MRowIndex,Index] is 1.0 and   //
// and the other elements are zero).  MRowIndex is the row of fM              //
// corresponding to column fM[*,Index] of fM and row fA[Index,*] of fA.       //
// ARowIndex is the row of fA corresponding to row fM[Index,*].  Unused       //
// entries in ARowIndex and MRowIndex are set to -1.  Hence MRowIndex and     //
// ARowIndex are used to represent inverse indexing functions on a partial    //
// domain.                                                                    //
// EXCEPTION: this rule is currently violated by Pivot.                       //
//============================================================================//

import java.util.*;
import java.io.*;

#include "QcDefines.jh"

class QcCoreTableau
	implements Serializable
{
	public QcSolvedFormMatrix fSF;	// Solved form.

	protected QcQuasiRowColStateVector fRowColState;
	protected QcQuasiRowStateVector fRowState;
	protected QcQuasiColStateVector fColState;
	protected QcOrigRowStateVector fOrigRowState;  // 0..fRows-1

	/** @invariant <tt>fRows</tt> &ge; 0 */
	protected int fRows;
	protected int fColumns;
	protected int fAllocRows;
	protected int fAllocColumns;

	//-----------------------------------------------------------------------//
	// Constructor.                                                          //
	//-----------------------------------------------------------------------//
	public QcCoreTableau(int hintRows, int hintCols, QcQuasiRowColStateVector rcs)
	{
		fSF = new QcSolvedFormMatrix();
		fOrigRowState = new QcOrigRowStateVector();
		fRowColState = rcs;
		fRowState = fRowColState.fRowState;
		fColState = fRowColState.fColState;
		initialize(hintRows, hintCols);
	}

	protected QcCoreTableau(QcQuasiRowColStateVector rcs)
	{
		fSF = new QcSolvedFormMatrix();
		fOrigRowState = new QcOrigRowStateVector();
		fRowColState = rcs;
		fRowState = fRowColState.fRowState;
		fColState = fRowColState.fColState;
	}

	public QcCoreTableau(DataInputStream s, QcQuasiRowColStateVector rcs)
		{ loadFromBinary(s, rcs); }

	public QcCoreTableau(StringTokenizer token, QcQuasiRowColStateVector rcs)
		{ loadFromString(token, rcs); }

	protected QcCoreTableau()
		// This constructor should only be used by the derived class.  The
		// derive class should ensure that all member data are initialized.
		{ }

	//-----------------------------------------------------------------------//
	// Query functions.                                                      //
	//-----------------------------------------------------------------------//
	public int getAllocRows()
		{ return fAllocRows; }

	public int getAllocColumns()
		{ return fAllocColumns; }

	public int getColumns()
		{ return fColumns; }

	public boolean getARowDeleted(int ci)
		{ return fOrigRowState.getARowDeleted(ci); }

	public int getARowIndex(int ci)
		{ return fRowState.getARowIndex(ci); }

	public boolean getMRowDeleted(int ci)
		{ return fRowState.getMRowDeleted(ci); }

	public int getMRowIndex(int ci)
		{ return fOrigRowState.getMRowIndex(ci); }

	public double getRawRHS(int ci) 		// The orginal constraint RHS
		{ return fOrigRowState.getRHS(ci); }

	public double getRHS(int ci)
		// Get RHS for solved form constraint ci.
	{
		if (getMRowDeleted(ci))
			return 0.0;
		return fSF.getRHS(ci);
	}

	public int getRows()
		{ return fRows; }

	public double getValue(int ci, int vi)
		// Get coefficient for variable vi in constraint ci.
		// getValue and getRHS can be used to access the coefficients
		// of deleted constraints (zero is returned).  This behaviour is
		// required to satisfy the RowCol interface amongst other reasons.
	{
		if (getMRowDeleted(ci))
			return 0.0f;
		return fSF.getValue(ci, vi);
	}

	public boolean isValidCIndex(int ci)
		{ return (ci >= 0 && ci < fRows && !getMRowDeleted(ci)); }

	public boolean isValidOCIndex(int ci)
		{ return (ci >= 0 && ci < fRows && !getARowDeleted(ci)); }

	public boolean isValidVIndex(int vi)
		{ return (vi >= 0 && vi < fColumns); }

	//-----------------------------------------------------------------------//
	// Manipulation functions.                                               //
	//-----------------------------------------------------------------------//
	/** Add a new row, allocating any new columns indicated.
	    @return The new row index.
	    @postcondition ret &ge; 0
	**/
	public int addRow(QcTableauRowIterator varCoeffs, double rhs)
	{
		int r = 0;     // Row of fA, col of fM, CIndex for original constraint
		int quasiRow;	// Row of fM, CIndex for solved form constraint

		// Look for a deleted row or create a new row if one not found
		while (r < fRows && !getARowDeleted(r))
			r++;

		if (r == fRows) {
			r = increaseRows();
			quasiRow = getMRowIndex(r);
		} else {
			quasiRow = getMRowIndex(r);
			setARowDeleted(r, false);
			setMRowDeleted(quasiRow, false);
			linkRow(quasiRow);
		}

		// Load the new values into the fA matrix and RHS
		setRawRHS(r, rhs);
		fSF.fRHS.fData[quasiRow] = rhs;

		while (!varCoeffs.atEnd()) {
			fSF.setValue(quasiRow, varCoeffs.fIndex, varCoeffs.fValue);
			varCoeffs.increment();
		}

		return r;
	}

	public void addScaledRow(int destRow, int srcRow, double factor)
		{ fSF.addScaledRow(destRow, srcRow, factor); }

	public void changeRHS(int ci, double rhs)
		// Change the RHS of the constraint ci.
		{ setRawRHS(ci, rhs); }

	public boolean check() // Check data integrity (assertions)
	{
		// A few important basic checks
		if (!(fRows >= 0 && fColumns >= 0 && fAllocRows >= fRows &&
          		fAllocColumns >= fColumns))
			return false;

		// Check matrices and vectors are the right size
		if (!(fRowState.getSize() == fRows && fOrigRowState.getSize() == fRows))
			return false;

		// Now check a few integrity constraints for row state data
		for (int r = 0; r < fRows; r++) {
			// First consider r as indexing rows of fM
			if (getMRowDeleted(r)) {
				if (!(getMRowIndex(getARowIndex(r)) == r &&
                    		getARowDeleted(getARowIndex(r))))
					return false;
			}

			if (!(getARowIndex(r) == QcTableau.fInvalidConstraintIndex &&
               		(getARowIndex(r) >= 0 && getARowIndex(r) < fRows)))
				return false;

			// Now consider r as indexing rows of fA (and hence columns of fM)
			if (getARowDeleted(r)) {
				if (!(getARowIndex(getMRowIndex(r)) == r &&
                    		getMRowDeleted(getMRowIndex(r))))
					return false;
			}

			if (!(getMRowIndex(r) == QcTableau.fInvalidConstraintIndex &&
					(getMRowIndex(r) >= 0 && getMRowIndex(r) < fRows)))
				return false;
		}

		return true;
	}

	public int deleteRow(int ci)
		// Deletes a row given an original constraint index by setting all
		// the coefficients to zero and marking the row as deleted.
		// Derived classes use this method to delete constraints
		// which are found to be inconsistant after a successful AddEq
		// (i.e. unable to successfully pivot).
	{
		int quasiRow = getMRowIndex(ci);
		setMRowDeleted(quasiRow, true);
		setARowDeleted(ci, true);
		fSF.zeroRow(quasiRow);
		setRawRHS(ci, 0.0f);
		fRowState.unlinkRow(quasiRow);
		return quasiRow;
	}

	/** Create a new non-basic variable, and return its index.
	 **/
	public int increaseColumns()
	{
		if (fColumns == fAllocColumns) {
			fAllocColumns += (fAllocColumns == 0) ? 1 : fAllocColumns;
			fSF.reserve(fAllocRows, fAllocColumns);
			fColState.reserve(fAllocColumns);
		}

		fColumns++;
		fSF.resize(fRows, fColumns);
		fColState.resize(fColumns);

		// Set the default data values for the newly added column
		fSF.zeroColumn(fColumns - 1);

		return fColumns - 1;
	}

	/** Creates a new redundant constraint.
	    @return 1 less than the new <tt>fRows</tt>.

	    @postcondition ret &ge; 0
	**/
	public int increaseRows()
	{
		if (fRows == fAllocRows) {
			fAllocRows += (fAllocRows == 0) ? 1 : fAllocRows;
			fSF.reserve(fAllocRows, fAllocColumns);
			fRowState.reserve(fAllocRows);
			fOrigRowState.reserve(fAllocRows);
		}

		fRows++;
		fSF.resize(fRows, fColumns);
		fRowState.resize(fRows);
		fOrigRowState.resize(fRows);

		// Set the default data values for the newly added row
		fRowState.linkRow(fRows - 1);
		fSF.zeroRow(fRows - 1);
		return fRows - 1;
	}

	public void initialize(int hintRows, int hintCols) // Call this after construction
	{
		if (hintRows < 0)
			hintRows = 0;

		if (hintCols < 0)
			hintCols = 0;

		fRowColState.reserve(hintRows, hintCols);
		fOrigRowState.reserve(hintRows);
		fAllocRows = hintRows;
		fAllocColumns = hintCols;
		fRows = 0;
		fColumns = 0;
		fSF.reserve(hintRows, hintCols);
	}

	public void linkRow(int ci)
		// Add row ci to the row list.
		{ fRowState.linkRow(ci); }

	public void normalize(int ci)
		// Normalize the corresponding row in fSF if the rhs is negative.  This
		// makes rhs positive.
	{
		if (QcUtility.isNegative(getRHS(ci)))
			fSF.scaleRow(ci, -1.0f);
	}

	public void restart() // Erase everything ready to start afresh.
	{
		fRows = 0;
		fColumns = 0;
		fSF.restart();
		fRowColState.restart();
		int size = fOrigRowState.getSize();

		for (int i = 0; i < size; i++)
			fOrigRowState.restart(i);

		fOrigRowState.resize(0);
	}

	public void restore()
		// The tableau is reinitialised as if all the original constraints
		// had been added to a fresh tableau using AddEq but with no
		// Elimination or Pivoting having been done.
	{
		int co;		// original constraint index (row of fA)
		int cs;		// solved form constraint index (row of fM)

		// Reset constraint state information
		for (cs = 0, co = 0; cs < fRows; cs++, co++) {
			// Find next undeleted constraint
			while (cs < fRows && getMRowDeleted(cs))
				cs++;

			while (co < fRows && getARowDeleted(co))
				co++;

			if (co < fRows) {
				fSF.zeroRow(cs);
				fSF.setRHS(cs, fOrigRowState.getRHS(co));
				setMRowIndex(co, cs);
				setARowIndex(cs, co);
				// N.B. We permit Reset to change the co<->cs relationship for
				// a previously invalid constraint.  As long as they
				// are empty it doesn't matter which are used.
			} else
				break;
		}

		// Fix row and column linked lists.
		fRowState.fixLinkage();
		fColState.fixLinkage();
	}

	public void scaleRow(int row, double factor)
		{ fSF.scaleRow(row, factor); }

	public void setARowDeleted(int ci, boolean d)
		{ fOrigRowState.setARowDeleted(ci, d); }

	public void setARowIndex(int ci, int i)
		{ fRowState.setARowIndex(ci, i); }

	public void setMRowDeleted(int ci, boolean d)
		{  fRowState.setMRowDeleted(ci, d); }

	public void setMRowIndex(int ci, int i)
		{ fOrigRowState.setMRowIndex(ci, i); }

	public void setRawRHS(int ci, double rhs) 	// The orginal constraint RHS
		{ fOrigRowState.setRHS(ci, rhs); }

	public void swapColumns(int c1, int c2)
		{ fSF.swapColumns(c1, c2); }

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//
	public void loadFromBinary(DataInputStream s, QcQuasiRowColStateVector rcs)
	{
		fRows = QcUtility.readBinaryInt(s);
		fColumns = QcUtility.readBinaryInt(s);
		qcAssertExternalPre(fRows >= 0);
		qcAssertExternalPre(fColumns >= 0);
		fAllocRows = fRows;
		fAllocColumns = fColumns;
		fOrigRowState = new QcOrigRowStateVector(s);
		fSF = new QcSolvedFormMatrix(s);
		fRowColState = rcs;
		fRowState = fRowColState.fRowState;
		fColState = fRowColState.fColState;
	}

	public void loadFromString(StringTokenizer token, QcQuasiRowColStateVector rcs)
	{
		fRows = QcUtility.integerValue(token.nextToken());
		fColumns = QcUtility.integerValue(token.nextToken());
		qcAssertExternalPre(fRows >= 0);
		qcAssertExternalPre(fColumns >= 0);
		fAllocRows = fRows;
		fAllocColumns = fColumns;
		fOrigRowState = new QcOrigRowStateVector(token);
		fSF = new QcSolvedFormMatrix(token);
		fRowColState = rcs;
		fRowState = fRowColState.fRowState;
		fColState = fRowColState.fColState;
	}

	public void print()
	{
		System.out.println("Begin[CoreTableau]");

		fOrigRowState.print();
		fRowColState.print();
		System.out.println();

		System.out.println("Begin[Solved Form]");
		fSF.print();
		System.out.println("End[Solved Form]");
		System.out.println();

		System.out.println("End[CoreTableau]");
	}

	public void saveAsBinary(DataOutputStream s)
	{
		QcUtility.writeBinaryInt(s, fRows);
		QcUtility.writeBinaryInt(s, fColumns);
		fOrigRowState.saveAsBinary(s);
		fSF.saveAsBinary(s);
	}

	public void saveAsText(OutputStream s)
	{
		String str = "" + fRows + " " + fColumns + "\n";
		QcUtility.writeTextString(s, str);
		fOrigRowState.saveAsText(s);
		fSF.saveAsText(s);
	}

	public void saveToString(StringBuffer s)
	{
		s.append(fRows);
		s.append(" ");
		s.append(fColumns);
		s.append("\n");
		fOrigRowState.saveToString(s);
		fSF.saveToString(s);
	}
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
