package qoca;
// $Id: QcDelCoreTableau.java.cpp,v 1.4 2000/11/21 05:35:16 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//============================================================================//

import java.util.*;
import java.io.*;

#include "QcDefines.jh"

class QcDelCoreTableau extends QcCoreTableau
{
	public QcSparseMatrix fM;		// Quasi-inverse matrix.
	public QcSparseMatrix fA; 		// Original constraint matrix.

	private Vector fCoeffCache;

	//-----------------------------------------------------------------------//
	// Constructor.                                                          //
	//-----------------------------------------------------------------------//
	public QcDelCoreTableau(int hintRows, int hintCols,
			QcQuasiRowColStateVector rcs)
	{
		super(rcs);
		fM = new QcSparseMatrix();
		fA = new QcSparseMatrix();
		fCoeffCache = new Vector(20);
		initialize(hintRows, hintCols);
	}

	public QcDelCoreTableau(DataInputStream s, QcQuasiRowColStateVector rcs)
	{
		super();
		loadFromBinary(s, rcs);
		fCoeffCache = new Vector(20);
	}

	public QcDelCoreTableau(StringTokenizer token, QcQuasiRowColStateVector rcs)
	{
		super();
		loadFromString(token, rcs);
		fCoeffCache = new Vector(20);
	}

	//-----------------------------------------------------------------------//
	// Query functions.                                                      //
	//-----------------------------------------------------------------------//
	public double getRHSVirtual(int ci)
	{
		if (getMRowDeleted(ci))
			return 0.0f;

		double sum = 0.0f;

		for (int i = 0; i < fRows; i++)
			sum += fM.getValue(ci, i) * getRawRHS(i);

		return QcUtility.zeroise(sum);
	}

	public double getValueVirtual(int ci, int vi)
	{
		if (getMRowDeleted(ci))
			return 0.0f;

		QcSparseMatrixColIterator varcol = new QcSparseMatrixColIterator(fA, vi);
		double sum = 0.0f;

		while (!varcol.atEnd()) {
			sum += fM.getValue(ci,varcol.fIndex) * varcol.fValue;
			varcol.increment();
		}

		return QcUtility.zeroise(sum);
	}

	public boolean isFree(int vi)
		// Indicates if a particular variable is in use (i.e. has a
		// non-zero coefficient in any solved form (or original) constraint.
	{
		QcSparseMatrixColIterator varCoeffs =
       			new QcSparseMatrixColIterator(fA, vi);
		return (varCoeffs.atEnd());
	}

	public boolean isValidSolvedForm()
		// For debugging purposes to ensure that the real solved form is
		// consistant with the virtual solved form.
	{
		for (int i = 0; i < fRows; i++) {
			for (int j = 0; j < fColumns; j++) {
     			if ((!QcUtility.isZero(getValueVirtual(i, j)) &&
              			QcUtility.isZero(getValue(i, j))) ||
					(!QcUtility.isZero(getValue(i, j)) &&
  	             		Math.abs(getValueVirtual(i, j) - getValue(i, j)) > 0.01))
          			return false;

      		if ((!QcUtility.isZero(getRHSVirtual(i)) &&
					QcUtility.isZero(getRHS(i))) ||
               			(!QcUtility.isZero(getRHS(i)) &&
               			Math.abs(getRHSVirtual(i) - getRHS(i)) > 0.01))
				return false;
			}
		}

		return true;
	}

	//-----------------------------------------------------------------------//
	// Manipulation functions.                                               //
	//-----------------------------------------------------------------------//
	public int addRow(QcTableauRowIterator varCoeffs, double rhs)
		// Add a new row, allocating any new columns indicated.
		// The new row index is returned.
     {
		int r = 0;     // Row of fA, col of fM, CIndex for original constraint
		int quasiRow;	// Row of fM, CIndex for solved form constraint

		// Look for a deleted row or create a new row if one not found
		while (r < fRows && !getARowDeleted(r))
			r++;

		if (r == fRows) {
			r = increaseRows();
			quasiRow = getMRowIndex(r);
		} else {
			quasiRow = getMRowIndex(r);
			fM.setValue(quasiRow, r, 1.0f);
			setARowDeleted(r, false);
			setMRowDeleted(quasiRow, false);
			linkRow(quasiRow);
		}

		// Load the new values into the fA matrix and RHS
		setRawRHS(r, rhs);
		fSF.fRHS.fData[quasiRow] = rhs;

		while (!varCoeffs.atEnd()) {
			fA.setValue(r, varCoeffs.fIndex, varCoeffs.fValue);
			fSF.setValue(quasiRow, varCoeffs.fIndex, varCoeffs.fValue);
			varCoeffs.increment();
		}

		return r;
	}

	public void addScaledRow(int destRow, int srcRow, double factor)
	{
		super.addScaledRow(destRow, srcRow, factor);
		fM.addScaledRow(destRow, srcRow, factor);
	}

	public void changeRHS(int ci, double rhs)
		// Change the RHS of the constraint ci.
	{
		double diff = rhs - getRawRHS(ci);

		if (!QcUtility.isZero(diff)) {		// Avoid doing redundance work.
			int nrows = getRows();
			for (int i = 0; i < nrows; i++) {
				if (!getMRowDeleted(i))
					fSF.increaseRHS(i, fM.getValue(i, ci) * diff);
			}
		}

		super.changeRHS(ci, rhs);
	}

	public boolean check() // Check data integrity (assertions)
	{
		// Check matrices and vectors are the right size
		if (!(fA.getRows() == fRows && fA.getColumns() == fColumns &&
				fM.getRows() == fRows && fM.getColumns() == fRows))
			return false;

		return super.check();
	}

	public void decreaseColumns(QcBiMapNotifier note)
  	// Get rid of the last (empty) column
	{
		fColumns--;
		fA.resize(fRows, fColumns);
		fColState.resize(fColumns);
		note.dropVariable(fColumns);
		fSF.resize(fRows, fColumns);
	}

#if 0 /* Not used.  TODO: Should we be worried by the fact that this isn't used?
         For example, deleteRow perhaps ought to call it. */

	/** Get rid of the last row - provided it is a deleted row for
	    both original and solved form constraint.
	    N.B. The deleted solved form and original constaint need
	    not correspond to each other.

	    @precondition <tt>fRows</tt> &ge; 1
	**/
	public void decreaseRows(QcBiMapNotifier note)
	{
		qcAssertPre(fRows >= 1);
		int AIndex = getARowIndex(fRows - 1);
		int MIndex = getMRowIndex(fRows - 1);

		// If the last (deleted) original and solved form constraints
		// don't correspond then swap the indices around so there will
		// be no invalid indices after this routine finishes.
		if (AIndex != fRows - 1) {
			setMRowIndex(AIndex, MIndex);
			setARowIndex(MIndex, AIndex);
		}

		fRows--;
		fM.resize(fRows, fRows);
		fA.resize(fRows, fColumns);
		fRowState.resize(fRows);
		fOrigRowState.resize(fRows);
		note.dropConstraint(fRows);
		fSF.resize(fRows, fColumns);
	}
#endif

	public int deleteRow(int ci)
		// Deletes a row given an original constraint index by setting all
		// the coefficients to zero and marking the row as deleted.
		// Derived classes use this method to delete constraints
		// which are found to be inconsistant after a successful AddEq
		// (i.e. unable to successfully pivot) and also used to remove a
		// constraint after is has been removed from the basis.
	{
		int quasiRow = getMRowIndex(ci);
		setMRowDeleted(quasiRow, true);
		setARowDeleted(ci, true);
		fM.zeroColumn(ci);
		fM.zeroRow(quasiRow);
		fA.zeroRow(ci);
		fSF.zeroRow(quasiRow);
		setRawRHS(ci, 0.0f);
		fRowState.unlinkRow(quasiRow);
		return quasiRow;
	}

	public int increaseColumns()	// Creates a new non-basic variable
	{
		if (fColumns == fAllocColumns) {
			fAllocColumns += (fAllocColumns == 0) ? 1 : fAllocColumns;
			fA.reserve(fAllocRows, fAllocColumns);
			fSF.reserve(fAllocRows, fAllocColumns);
			fColState.reserve(fAllocColumns);
		}

		fColumns++;
		fA.resize(fRows, fColumns);
		fA.zeroColumn(fColumns - 1);
		fColState.resize(fColumns);
		fSF.resize(fRows, fColumns);

		// Set the default data values for the newly added row
		fSF.zeroColumn(fColumns - 1);
		fA.zeroColumn(fColumns - 1);

		return fColumns - 1;
	}

	public int increaseRows()	// Creates a new redundant constraint
	{
		if (fRows == fAllocRows) {
			fAllocRows += (fAllocRows == 0) ? 1 : fAllocRows;
			fA.reserve(fAllocRows, fAllocColumns);
			fM.reserve(fAllocRows, fAllocRows);
			fSF.reserve(fAllocRows, fAllocColumns);
			fRowState.reserve(fAllocRows);
			fOrigRowState.reserve(fAllocRows);
		}

		fRows++;			     	// logically, these 3 lines are where the
		fA.resize(fRows, fColumns);	// number of rows increases, the optimiser
		fM.resize(fRows, fRows);		// should cache the (fRows-1) that follow.
		fSF.resize(fRows, fColumns);
		fRowState.resize(fRows);
		fOrigRowState.resize(fRows);

		// Set the default data values for the newly added row
		fA.zeroRow(fRows - 1);
		fM.zeroRow(fRows - 1);
		fM.zeroColumn(fRows - 1);
		fM.setValue(fRows - 1, fRows - 1, 1.0f);
		fSF.zeroRow(fRows - 1);
		fRowState.linkRow(fRows - 1);

		return fRows - 1;
	}

	public void initialize(int hintRows, int hintCols)
		// Call this after construction
	{
		super.initialize(hintRows, hintCols);
		fM.reserve(hintRows, hintRows);
		fA.reserve(hintRows, hintCols);
	}

	public void normalize(int ci)
		// Normal the corresponding row in fM and fSF if the rhs is negative.
		// This makes rhs positive.
	{
		if (QcUtility.isNegative(getRHS(ci))) {
			fM.scaleRow(ci, -1.0f);
			fSF.scaleRow(ci, -1.0f);
		}
	}

	public void pivotQuasiInverse(int row, int col)
	{
		double p = fM.getValue(row, col);

		scaleRow(row, 1.0f / p);

		// Cache column coefficients because it is problematic due to the
		// possible change of column coefficients.
		QcSparseMatrixColIterator varCoeffs =
       		new QcSparseMatrixColIterator(fM, col);
		fCoeffCache.removeAllElements();

		while (!varCoeffs.atEnd()) {
			fCoeffCache.addElement(
            		new QcSparseCoeff(varCoeffs.fValue, varCoeffs.fIndex));
			varCoeffs.increment();
		}

		Enumeration cIt = fCoeffCache.elements();

		while (cIt.hasMoreElements()) {
			QcSparseCoeff coeff = (QcSparseCoeff)cIt.nextElement();

			if (coeff.fIndex != row)
				addScaledRow(coeff.fIndex, row, -coeff.fValue);
		}
	}

	public void restart() // Erase everything ready to start afresh.
	{
		super.restart();
		fM.resize(0, 0);
		fA.resize(0, 0);
	}

	public void restore()
		// The tableau is reinitialised as if all the original constraints
		// had been added to a fresh tableau using AddEq but with no
		// Elimination or Pivoting having been done.
	{
		int co;		// original constraint index (row of fA)
		int cs;		// solved form constraint index (row of fM)

		// Reset constraint state information
		for (cs = 0, co = 0; cs < fRows; cs++, co++) {
			// Find next undeleted constraint
			while (cs < fRows && getMRowDeleted(cs))
				cs++;

			while (co < fRows && getARowDeleted(co))
				co++;

			if (co < fRows) {
				fM.zeroRow(cs);
				fM.zeroColumn(co);
				fM.setValue(cs, co, 1.0f);
				fSF.copyRow(cs, new QcSparseMatrixRowIterator(fA, co));
				fSF.setRHS(cs, fOrigRowState.getRHS(co));
				setMRowIndex(co, cs);
				setARowIndex(cs, co);
				// N.B. We permit Reset to change the co<->cs relationship for
				// a previously invalid constraint.  As long as they
				// are empty it doesn't matter which are used.
			} else
				break;
		}

		// Fix row and column linked lists.
		fRowState.fixLinkage();
		fColState.fixLinkage();
	}

	public void scaleRow(int row, double factor)
	{
		super.scaleRow(row, factor);
		fM.scaleRow(row, factor);
	}

	public void swapColumns(int c1, int c2)
	{
		super.swapColumns(c1, c2);
		fA.swapColumns(c1, c2);
	}

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//
	public void loadFromBinary(DataInputStream s, QcQuasiRowColStateVector rcs)
	{
		super.loadFromBinary(s, rcs);
 		fM = new QcSparseMatrix(s);
		fA = new QcSparseMatrix(s);
	}

	public void loadFromString(StringTokenizer token, QcQuasiRowColStateVector rcs)
	{
		super.loadFromString(token, rcs);
 		fM = new QcSparseMatrix(token);
		fA = new QcSparseMatrix(token);
	}

	public void print()
	{
		System.out.println("Begin[CoreTableau]");

		fOrigRowState.print();
		fRowColState.print();
		System.out.println();

		System.out.println("Begin[Quasi Inverse]");
		fM.print();
		System.out.println("End[Quasi Inverse]");
		System.out.println();

		System.out.println("Begin[Original Constraint]");
		fA.print();
		System.out.println("End[Original Constraint]");
		System.out.println();

		System.out.println("Begin[Solved Form]");
		fSF.print();
		System.out.println("End[Solved Form]");
		System.out.println();

		System.out.println("End[CoreTableau]");
	}

	public void transitiveClosure(QcIntVector vars)
		// Compute the transitive closure of a set of variables, using the
		// relation "belong to the same original constraint".
	{
		boolean checkAgain = true;

		if (fColumns == 0 || fRows == 0)
			return;

		QcSparseMatrixColIterator varCol = new QcSparseMatrixColIterator(fA, 0);
		QcSparseMatrixRowIterator varRow = new QcSparseMatrixRowIterator(fA, 0);
		int sfSize = fColumns;
		boolean scatchFlags[] = new boolean[sfSize];
		int size = vars.size();

		for (int i = 0; i < sfSize; i++)
			scatchFlags[i] = false;

		for (int i = 0; i < size; i++)
			scatchFlags[vars.fData[i]] = true;

		while (checkAgain) {
			checkAgain = false;

			for (int i = 0; i < sfSize; i++) {
				if (scatchFlags[i]) {
					varCol.setToBeginOf(i);

					while (!varCol.atEnd()) {
						varRow.setToBeginOf(varCol.fIndex);
						while (!varRow.atEnd()) {
							if (!scatchFlags[varRow.fIndex]) {
								scatchFlags[varRow.fIndex] = true;
								vars.addElement(varRow.fIndex);
								checkAgain = true;
							}
							varRow.increment();
						}
						varCol.increment();
					}
				}
			}
		}
	}

	public void saveAsBinary(DataOutputStream s)
	{
		super.saveAsBinary(s);
		fM.saveAsBinary(s);
		fA.saveAsBinary(s);
	}

	public void saveAsText(OutputStream s)
	{
		super.saveAsText(s);
		fM.saveAsText(s);
		fA.saveAsText(s);
	}

	public void saveToString(StringBuffer s)
	{
		super.saveToString(s);
		fM.saveToString(s);
		fA.saveToString(s);
	}
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
