package qoca;
// $Id: QcFloat.java.cpp,v 1.10 2000/11/21 05:35:16 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//============================================================================//

import java.util.*;
import java.io.*;

/** A variable having a <code>double</code> value; also has a desired value and
    weights for goal-seeking purposes. */
public class QcFloat
	implements Cloneable, QcDupable, Serializable
{
	protected final static boolean fCheckPre = QcConstants.fCheckPre;
	private final static double fDefaultStayWeight = 1.0f;
	private final static double fDefaultEditWeight = 10000.0f;

	protected static int fNextValidId = 1;
		// fInvalidId is currently set to 0. All negative values are reserved
    		// for future use.

	protected double fStayWeight;
	protected double fEditWeight;

	protected int fId;			// used as index key

	// TODO: The below vars used to be public.  Possibly some of them (fName?) can
	// be made public again.  In any case, need to provide access functions.

	/** True iff var is restricted to non-negative values.  Can be set only
            at construction time.  Note: <tt>fRestricted</tt> is currently never
            used except to write its value in <tt>saveToBinary</tt>,
            <tt>saveToString</tt>, etc. */
	boolean fRestricted;
	public double fValue;
	public double fWeight;
	public double fDesireValue;
	public String fName;

	//-----------------------------------------------------------------------//
	// Constructors.                                                         //
	//-----------------------------------------------------------------------//
	public QcFloat()
		{ initialize("", 0.0f, fDefaultStayWeight, fDefaultEditWeight, false); }

	public QcFloat(boolean restrictedToNonNegative)
	{
		initialize("", 0.0f, fDefaultStayWeight, fDefaultEditWeight,
			restrictedToNonNegative);
	}

	public QcFloat(String name)
	{
		initialize(name, 0.0f, fDefaultStayWeight, fDefaultEditWeight, false);
	}

	public QcFloat(String name, boolean restrictedToNonNegative)
	{
		initialize(name, 0.0f, fDefaultStayWeight, fDefaultEditWeight,
			restrictedToNonNegative);
	}

	public QcFloat(String name, double desval)
	{
		initialize(name, desval, fDefaultStayWeight, fDefaultEditWeight,
			false);
	}

	public QcFloat(String name, double desval, boolean restrictedToNonNegative)
	{
		initialize(name, desval, fDefaultStayWeight, fDefaultEditWeight,
			restrictedToNonNegative);
	}

	/**
	   @param desval Desired value.
	   @param sw Stay weight.
	   @param ew Edit weight.
	**/
	public QcFloat(String name, double desval, double sw, double ew)
		{ initialize(name, desval, sw, ew, false); }

	public QcFloat(String name, double desval, double sw, double ew,
     		boolean restrictedToNonNegative)
		{ initialize(name, desval, sw, ew, restrictedToNonNegative); }

	public QcFloat(DataInputStream s)
		{ loadFromBinary(s); }

	public QcFloat(StringTokenizer token)
		{ loadFromString(token); }

#if qcCheckInternalInvar
	public void assertInvar()
	{
		qcAssertInvar (fId > 0);
		qcAssertInvar (fStayWeight >= 0);
		qcAssertInvar (fEditWeight >= 0);
		qcAssertInvar (fWeight >= 0);
	}
#endif

	//-----------------------------------------------------------------------//
	// Query functions.                                                      //
	//-----------------------------------------------------------------------//

	/** Returns the value assigned by the last solve/resolve/reset operation. */
	public double getSolvedValue() { return fValue; }
	public double getEditWeight() { return fEditWeight; }
	public double getStayWeight() { return fStayWeight; }
	public double getDesiredValue() { return fDesireValue; }
	public String getName() { return fName; }
	public boolean isRestricted() { return fRestricted; }

	public double editWeight()
		{ return fEditWeight; }

	public int intDesireValue()
		{ return Math.round((float)fDesireValue); }

	public int intValue()
		{ return Math.round((float)fValue); }

	public long longDesireValue()
		{ return Math.round(fDesireValue); }

	public long longValue()
		{ return Math.round(fValue); }

	public double stayWeight()
		{ return fStayWeight; }

	//-----------------------------------------------------------------------//
	// Manipulatiom functions.                                               //
	//-----------------------------------------------------------------------//
	public void initialize(String name, double dv, double sw, double ew,
  		boolean restricted)
	{
		if (fCheckPre) { //#if qcCheckPre
			if (sw < 0.0 || ew < 0.0)
				throw new PreconditionException("negative weight; sw=" + sw + ", ew=" + ew);
		} //#endif
		fId = fNextValidId++;
		fRestricted = restricted;
		fWeight = sw;
		fDesireValue = dv;
		fName = name.trim();
		fStayWeight = sw;
		fEditWeight = ew;
		fValue = fDesireValue;
	}

	public void reset()
		// Set implementation variables to defaults
		{ fId = fNextValidId++; }

	/** For nomadic variables (i.e.&nbsp;not <tt>QcFixedFloat</tt>s), set
	    the desired value to the solved value.  For <tt>QcFixedFloat</tt>,
	    does nothing.
	**/
	public void restVariable()
		{ fDesireValue = fValue; }

	public void setToEditWeight()
		{ fWeight = fEditWeight; }

	void setName(String n)
		{ fName = n; }

	public void setToStayWeight()
		{ fWeight = fStayWeight; }

	public void setVariable()
		{ fValue = fDesireValue; }

	void setValue(double w)
		{ fValue = w; }

	public void setWeight(double w)
		{ fWeight = w; }

	public void suggestValue(double dv)
		{ fDesireValue = dv; }

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//
	public boolean equal(QcFloat other)
		{ return (this == other); }

	/** Used primarily by Kakapo.  Returns a small positive integer suitable for use in an IntMap. */
	public int getId()
		{ return fId; }

	public QcFloat duplicate(int depth)
	{
		if (depth == NOT_QCFLOAT)
			return this;
		else {
			try {
				return (QcFloat) clone();
			} catch (CloneNotSupportedException e) {
				throw new RuntimeException(e.toString());
			}
		}
	}

	public void loadFromBinary(DataInputStream s)
	{
		fName = QcUtility.readBinaryString(s);
		fDesireValue = QcUtility.readBinaryDouble(s);
		fStayWeight = QcUtility.readBinaryDouble(s);
		fEditWeight = QcUtility.readBinaryDouble(s);
		fRestricted = QcUtility.readBinaryBoolean(s);
		fWeight = fStayWeight;
		fId = fNextValidId++;
	}

	public void loadFromString(StringTokenizer token)
	{
		fName = token.nextToken();
		fDesireValue = QcUtility.doubleValue(token.nextToken());
		fStayWeight = QcUtility.doubleValue(token.nextToken());
		fEditWeight = QcUtility.doubleValue(token.nextToken());
		fRestricted = (token.nextToken().equals("t"));
		fWeight = fStayWeight;
		fId = fNextValidId++;
	}

	public String toString()
	{
		/* We cast to float (which correspondingly reduces the number of
		   digits in the output) because in practice we tend to get only
		   about float accuracy in the values anyway.  If users want the
		   full precision, they can use getValue() etc. */
		return (fName + '#' + fId + '('
			+ (float) fValue + ", " + (float) fDesireValue + ", " + fWeight
			+ ')');
	}

	public void print()
	{
		System.out.print(toString());
	}

	public void saveAsBinary(DataOutputStream s)
	{
		QcUtility.writeBinaryString(s, "n");

		if (fName.equals(""))
			QcUtility.writeBinaryString(s, "$" + fId);
		else
			QcUtility.writeBinaryString(s, fName);

		QcUtility.writeBinaryDouble(s, fDesireValue);
		QcUtility.writeBinaryDouble(s, fStayWeight);
		QcUtility.writeBinaryDouble(s, fEditWeight);
		QcUtility.writeBinaryBoolean(s, fRestricted);
	}

	public void saveAsText(OutputStream s)
	{
		StringBuffer str = new StringBuffer();
		saveToString(str);
		QcUtility.writeTextString(s, str.toString());
	}

	public void saveToString(StringBuffer s)
	{
		s.append("n ");

		if (fName.equals("")) {
			s.append("$");
			s.append(fId);
		} else
			s.append(fName);

		s.append(" ");
		s.append(QcUtility.toShortestString(fDesireValue));
		s.append(" ");
		s.append(QcUtility.toShortestString(fStayWeight));
		s.append(" ");
		s.append(QcUtility.toShortestString(fEditWeight));
		s.append(" ");
		s.append((fRestricted) ? "t" : "f");
	}
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
