package qoca;
// $Id: QcLinEqSolver.java.cpp,v 1.8 2000/11/21 04:26:13 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//----------------------------------------------------------------------------//
// This interface needs to be read in conjuction with that of Solver.         //
//============================================================================//

import java.util.*;
import java.io.*;

public class QcLinEqSolver extends QcSolver
{
	/** Holds the deriv==0 (i.e.&nbsp;optimality) constraints. */
	protected QcLinEqSystem fSubSystem;
	/** Holds user constraints. */
	protected QcDelLinEqSystem fSystem;
	protected Vector fDesValVar;
	protected Vector fSaveDesireValue;
	protected QcIntVector fDepPars;
	protected QcIntVector fVarsByIndex;
	protected QcBooleanVector fDepBasicVars;

	protected QcVariableBiMap fVBiMap;
	protected QcBiMapNotifier fNotifier;
	protected QcConstraintBiMap fOCBiMap;
	protected QcDelLinEqTableau fTableau;

	//-----------------------------------------------------------------------//
	// Constructor.                                                          //
	//-----------------------------------------------------------------------//
	public QcLinEqSolver()
	{
		super();
		fSubSystem = new QcLinEqSystem();
		fSystem = new QcDelLinEqSystem();
		init(QcConstants.fVectorInitCapacity);
	}

	public QcLinEqSolver(int hintNumConstraints, int hintNumVariables)
	{
		super();
		fSubSystem = new QcLinEqSystem(hintNumConstraints, hintNumVariables);
		fSystem = new QcDelLinEqSystem(hintNumConstraints, hintNumVariables);
		init(hintNumVariables);
	}

	private void init(int hintNumVariables)
	{
		fVBiMap = fSystem.fVBiMap;
		fNotifier = fSystem.fNotifier;
		fOCBiMap = fSystem.fOCBiMap;
		fTableau = (QcDelLinEqTableau)fSystem.fTableau;

		fSaveDesireValue = new Vector(hintNumVariables,
					      QcConstants.fVectorCapacityIncrement);
		// Why don't the rest of these use hintNumVariables?
		fDesValVar = new Vector(QcConstants.fVectorInitCapacity,
					QcConstants.fVectorCapacityIncrement);
		fDepPars = new QcIntVector(QcConstants.fVectorInitCapacity,
					   QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
					       QcConstants.fVectorCapacityIncrement);
		fDepBasicVars = new QcBooleanVector(QcConstants.fVectorInitCapacity,
						    QcConstants.fVectorCapacityIncrement);
	}

	//-----------------------------------------------------------------------//
	// Variable management methods                                           //
	//-----------------------------------------------------------------------//
	public void addVar(QcFloat v)
		{ fSystem.addVar(v); }

	public void suggestValue(QcFloat v, double desval)
		{ fSystem.suggestValue(v, desval); }

	public boolean removeVar(QcFloat v)
		{ return fSystem.removeVar(v); }

	public void restSolver()
		{ fSystem.restSolver(); }

	//-----------------------------------------------------------------------//
	// Enquiry functions for variables                                       //
	//-----------------------------------------------------------------------//
	public boolean isRegistered(QcFloat v)
		{ return fSystem.isRegistered(v); }

	public Enumeration getVariables()
		{ return fSystem.getVariables(); }

	public boolean isFree(QcFloat v)
		{ return fSystem.isFree(v); }

	public boolean isBasic(QcFloat v)
		{ return fSystem.isBasic(v); }

	public boolean isDepBasicVar(int vi)
		{ return fDepBasicVars.fData[vi]; }

	public boolean isDepPar(int vi)
	{
		for (int it = 0; it < fDepPars.fCount; it++)
			if (fDepPars.fData[it] == vi)
				return true;

		return false;
	}

	//-----------------------------------------------------------------------//
	// High Level Edit Variable Interface for use by solver clients.         //
	//-----------------------------------------------------------------------//
	public void beginEdit()
	{
		// build a version of fEditVars by index.
		fVarsByIndex.removeAllElements();
		int evSize = fEditVars.size();

		for (int voiIt = 0; voiIt < evSize; voiIt++)
			fVarsByIndex.addElement(fVBiMap.index(
				(QcFloat)fEditVars.elementAt(voiIt)));

		fSubSystem.restart();
		transitiveClosure();	// Calculates fDepPars and fDepBasicVars

		// Create a variable for the desired value of each EditVars.
		// Also set fSubSystem fEditVars to this.fEditVars.
		fDesValVar.setSize(fTableau.getColumns());

		for (int i = 0; i < fVarsByIndex.fCount; i++) {
			int vi = fVarsByIndex.fData[i];
			QcFloat v = new QcFloat("desired value");
			fSubSystem.registerEditVar(v);
			fDesValVar.setElementAt(v, vi);
		}

		// Make a copy of a matrix of coefficients to speed up GenerateKT1.
 		QcSparseMatrix ccache = new QcSparseMatrix();
		ccache.resize(fTableau.getRows(), fTableau.getColumns());
		QcParamVarIndexIterator vIt = new QcParamVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			QcTableauColIterator colCoeffs =
				new QcTableauColIterator(fTableau, vIt.fIndex);

			while (!colCoeffs.atEnd()) {
				ccache.setValue(colCoeffs.fIndex, vIt.fIndex, colCoeffs.fValue);
				colCoeffs.increment();
			}
			vIt.increment();
		}

		// Generate subsystem KT constraint for each dependent parameter
		// Also set fSystem fEditVars to fDepPars.
		fSystem.clearEditVars();

		for (int viIt = 0; viIt < fDepPars.fCount; viIt++) {
			QcFloat v = fVBiMap.identifier(fDepPars.fData[viIt]);
			// N.B. constraint may be needed even when v is free
			generateKT2(v, ccache);
			fSystem.registerEditVar(v);
		}

		fSubSystem.beginEdit();
		fEditVarsSetup = true;
	}

	public void endEdit()
	{
		for (int vi = 0; vi < fDepBasicVars.fCount; vi++)
			if (fDepBasicVars.fData[vi])
				fVBiMap.identifier(vi).restVariable();

		for (int viIt = 0; viIt < fDepPars.fCount; viIt++)
			fVBiMap.identifier(fDepPars.fData[viIt]).restVariable();

		super.endEdit();
	}

	//-----------------------------------------------------------------------//
	// Constraint management methods                                         //
	//-----------------------------------------------------------------------//
	public boolean addConstraint(QcConstraint c)
	{
		boolean success = fSystem.addConstraint(c);
		if (success) {
			if (fCheckPost) {//#ifdef qcCheckPost
				addCheckedConstraint(c);
			} //#endif
			if (fAutoSolve)
				solve();
		}
		return success;
	}

	public boolean addConstraint(QcConstraint c, QcFloat hint)
	{
		boolean success = fSystem.addConstraint(c, hint);
		if (success) {
			if (fCheckPost) { //#ifdef qcCheckPost
				addCheckedConstraint(c);
			} //#endif
			if (fAutoSolve)
				solve();
		}
		return success;
	}

	public boolean changeConstraint(QcConstraint c, double rhs)
	{
		boolean success = fSystem.changeConstraint(c, rhs);
		if (fCheckPost) { //#ifdef qcCheckPost
			if (success)
				changeCheckedConstraint(c, rhs);
		} //#endif
		return success;
	}

	public boolean removeConstraint(QcConstraint c)
	{
		boolean success = fSystem.removeConstraint(c);

		if (success) {
			if (fCheckPost) { //#ifdef qcCheckPost
				removeCheckedConstraint(c);
			} //#endif
			if (fAutoSolve)
				solve();
		}
		return success;
	}

	public boolean reset()
		{ return fSystem.reset(); }

	//-----------------------------------------------------------------------//
	// Constraint Solving methods                                            //
	//-----------------------------------------------------------------------//
	public void solve()
		// IMPORTANT NOTE.
		// It is is not necessary to call QcLinEqSystem::solve() before
		// creating and solving the subsystem since the subsystem
		// is created without using the values of any variables - just
		// desired values and weights.
	{
		// Make a copy of a matrix of coefficients to speed up GenerateKT1.
 		QcSparseMatrix ccache = new QcSparseMatrix();
		ccache.resize(fTableau.getRows(), fTableau.getColumns());
		QcParamVarIndexIterator vIt = new QcParamVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			QcTableauColIterator colCoeffs
				= new QcTableauColIterator(fTableau, vIt.fIndex);

			while (!colCoeffs.atEnd()) {
				ccache.setValue(colCoeffs.fIndex, vIt.fIndex, colCoeffs.fValue);
				colCoeffs.increment();
			}
			vIt.increment();
		}

		// Generate a subsystem constraint for each parametric variable.
		fSubSystem.restart();

		vIt.reset();

		while (!vIt.atEnd()) {
			// The test avoids entering a redundant constraint in
			// the subsystem to save time.  Also, there is no need to
			// call v.Set when the variable is free since this will be
			// done when LinEq::Solve() is called later.
			if (!fTableau.isFree(vIt.fIndex))
				generateKT1(fVBiMap.identifier(vIt.fIndex), ccache);
			vIt.increment();
		}

		fSubSystem.solve();

		// At this point we have final values for the non-free parametric
		// variables.  Set the desired values and solve the original system
		// again to get the values for the basic variables.
		// Make a record of the altered desired values.
		QcBasicVarIndexIterator bvIt =
			new QcBasicVarIndexIterator(fSubSystem.fTableau);

		while (!bvIt.atEnd()) {
			QcFloat v = fSubSystem.fVBiMap.identifier(bvIt.fIndex);
			fSaveDesireValue.addElement(new QcDesireValueStore(v));
			// save the desired value
			v.suggestValue(v.fValue);
			bvIt.increment();
		}

		fSystem.solve();
		int sdvSize = fSaveDesireValue.size();

		// Restore the altered desired values
		for (int i = 0; i < sdvSize; i++)
			((QcDesireValueStore)fSaveDesireValue.elementAt(i)).restore();

		fSaveDesireValue.removeAllElements();
		// And now the variable values provide the solution.

		if (fCheckPost) { //#ifdef qcCheckPost
			checkSatisfied();
		} //#endif
	}

	public void resolve()
	{
		if (!fEditVarsSetup)
			beginEdit();

		// Setup the desired value variables.
		for (int i = 0; i < fVarsByIndex.fCount; i++) {
			int vi = fVarsByIndex.fData[i];
			((QcFloat)fDesValVar.elementAt(vi)).suggestValue(
            		((QcFloat)fEditVars.elementAt(i)).fDesireValue);
		}

		fSubSystem.resolve();

		// The dependent variables have their desired values altered
		// according to the value found by the subsystem solve.
		for (int vIt = 0; vIt < fDepPars.fCount; vIt++) {
			QcFloat v = fVBiMap.identifier(fDepPars.fData[vIt]);
			fSaveDesireValue.addElement(new QcDesireValueStore(v));
            	// save the desired value
			v.fDesireValue = v.fValue;
		}

		fSystem.resolve();
		int sdvSize = fSaveDesireValue.size();

		// Restore the altered desired values
		for (int i = 0; i < sdvSize; i++)
			((QcDesireValueStore)fSaveDesireValue.elementAt(i)).restore();

		fSaveDesireValue.removeAllElements();

		if (fCheckPost) { //#ifdef qcCheckPost
			checkSatisfied();
		} //#endif
	}

	/** Returns the current value of objective function.  Note that this
	    function is not currently used within QOCA, so is subject to
	    removal; let the QOCA maintainer know if you use this function and
	    want it to stay.
	**/
	public double objective()
	{
		double opt = 0.0f;

		QcVariableIndexIterator vIt = new QcVariableIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			QcFloat v = fVBiMap.identifier(vIt.fIndex);
			double delta = v.fValue - v.fDesireValue;
			opt += v.fWeight * delta * delta;
			vIt.increment();
		}

		return opt;
	}

	//-----------------------------------------------------------------------//
	// Sub-system specific functions.                                        //
	//-----------------------------------------------------------------------//

	/** Generate the constraint df'/dp == 0, where f' = least squares
	    objective function with basic variables eliminated.  Add the
	    constraint to the subsystem fSubSystem.  It is important that this
	    routine does not rely on the values of any of the variables being a
	    solution of LinEq.  See note in <tt>QcLinEqSolver::Solve</tt>.

	    @precondition <tt>p != null</tt>
	**/
	protected void generateKT1(QcFloat p, QcSparseMatrix ccache)
	{
		QcLinPoly lpoly = new QcLinPoly();
		double rhs = 0.0f;
		int pIndex = fVBiMap.index(p);
		double pWeight2 = 2.0f * p.fWeight;
		double pAccumulator = pWeight2;

		QcBasicVarIndexIterator bvIt =
			new QcBasicVarIndexIterator(fTableau);

		while (!bvIt.atEnd()) {
			QcFloat bv = fVBiMap.identifier(bvIt.fIndex);
			int c = bvIt.fCurrent.fIsBasicIn;	// use constraint c to eliminate bv
	 		double pCoeff = ccache.getValue(c, pIndex);
			double factor = 2.0f * bv.fWeight * pCoeff;
			pAccumulator += factor * pCoeff;

			QcSparseMatrixRowIterator varCoeffs =
				new QcSparseMatrixRowIterator(ccache, c);

			while (!varCoeffs.atEnd()) {
				QcFloat pv = fVBiMap.identifier(varCoeffs.fIndex);

				if (varCoeffs.fIndex != pIndex)
					lpoly.plus(
						new QcLinPolyTerm(factor * varCoeffs.getValue(), pv));

				varCoeffs.increment();
			}

			rhs += factor * (fTableau.getRHS(c) - bv.fDesireValue);
			bvIt.increment();
		}

		lpoly.addTerm(new QcLinPolyTerm(pAccumulator, p));
		rhs += pWeight2 * p.fDesireValue;
		fSubSystem.addConstraint(
			new QcConstraint("", lpoly, QcConstraint.fOpEqual, rhs));
	}

	protected void generateKT2(QcFloat p, QcSparseMatrix ccache)
		// Generate the constraint df'/dp == 0 where parameters of interest
		// have variable desired values and where f' = least squares
		// objective function with basic variables eliminated.
		// Add the constraint to the subsystem fSubSystem.
	{
		QcLinPoly lpoly = new QcLinPoly();
       	// N.B. lpoly is destroyed when this routine returns
		double rhs = 0.0f;
		int pIndex = fVBiMap.index(p);
		double pWeight2 = 2.0f * p.fWeight;
		double pAccumulator = pWeight2;
		QcFloat pDV = (QcFloat)fDesValVar.elementAt(pIndex);

		QcBasicVarIndexIterator bvIt =
       		new QcBasicVarIndexIterator(fTableau);

		while (!bvIt.atEnd()) {
			int bvi = bvIt.fIndex;
			QcFloat bv = fVBiMap.identifier(bvi);
			int c = bvIt.fCurrent.fIsBasicIn;	// use constraint c to eliminate bv
	 		double pCoeff = ccache.getValue(c, pIndex);
			double factor = 2.0f * bv.fWeight * pCoeff;
			pAccumulator += factor * pCoeff;

			QcSparseMatrixRowIterator varCoeffs =
            		new QcSparseMatrixRowIterator(ccache, c);

			while (!varCoeffs.atEnd()) {
				QcFloat pv = fVBiMap.identifier(varCoeffs.fIndex);

				if (varCoeffs.fIndex != pIndex)
					lpoly.plus(
						new QcLinPolyTerm(factor * varCoeffs.getValue(), pv));

				varCoeffs.increment();
			}

			if (isDepBasicVar(bvi) && isEditVar(bv)) {
				rhs += factor * fTableau.getRHS(c);
				lpoly.plus(new QcLinPolyTerm(factor,
					(QcFloat)fDesValVar.elementAt(bvi)));
			} else
				rhs += factor * (fTableau.getRHS(c) - bv.fDesireValue);

			bvIt.increment();
		}

		lpoly.addTerm(new QcLinPolyTerm(pAccumulator, p));

		if (isEditVar(p))
			lpoly.addTerm(new QcLinPolyTerm(-pWeight2, pDV));
		else
			rhs += pWeight2 * p.fDesireValue;

		fSubSystem.addConstraint(
			new QcConstraint("", lpoly, QcConstraint.fOpEqual, rhs));
	}

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//
	public QcConstraint getConstraint(String n)
		{ return fSystem.getConstraint(n); }

	public QcFloat getVariable(String n)
		{ return fSystem.getVariable(n); }

	public void getVariableSet(QcVariableSet vars)
		{ fSystem.getVariableSet(vars); }

	public void loadFromBinary(DataInputStream s, QcVariableSet vars)
	{
		super.loadFromBinary(s, vars);
		fSystem.loadFromBinary(s, vars);

		fVBiMap = fSystem.fVBiMap;
		fNotifier = fSystem.fNotifier;
		fOCBiMap = fSystem.fOCBiMap;
		fTableau = (QcDelLinEqTableau)fSystem.fTableau;
	}

	public void loadFromString(StringTokenizer token, QcVariableSet vars)
	{
		super.loadFromString(token, vars);
		fSystem.loadFromString(token, vars);

		fVBiMap = fSystem.fVBiMap;
		fNotifier = fSystem.fNotifier;
		fOCBiMap = fSystem.fOCBiMap;
		fTableau = (QcDelLinEqTableau)fSystem.fTableau;
	}

	public void print()
	{
		fSystem.print();

 		if (fEditVarsSetup) {
			System.out.println("Sub-System: ");
			fSubSystem.print();
			System.out.println();
			System.out.println("Dependant basic variables: ");

			for (int i = 0; i < fDepBasicVars.fCount; i++)
				if (fDepBasicVars.fData[i])
					System.out.println(i + ", ");

			System.out.println();
			System.out.println("Dependant parameter variables: ");

			for (int i = 0; i < fDepPars.fCount; i++)
				System.out.println(fDepPars.fData[i] + ", ");

			System.out.println();
		}
	}

	public void saveAsBinary(DataOutputStream s, QcVariableSet vars)
	{
		super.saveAsBinary(s, vars);
		fSystem.saveAsBinary(s, vars);
	}

	public void saveAsText(OutputStream s, QcVariableSet vars)
	{
		super.saveAsText(s, vars);
		fSystem.saveAsText(s, vars);
	}

	public void saveToString(StringBuffer s, QcVariableSet vars)
	{
		super.saveToString(s, vars);
		fSystem.saveToString(s, vars);
	}

	/** Calculate fDepBasicVars and fDepPars as transitive closure of
	    variables in constraints which mention EditVars. */
	protected void transitiveClosure()
	{
		/* TODO: The C++ version uses a straight assignment.  Check
		   whether that would be sufficient here.  If not, check whether
		   the C++ version should be using a deep copy.

		   Same goes for QcLinInEqSolver.
		*/
		QcIntVector fDepVars = fVarsByIndex.duplicate(QcDupable.DEEP);
		fSystem.transitiveClosure(fDepVars);

		// Shift fDepVars into fDepPars and fDepBasicVars.
		fDepBasicVars.setSize(fTableau.getColumns());

		for (int i = 0; i < fDepBasicVars.fCount; i++)
			fDepBasicVars.fData[i] = false;

		fDepPars.removeAllElements();

		for (int i = 0; i < fDepVars.fCount; i++) {
			int vi = fDepVars.fData[i];

			if (fTableau.isBasic(vi))
				fDepBasicVars.fData[vi] = true;
			else
				fDepPars.addElement(vi);
		}
	}
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
