package qoca;
// $Id: QcLinEqSystem.java.cpp,v 1.7 2000/11/25 13:56:17 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//----------------------------------------------------------------------------//
// This interface needs to be read in conjuction with that of Solver.         //
//============================================================================//

import java.util.*;
import java.io.*;

#include "QcDefines.jh"

class QcLinEqSystem extends QcSolver
{
	protected QcIntVector fDepVars;
	protected QcIntVector fParsOfInterest;
	protected QcIntVector fVarsByIndex;    // Indices of variables of interest.
	protected QcFloatVector fDepVarBaseVals;  // Used for Resolve
	protected Vector fCoeffCache;
	protected Vector fDepVarCoeffs[];
		// fDepVarCoeffs is a partial map from dependant variable index
		// to a vector (used as a list) of sparse coefficients.

	public QcVariableBiMap fVBiMap;
	public QcBiMapNotifier fNotifier;
	public QcConstraintBiMap fOCBiMap;
	public QcLinEqTableau fTableau;

	//-----------------------------------------------------------------------//
	// Constructor.                                                          //
	//-----------------------------------------------------------------------//
	public QcLinEqSystem()
	{
		super();
		fNotifier = new QcBiMapNotifier();
		fVBiMap = fNotifier.fVMap;
		fOCBiMap = fNotifier.fOCMap;
		fTableau = new QcLinEqTableau(0, 0, fNotifier);
		fDepVars = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fParsOfInterest = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fDepVarBaseVals = new QcFloatVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fCoeffCache = new Vector(20);
	}

	public QcLinEqSystem(int hintRows, int hintCols)
	{
		super();
		fNotifier = new QcBiMapNotifier();
		fVBiMap = fNotifier.fVMap;
		fOCBiMap = fNotifier.fOCMap;
		fTableau = new QcLinEqTableau(hintRows, hintCols, fNotifier);
		fDepVars = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fParsOfInterest = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fDepVarBaseVals = new QcFloatVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fCoeffCache = new Vector(20);
	}

	public QcLinEqSystem(int hintRows, int hintCols, QcLinEqTableau tab)
	{
		super();
		fNotifier = tab.fNotifier;
		fVBiMap = fNotifier.fVMap;
		fOCBiMap = fNotifier.fOCMap;
		fTableau = tab;
		fDepVars = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fParsOfInterest = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
	         	QcConstants.fVectorCapacityIncrement);
		fDepVarBaseVals = new QcFloatVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fCoeffCache = new Vector(20);
	}

	public QcLinEqSystem(DataInputStream s, QcVariableSet vars)
	{
		loadFromBinary(s, vars);
		fCoeffCache = new Vector(20);
	}

	public QcLinEqSystem(StringTokenizer token, QcVariableSet vars)
	{
		loadFromString(token, vars);
		fCoeffCache = new Vector(20);
	}

#ifndef NDEBUG
	public void assertInvar()
	{
		super.assertInvar();
		fVBiMap.assertInvar();
		fNotifier.assertInvar();
		fOCBiMap.assertInvar();
		fTableau.assertInvar();
	}
#endif

	//-----------------------------------------------------------------------//
	// Variable management methods (see QcSolver.java for descriptions)      //
	//-----------------------------------------------------------------------//
	public void addVar(QcFloat v)
		// N.B. AddVar does nothing if the variable is already registered.
	{
		if (!fVBiMap.identifierPresent(v)) {
			newVariable(v);
			dbg(assertInvar());
		}
	}

	public void suggestValue(QcFloat v, double desval)
	{
		if (fVBiMap.identifierPresent(v))
			v.suggestValue(desval);
	}

	public void restSolver()
	{
		QcVariableIndexIterator vIt = new QcVariableIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			fVBiMap.identifier(vIt.fIndex).restVariable();
			vIt.increment();
		}
	}

	//-----------------------------------------------------------------------//
	// Enquiry functions for variables                                       //
	//-----------------------------------------------------------------------//
	public boolean isRegistered(QcFloat v)
		{ return (fVBiMap.identifierPresent(v)); }

	public Enumeration getVariables()
		{ return fVBiMap.getIdentifiers(); }

	public boolean isBasic(QcFloat v)
		{ return fTableau.isBasic(fVBiMap.index(v)); }

	// -----------------------------------------------------------------------//
	// High Level Edit Variable Interface for use by solver clients.  The
	// methods <tt>addEditVar</tt> and <tt>endEdit</tt> are more efficient
	// than the ones inherited from <tt>QcSolver</tt> since variable weights
	// have no effect on <tt>QcLinEqSystem.resolve</tt>.
	// -----------------------------------------------------------------------//
	public void beginEdit()
		// Implementation note.
		// It does not appear to be necessary to call solve before
		// beginEdit, so don't depend on this in the first loop.
	{
		// build a version of fEditVars by index.
		fVarsByIndex.removeAllElements();
		int evSize = fEditVars.size();

		for (int voiIt = 0; voiIt < evSize; voiIt++) {
			QcFloat f = (QcFloat)fEditVars.elementAt(voiIt);
			if (isRegistered(f))
				fVarsByIndex.addElement(fVBiMap.index(f));
		}

		// try to make all variables of interest into parameters.
		for (int viIt = 0; viIt < fVarsByIndex.fCount; viIt++) {
			int vi = fVarsByIndex.fData[viIt];
			if (fTableau.isBasic(vi)) {
				int ci = fTableau.isBasicIn(vi);
				// Cache tableau column coefficients since it is possible that
				// the iterator may be invalidated during iteration.
				QcTableauRowIterator varCoeffs =
					new QcTableauRowIterator(fTableau, ci);
				fCoeffCache.removeAllElements();

				while (!varCoeffs.atEnd()) {
					fCoeffCache.addElement(
            				new QcSparseCoeff(varCoeffs.fValue, varCoeffs.fIndex));
					varCoeffs.increment();
				}

				Enumeration cIt = fCoeffCache.elements();

				while (cIt.hasMoreElements()) {
					QcSparseCoeff coeff = (QcSparseCoeff)cIt.nextElement();
					boolean reject = false;

					for (int viIt2 = 0; viIt2 < fVarsByIndex.fCount; viIt2++)
						if (coeff.fIndex == fVarsByIndex.fData[viIt2])
							reject = true;

					if (!reject) {
						fTableau.pivot(ci, coeff.fIndex);
						break;
					}
				}
			}
		}

		solve();							// Compute new solution here.
		fParsOfInterest.removeAllElements(); 	// setup fParsOfInterest

		for (int viIt = 0; viIt < fVarsByIndex.fCount; viIt++) {
			int vi = fVarsByIndex.fData[viIt];
			if (!fTableau.isBasic(vi) && !isParOfInterest(vi))
				fParsOfInterest.addElement(vi);
		}

		fDepVars.removeAllElements(); 		// setup fDepVars

		// setup fDepVarCoeffs.  Extend fDepVarCoeffs so 0..fTableau.Cols()-1
		// are legal indices.
		fDepVarCoeffs = new Vector[fTableau.getColumns()];
		fDepVarBaseVals.setSize(fTableau.getColumns());
		QcBasicVarIndexIterator bvIt = new QcBasicVarIndexIterator(fTableau);

		for (int viIt = 0; viIt < fDepVarCoeffs.length; viIt++)
			fDepVarCoeffs[viIt] = new Vector(QcConstants.fVectorInitCapacity,
				QcConstants.fVectorCapacityIncrement);

		while (!bvIt.atEnd()) {
			int bvi = bvIt.fIndex;
			int ci = bvIt.fCurrent.fIsBasicIn;
			fDepVarBaseVals.fData[bvi] = fTableau.getRHS(ci);
			QcTableauRowIterator varCoeffs =
				new QcTableauRowIterator(fTableau, ci);

			while (!varCoeffs.atEnd()) {
				if (isParOfInterest(varCoeffs.fIndex)) {
					if (!isDepVar(bvi))
						fDepVars.addElement(bvi);
					fDepVarCoeffs[bvi].addElement(
						new QcSparseCoeff(varCoeffs.fValue, varCoeffs.fIndex));
				} else if (bvi != varCoeffs.fIndex) {
					double val = fVBiMap.identifier(varCoeffs.fIndex).fValue;
					fDepVarBaseVals.fData[bvi] -= val * varCoeffs.fValue;
				}
				varCoeffs.increment();
			}
			bvIt.increment();
		}

		fEditVarsSetup = true;
	}

	public void endEdit()
		// Specialised version of EndEdit since LinEqSystem does not adjust
		// weights.
	{
		for (int vIt = 0; vIt < fDepVars.fCount; vIt++)
			fVBiMap.identifier(fDepVars.fData[vIt]).restVariable();

		for (int vIt = 0; vIt < fParsOfInterest.fCount; vIt++)
			fVBiMap.identifier(fParsOfInterest.fData[vIt]).restVariable();

		super.endEdit();
	}

	//-----------------------------------------------------------------------//
	// Constraint management methods                                         //
	//-----------------------------------------------------------------------//

	/** Extracts the various components and calls Insert, as well as
	    updating the bimapnotifier for the tableau if successful.  The
	    return value indicates success or failure.  If
	    <tt>addConstraint</tt> fails then the state of the tableau is
	    unchanged. */
	public boolean addConstraint(QcConstraint c)
	{
		return doAddConstraint(c, null);
	}

	public boolean addConstraint(QcConstraint c, QcFloat hint)
	{
		if (fCheckInternalPre) { //#if qcCheckInternalPre
			if (hint == null)
				throw new InternalPreconditionException("null hint");
		} //#endif
		return doAddConstraint(c, hint);
	}

	private boolean doAddConstraint(QcConstraint c, QcFloat hint)
	{
		/* If in batch mode and the one of more of the preceeding add
		   constraint operation failed, then should not proceed to do
		   anymore constraint addition. */
		if (fBatchAddConst && fBatchAddConstFail)
			return false;

		QcRowAdaptor ra = new QcRowAdaptor(c.fLinPoly, fVBiMap, fTableau);
		boolean success = true;
		int origConsIx; 	// original constraint index
		int solvedConsIx;	// solved form constraint index

		origConsIx = fTableau.addEq(ra, c.fRHS);
		solvedConsIx = fTableau.eliminate(origConsIx);
		if (hint != null) {
			fTableau.fPivotHints.removeAllElements();
			fTableau.fPivotHints.addElement(fVBiMap.index(hint));
		}

		if (solvedConsIx == QcTableau.fInvalidConstraintIndex) {
			fInconsistant.addElement(c);
			success = false;
		} else if (!fTableau.isRedundant(solvedConsIx))
			applyHints(solvedConsIx);

		if (success) {
			fOCBiMap.update(c, origConsIx);
			fBatchConstraints.addElement(c);
		} else if (fBatchAddConst)
			fBatchAddConstFail = true;

		fEditVarsSetup = false;
		dbg(assertInvar());
		return success;
	}

	public boolean changeConstraint(QcConstraint c, double rhs)
	{
		int ix = fOCBiMap.safeIndex(c);
		if (ix < 0)
			return false; // TODO: perhaps throw a PreconditionException
		fTableau.changeRHS(ix, rhs);
		// TODO: if (!fBatchAddConst) { maybe return false; }
		return true;
	}

	public boolean reset()
		// While this routine is in progress the tableau is slowly rebuilt from
		// row 0 forwards retaining the original constraint matrix fA.
		// The implementation used here only relies on the rest of the tableau
		// methods being able to cope with multiple undetermined constraints.
		// Reset starts by calling Restore to establish a state equivalent
		// to the original constraints having been all entered using AddEq
		// and nothing else apart from the deleted entries being left alone.
	{
		boolean result = true;
		int cols = fTableau.getColumns();
		int rows = fTableau.getRows();

		// Record the columns of the previous basic variables so to make them
		// basic after this reset() method.
		boolean basic[] = new boolean[cols];

	     for (int c = 0; c < cols; c++)
	     	basic[c] = fTableau.isBasic(c);

		// Establish solved form and quasi-inverse from original constraints.
		for (int c = 0; c < cols; c++) {
			if (basic[c]) {
				int maxrow = -1;
				double max = 0.0;
				QcTableauColIterator iC =
					new QcTableauColIterator(fTableau, c);

				while (!iC.atEnd()) {
					if (max < Math.abs(iC.fValue)) {
						max = iC.fValue;
						maxrow = iC.fIndex;
					}

					iC.increment();
				}

				// If there is at least one non-zero coefficient in the
				// current column.
				if (maxrow != -1)
					fTableau.pivot(maxrow, c);
			}
		}

		// Make sure that all solved form constraints are assigned a basic
		// variable.
		for (int r = 0; r < rows; r++) {
			if (fTableau.getRowCondition(r) == QcLinEqRowState.fNormalised) {
				int maxcol = -1;
				double max = 0.0;
				QcTableauRowIterator iR =
					new QcTableauRowIterator(fTableau, r);

				while (!iR.atEnd()) {
					if (max < Math.abs(iR.fValue)) {
						max = iR.fValue;
						maxcol = iR.fIndex;
					}

					iR.increment();
				}

				// If there is at least one non-zero coefficient in the
				// current column.
				if (maxcol != -1)
					fTableau.pivot(r, maxcol);
				else {
					// Verify that the Reset operation is successful.  It
					// fails if one of more of the solved form row is
					// marked as "invalid".
					if (QcUtility.isZero(fTableau.getRHS(r)))
						fTableau.setRowCondition(r, QcLinEqRowState.fRedundant);
					else {
						fTableau.setRowCondition(r, QcLinEqRowState.fInvalid);
						result = false;
					}
				}
			}
		}

		return result;
	}

	//-----------------------------------------------------------------------//
	// Constraint Solving methods                                            //
	//-----------------------------------------------------------------------//
	public void solve()
		// Using the tableau's desired values cache, calculate the value
		// for each variable depending upon whether it is basic or parametric.
	{
		loadDesValCache();
		QcVariableIndexIterator vIt = new QcVariableIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			int vi = vIt.fIndex;
			QcFloat v = fVBiMap.identifier(vi);

			if (fTableau.isBasic(vi)) {
				double value = fTableau.evalBasicVar(vi);
				v.fValue = QcUtility.zeroise(value);
			} else
				v.setVariable();

			vIt.increment();
		}
	}

	public void resolve()
	{
		if (!fEditVarsSetup)
			beginEdit();

		// adjust the dependent basic variables
		for (int vIt = 0; vIt < fDepVars.fCount; vIt++) {
			int vi = fDepVars.fData[vIt];
			QcFloat dv = fVBiMap.identifier(vi);
			Vector Coeffs = fDepVarCoeffs[vi];
			double dvValue = fDepVarBaseVals.fData[vi];
			int cfSize = Coeffs.size();

			for (int pIt = 0; pIt < cfSize; pIt++) {
				QcSparseCoeff cv = (QcSparseCoeff)Coeffs.elementAt(pIt);
				QcFloat p = fVBiMap.identifier(cv.fIndex);
				dvValue -= cv.fValue * p.fDesireValue;
			}

			dv.fValue = QcUtility.zeroise(dvValue);
		}

		// adjust the parameters of interest
		for (int vIt = 0; vIt < fParsOfInterest.fCount; vIt++)
			fVBiMap.identifier(fParsOfInterest.fData[vIt]).setVariable();
	}

	//-----------------------------------------------------------------------//
	// Read access to the Tableau and BimapNotifier are a low level aspect of//
	// the public interface.                                                 //
	//-----------------------------------------------------------------------//
	protected void loadDesValCache()
		// Used for Solve
	{
		QcVariableIndexIterator vIt = new QcVariableIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			fTableau.setDesireValue(vIt.fIndex,
            	fVBiMap.identifier(vIt.fIndex).fDesireValue);
			vIt.increment();
		}
	}

	protected boolean applyHints(int cs)
		// ApplyHints calls Pivot according to the PivotHints and should
		// only be called after a successful Eliminate call (i.e. the
		// new constraint is consistent and independent).  If all the
		// PivotHints fail then ApplyHints trys pivoting at all the
		// parametric variables with non-zero coeffs for cs.
	{
		boolean pivoted = false;

		for (int i = 0; i < fTableau.fPivotHints.fCount && !pivoted; i++)
			pivoted = fTableau.pivot(cs, fTableau.fPivotHints.fData[i]);

		// If PivotHints failed, this means all the pivot hints variables
		// have zero coeffs.  Now try all the rest of the parametric
		// variables with non-zero coeffs in the solved form constraint.
		if (!pivoted) {
			// Cache tableau column coefficients since it is possible that
			// the iterator may be invalidated during iteration.
			QcTableauRowIterator varCoeffs =
				new QcTableauRowIterator(fTableau, cs);
			fCoeffCache.removeAllElements();

			while (!varCoeffs.atEnd()) {
				fCoeffCache.addElement(
					new QcSparseCoeff(varCoeffs.fValue, varCoeffs.fIndex));
				varCoeffs.increment();
			}

			// Note this iterator will not return zero coeffs, so
			// there is no need to exclude the hints.
			Enumeration cIt = fCoeffCache.elements();

			while (cIt.hasMoreElements() && !pivoted) {
				QcSparseCoeff coeff = (QcSparseCoeff)cIt.nextElement();

				if (!fTableau.isBasic(coeff.fIndex))
					pivoted = fTableau.pivot(cs, coeff.fIndex);
			}
		}

		return true;
	}

	public boolean isParOfInterest(int vi)
	{
		for (int i = 0; i < fParsOfInterest.fCount; i++)
			if (fParsOfInterest.fData[i] == vi)
				return true;

		return false;
	}

	public boolean isDepVar(int vi)
		// Used by EditVars and Resolve optimised linkage.
	{
		for (int i = 0; i < fDepVars.fCount; i++)
			if (fDepVars.fData[i] == vi)
				return true;

		return false;
	}

	public void newVariable(QcFloat v)
		{ fVBiMap.update(v, fTableau.newVariable()); }

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//
	public QcConstraint getConstraint(String n)
		{ return fOCBiMap.identifier(n); }

	public QcFloat getVariable(String n)
		{ return fVBiMap.identifier(n); }

	public void getVariableSet(QcVariableSet vars)
		{ fVBiMap.getVariableSet(vars); }

	public void loadFromBinary(DataInputStream s, QcVariableSet vars)
	{
		fNotifier = new QcBiMapNotifier(s, vars);
		fVBiMap = fNotifier.fVMap;
		fOCBiMap = fNotifier.fOCMap;
		fTableau = new QcDelLinEqTableau(s, fNotifier);
		fDepVars = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fParsOfInterest = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fDepVarBaseVals = new QcFloatVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		super.loadFromBinary(s, vars);
	}

	public void loadFromBinary(DataInputStream s, QcDelLinEqTableau tab,
		QcVariableSet vars)
	{
		fNotifier = tab.fNotifier;
		fVBiMap = fNotifier.fVMap;
		fOCBiMap = fNotifier.fOCMap;
		fTableau = tab;
		fDepVars = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fParsOfInterest = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fDepVarBaseVals = new QcFloatVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		super.loadFromBinary(s, vars);
	}

	public void loadFromString(StringTokenizer token, QcVariableSet vars)
	{
		fNotifier = new QcBiMapNotifier(token, vars);
		fVBiMap = fNotifier.fVMap;
		fOCBiMap = fNotifier.fOCMap;
		fTableau = new QcDelLinEqTableau(token, fNotifier);
		fDepVars = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fParsOfInterest = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fDepVarBaseVals = new QcFloatVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		super.loadFromString(token, vars);
	}

	public void loadFromString(StringTokenizer token, QcDelLinEqTableau tab,
		QcVariableSet vars)
	{
		fNotifier = tab.fNotifier;
		fVBiMap = fNotifier.fVMap;
		fOCBiMap = fNotifier.fOCMap;
		fTableau = tab;
		fDepVars = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fParsOfInterest = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fDepVarBaseVals = new QcFloatVector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		super.loadFromString(token, vars);
	}

	public void print()
	{
		super.print();
		fNotifier.print();
		System.out.println();
		fTableau.print();
	}

	public void restart()
		// Erase everything ready to start afresh.
	{
		super.restart();
		fTableau.restart();
		fNotifier.restart();
		fDepVars.removeAllElements();
		fParsOfInterest.removeAllElements();
		fDepVarCoeffs = null;
	}

	public void saveAsBinary(DataOutputStream s, QcVariableSet vars)
	{
		fNotifier.saveAsBinary(s, vars);
		fTableau.saveAsBinary(s);
		super.saveAsBinary(s, vars);
	}

	public void saveAsText(OutputStream s, QcVariableSet vars)
	{
		fNotifier.saveAsText(s, vars);
		fTableau.saveAsText(s);
		super.saveAsText(s, vars);
	}

	public void saveToString(StringBuffer s, QcVariableSet vars)
	{
		fNotifier.saveToString(s, vars);
		fTableau.saveToString(s);
		super.saveToString(s, vars);
	}
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
