package qoca;
// $Id: QcLinInEqSolver.java.cpp,v 1.7 2000/11/21 05:35:16 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//----------------------------------------------------------------------------//
// This interface needs to be read in conjuction with that of Solver.         //
//============================================================================//

import java.util.*;
import java.io.*;

#include "QcDefines.jh"

/**
 LinInEqSolver optimizes the square of the Euclidean distance (like LinEqSolver)
 but allows linear inequalities as well as linear equalities.  It is based on
 linear complementary pivoting.
**/
public class QcLinInEqSolver extends QcSolver
{
	protected QcDelLinInEqSystem fSystem;
	protected Vector fDual;			// The complementary variables.
	protected QcIntVector fDualIndex;	// The complementary variable indices.
		// Each variable in the dual problem (fSubSystem) has a complementary
		// variable.  This correspondance is that for each parameter
		// (index j) in the primal problem, there is a corresponding dual
		// constraint and fDual[j] is its slack variable, and for each basic
		// variable in the primal problem (index j) there is a corresponding
		// dual variable fDual[j] (as used in AddDualCoeffs).
		// The total number of variables in the dual problem is hence
		// (primal parametric vars + 2 * primal basic vars + artificial var).
		// The artificial variable is entered as fDual[fAI], though this entry
		// is not used anywhere except for debug state output.
		// For each i, fDualIndex[i] provides the index of fDual[i] in fSubSystem.

	protected QcFloat fAV;	// Artificial variable used for complementary pivot.
	protected int fAI;		// Index of artificial variable in fSubSystem.
	protected QcIntVector fDepPars;
	protected QcIntVector fVarsByIndex;
	protected QcBooleanVector fDepBasicVars;
	protected QcBooleanVector fDepParVars;
	protected Vector fDesValVar;
	protected QcIntVector fDesValIndex;

	protected QcVariableBiMap fVBiMap;
	protected QcBiMapNotifier fNotifier;
	protected QcConstraintBiMap fOCBiMap;
	protected QcDelLinInEqTableau fTableau;

	protected QcCompPivotSystem fSubSystem;
	protected QcCompPivotTableau fSubTableau;
	protected QcSparseMatrix fCacheCoeff;

	protected QcVariableBiMap fSubVBiMap;
	protected QcBiMapNotifier fSubNotifier;
	protected QcConstraintBiMap fSubOCBiMap;

	//-----------------------------------------------------------------------//
	// Constructor.                                                          //
	//-----------------------------------------------------------------------//
	public QcLinInEqSolver()
	{
		super();

		fSystem = new QcDelLinInEqSystem();
		fAV = new QcFloat("Artificial Var");
		fAI = QcTableau.fInvalidVariableIndex;

		fDual = new Vector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDualIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDepPars = new QcIntVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDepBasicVars = new QcBooleanVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDepParVars = new QcBooleanVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDesValIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDesValVar = new Vector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);

		fVBiMap = fSystem.fVBiMap;
		fNotifier = fSystem.fNotifier;
		fOCBiMap = fSystem.fOCBiMap;
		fTableau = (QcDelLinInEqTableau)fSystem.fTableau;

		fSubSystem = new QcCompPivotSystem();
		fSubNotifier = fSubSystem.fNotifier;
		fSubTableau = fSubSystem.fTableau;
		fSubVBiMap = fSubNotifier.fVMap;
		fSubOCBiMap = fSubNotifier.fOCMap;

		fCacheCoeff = new QcSparseMatrix();
	}

	public QcLinInEqSolver(int hintNumConstraints, int hintNumVariables)
	{
		super();

		fSystem = new QcDelLinInEqSystem(hintNumConstraints, hintNumVariables);
		fAV = new QcFloat("Artificial Var");
		fAI = QcTableau.fInvalidVariableIndex;

		fDual = new Vector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDualIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDepPars = new QcIntVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fVarsByIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDepBasicVars = new QcBooleanVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDepParVars = new QcBooleanVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDesValIndex = new QcIntVector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);
		fDesValVar = new Vector(QcConstants.fVectorInitCapacity,
       		QcConstants.fVectorCapacityIncrement);

		fVBiMap = fSystem.fVBiMap;
		fNotifier = fSystem.fNotifier;
		fOCBiMap = fSystem.fOCBiMap;
		fTableau = (QcDelLinInEqTableau)fSystem.fTableau;

		fSubSystem = new QcCompPivotSystem(hintNumConstraints,
       		hintNumVariables);
		fSubNotifier = fSubSystem.fNotifier;
		fSubTableau = fSubSystem.fTableau;
		fSubVBiMap = fSubNotifier.fVMap;
		fSubOCBiMap = fSubNotifier.fOCMap;

		fCacheCoeff = new QcSparseMatrix();
	}

	//-----------------------------------------------------------------------//
	// Variable management methods                                           //
	//-----------------------------------------------------------------------//
	public void addVar(QcFloat v)
		{ fSystem.addVar(v); }

	public void suggestValue(QcFloat v, double desval)
		{ fSystem.suggestValue(v, desval); }

	public boolean removeVar(QcFloat v)
		{ return fSystem.removeVar(v); }

	public void restSolver()
		{ fSystem.restSolver(); }

	//-----------------------------------------------------------------------//
	// Enquiry functions for variables                                       //
	//-----------------------------------------------------------------------//
	public boolean isRegistered(QcFloat v)
		{ return fSystem.isRegistered(v); }

	public Enumeration getVariables()
		{ return fSystem.getVariables(); }

	public boolean isFree(QcFloat v)
		{ return fSystem.isFree(v); }

	public boolean isBasic(QcFloat v)
		{ return fSystem.isBasic(v); }

	public boolean isDepBasicVar(int vi)
		{ return fDepBasicVars.fData[vi]; }

	public boolean isDepParVar(int vi)
		{ return fDepParVars.fData[vi]; }

	public boolean isDepVar(int vi)
		{ return fDepParVars.fData[vi] || fDepBasicVars.fData[vi]; }

	//-----------------------------------------------------------------------//
	// High Level Edit Variable Interface for use by solver clients.         //
	//-----------------------------------------------------------------------//
	public void beginEdit()
	{
		fAI = QcTableau.fInvalidVariableIndex;
		fSubSystem.restart();

		// build a version of fEditVars by index.
		fVarsByIndex.removeAllElements();
		int evSize = fEditVars.size();

		for (int voiIt = 0; voiIt < evSize; voiIt++)
			fVarsByIndex.addElement(fVBiMap.index(
            		((QcFloat)fEditVars.elementAt(voiIt))));

		// Make a copy of the main tableau coefficients to speed up generateKT.
		cacheTableau();

		// Ensure variable indices are the same by adding variables into the
		// sub-tableau.  This may be improved later by just copying the bimaps.
		prepareVariables();

      	// Calculates fDepPars, fDepBasicVars and fDepParVars.
		transitiveClosure();

		// Place constraints with slack basic variables with dependant variable
		// in fDepParVars into fSubLS tableau.
		QcBasicVarIndexIterator bvIt = new QcBasicVarIndexIterator(fTableau);

		while (!bvIt.atEnd()) {
			if (fTableau.isSlack(bvIt.fIndex)) {
				// Check if the current constraint has a dependant parameter
				// variable.
				int c = bvIt.fCurrent.fIsBasicIn;
				boolean hasDepPar = false;
				QcSparseMatrixRowIterator varCoeffs =
                 		new QcSparseMatrixRowIterator(fCacheCoeff, c);

				while (!varCoeffs.atEnd()) {
					if (isDepVar(varCoeffs.fIndex)) {
						hasDepPar = true;
						break;
					}
					varCoeffs.increment();
				}

				// If so, add the solved form constraint into fSubLS tableau.
				if (hasDepPar) {
					varCoeffs.setToBeginOf(c);
					QcLinPoly lpoly = new QcLinPoly();
					lpoly.plus(new QcLinPolyTerm(
						fTableau.getValue(c, bvIt.fIndex),
		              		fVBiMap.identifier(bvIt.fIndex)));

					while (!varCoeffs.atEnd()) {
						QcFloat v = fVBiMap.identifier(varCoeffs.fIndex);
						lpoly.plus(new QcLinPolyTerm(varCoeffs.fValue, v));
						varCoeffs.increment();
					}

					QcConstraint lc = new QcConstraint("", lpoly,
                      		QcConstraint.fOpEqual, fTableau.getRHS(c));
					fSubSystem.copyConstraint(lc, bvIt.fIndex);
				}
			}
			bvIt.increment();
		}

		// This needs the original constraints copied into subsystem.
		// The dual variables are preallocated.  The indices are assigned
		// below.
		createDualVars2();

		// Create a variable for the desired value of each EditVars.
		createDesValVars();

		fAI = fSubTableau.addArtificial(-1.0f);
		fAV.setWeight(0.0f);	// Just in case it somehow get taken notice of
		fDual.addElement(fAV);	// This shouldn't need to be accessed.
		fDualIndex.addElement(fAI);
		fSubVBiMap.update(fAV, fAI);

		fSystem.clearEditVars();	// Add Edit Variables.

		for (int viIt = 0; viIt < fDepPars.fCount; viIt++) {
			QcFloat v = fSubVBiMap.identifier(fDepPars.fData[viIt]);
			fSystem.registerEditVar(v);
		}

		generateKT2();

		// Enter fDualIndex values for non slack variables
		int ssCols = fSubTableau.getColumns();

		for (int vi = 0; vi < ssCols; vi++) {
			QcFloat dualVar = (QcFloat)fDual.elementAt(vi);
			int dualVarIx = fSubVBiMap.safeIndex(dualVar);
			if (dualVarIx >= 0)
				fDualIndex.fData[vi] = dualVarIx;
		}

		// The subsystem has both the dual and restricted primal constraints.
		loadDesVarCache();
		makeFeasibleComplex();
		fSubSystem.inEqRawSolveComplex();
		fSubSystem.eqRawSolveComplex();

		// And now the variable values provide the solution.
		fSystem.eqSolve();	// Solve using slack values as well
		fEditVarsSetup = true;
	}

	public void endEdit()
	{
		// The EndEdit in the QcSolver class is called so that the weight of
		// all the edit variable are set back to stay weight, and for clearEditVars.
		super.endEdit();

		for (int vi = 0; vi < fDepBasicVars.fCount; vi++)
			if (fDepBasicVars.fData[vi])
				fVBiMap.identifier(vi).restVariable();

		for (int viIt = 0; viIt < fDepPars.fCount; viIt++)
			fVBiMap.identifier(fDepPars.fData[viIt]).restVariable();

		fDual.removeAllElements();
		fDualIndex.removeAllElements();
	}

	//-----------------------------------------------------------------------//
	// Constraint management methods                                         //
	//-----------------------------------------------------------------------//
	public boolean addConstraint(QcConstraint c)
	{
		boolean success = fSystem.addConstraint(c);
		if (success) {
			if (fCheckPost) { //#ifdef qcCheckPost
				addCheckedConstraint(c);
			} //#endif
			if (fAutoSolve)
				solve();
		}
		return success;
	}

	public boolean addConstraint(QcConstraint c, QcFloat hint)
	{
		boolean success = fSystem.addConstraint(c, hint);
		if (success) {
			if (fCheckPost) { //#ifdef qcCheckPost
				addCheckedConstraint(c);
			} //#endif
			if (fAutoSolve)
				solve();
		}
		return success;
	}

	public void beginAddConstraint()
		{ fSystem.beginAddConstraint(); }

	public boolean changeConstraint(QcConstraint oldc, double rhs)
	{
		boolean success = fSystem.changeConstraint(oldc, rhs);
		if (fCheckPost) { //#ifdef qcCheckPost
			if (success)
				changeCheckedConstraint(oldc, rhs);
		} //#endif
		return success;
	}

	public boolean endAddConstraint()
		{ return fSystem.endAddConstraint(); }

	public boolean removeConstraint(QcConstraint c)
	{
		boolean success = fSystem.removeConstraint(c);
		if (success) {
			if (fCheckPost) { //#ifdef qcCheckPost
				removeCheckedConstraint(c);
			} //#endif
			if (fAutoSolve)
				solve();
		}
		return success;
	}

	public boolean reset()
		{ return fSystem.reset(); }

	//-----------------------------------------------------------------------//
	// Constraint Solving methods                                            //
	//-----------------------------------------------------------------------//
	public void solve()
		// IMPORTANT NOTE.
		// It is not necessary to call fSystem.solve() before creating and
		// solving the subsystem since the subsystem is created without using
		// the values of any variables - just desired values and weights.
	{
 		fAI = QcTableau.fInvalidVariableIndex;
		fSubSystem.restart();

		// Make a copy of the main tableau coefficients to speed up GenerateKT.
		// This is only performed if the virtual tableau or virtual rhs is used.
		cacheTableau();

		// Ensure variable indices are the same by adding variables into the
		// sub-tableau.  This may be improved later by just copying the bimaps.
		prepareVariables();

		// Place constraints with slack basic variables into fSubTableau.
		QcBasicVarIndexIterator bvIt = new QcBasicVarIndexIterator(fTableau);

		while (!bvIt.atEnd()) {
			if (fTableau.isSlack(bvIt.fIndex)) {
				int c = bvIt.fCurrent.fIsBasicIn;
	      		QcLinPoly lpoly = new QcLinPoly();
				lpoly.plus(new QcLinPolyTerm(
					fTableau.getValue(c, bvIt.fIndex),
					fVBiMap.identifier(bvIt.fIndex)));
				QcSparseMatrixRowIterator varCoeffs =
					new QcSparseMatrixRowIterator(fCacheCoeff, c);

				while (!varCoeffs.atEnd()) {
					lpoly.plus(new QcLinPolyTerm(varCoeffs.fValue,
								     fVBiMap.identifier(varCoeffs.fIndex)));
					varCoeffs.increment();
				}

				QcConstraint lc = new QcConstraint("", lpoly,
                 		QcConstraint.fOpEqual, fTableau.getRHS(c));
				fSubSystem.copyConstraint(lc, bvIt.fIndex);
			}
			bvIt.increment();
		}

		// This needs the original constraints copied into subsystem.
		// The dual variables are preallocated.  The indices are assigned below.
		createDualVars1();

		fAI = fSubTableau.addArtificial(-1.0f);
		fAV.setWeight(0.0f);	// Just in case it somehow get taken notice of
		fDual.addElement(fAV);	// This shouldn't need to be accessed.
		fDualIndex.addElement(fAI);
		fSubVBiMap.update(fAV, fAI);

		// Add Kuhn-Tucker constraints to the subsystem
		generateKT1();

		// Enter fDualIndex values for non slack variables
		int ssCols = fSubTableau.getColumns();

		for (int vi = 0; vi < ssCols; vi++) {
			QcFloat dualVar = (QcFloat)fDual.elementAt(vi);
			int dualVarIx = fSubVBiMap.safeIndex (dualVar);
			if (dualVarIx >= 0)
				fDualIndex.fData[vi] = dualVarIx;
		}

		// The subsystem has both the dual and restricted primal constraints.
		makeFeasible();
		fSubSystem.inEqRawSolve();
		fSubSystem.eqRawSolve();
		subSysEqSolve();		// Solve using slack values as well

		// Rest variables by setting their desire value to the actual value
		// computed above.
		QcVariableIndexIterator vIt = new QcVariableIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			QcFloat v = fVBiMap.identifier(vIt.fIndex);
			v.restVariable();
			vIt.increment();
		}

		// And now the variable values provide the solution.
		fDual.removeAllElements();
		fDualIndex.removeAllElements();

		if (fCheckPost) { //#ifdef qcCheckPost
			checkSatisfied();
		} //#endif
	}

	public void resolve()
	{
		if (!fEditVarsSetup)
			beginEdit();

		// The subsystem has both the dual and restricted primal constraints.
		loadDesVarCache();
		makeFeasibleComplex();

		fSubSystem.inEqRawSolveComplex();
		fSubSystem.eqRawSolveComplex();
		subSysEqSolve();	// Solve using slack values as well

		if (fCheckPost) { //#ifdef qcCheckPost
			checkSatisfied();
		} //#endif
	}

	/** Returns the current value of objective function.  Note that this
	    function is not currently used within QOCA, so is subject to
	    removal; let the QOCA maintainer know if you use this function and
	    want it to stay.
	**/
	public double objective()
	{
		double opt = 0.0f;
		QcVariableIndexIterator vIt = new QcVariableIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			QcFloat v = fVBiMap.identifier(vIt.fIndex);
			double delta = v.fValue - v.fDesireValue;
			opt += v.fWeight * delta * delta;
			vIt.increment();
		}

		return opt;
	}

	//-----------------------------------------------------------------------//
	// Sub-system specific functions.                                        //
	//-----------------------------------------------------------------------//
	private void addDualCoeffs1(QcLinPoly lpoly, int p)
		// Used by Generate constraints to add the dual variable
		// coefficients to a linpoly corresponding to parameter p.
		// Assumes the dual variables have already been allocated.
	{
		QcTableauColIterator varCoeffs = new QcTableauColIterator(fTableau, p);

		while (!varCoeffs.atEnd()) {
			if (fTableau.isRestricted(varCoeffs.fIndex)) {
				int cvi = fTableau.getBasicVar(varCoeffs.fIndex);
				QcFloat dualVar = (QcFloat)fDual.elementAt(cvi);

				// Register the new dual index
				if (!fSubVBiMap.identifierPresent(dualVar)) {
					int dualIndex = fSubTableau.addDual();
                         	// Should have linsimp method?
					fSubVBiMap.update(dualVar, dualIndex);
					QcFloat primeVar = fSubVBiMap.identifier(cvi);
					fDual.addElement(primeVar);
					fDualIndex.addElement(cvi);
					// This code is far from ideal for a number of reasons:
					// 1) It uses knowledge about how LinSimp is
					// implemented - specifically how the index for the new
					// dual is allocated on the fly.
					// 2) Looking up the handle for the corresponding
					// variable via the variable bimap seems a bit heavy
					// handed.
					//
					// This will be rethought when the interfaces are
					// revamped.
				}
				lpoly.plus(new QcLinPolyTerm(varCoeffs.fValue, dualVar));
			}
			varCoeffs.increment();
		}
	}

	private void addDualCoeffs2(QcLinPoly lpoly, int p)
		// Used by Generate constraints to add the dual variable
		// coefficients to a linpoly corresponding to parameter p.
		// Assumes the dual variables have already been allocated.
	{
		QcTableauColIterator varCoeffs = new QcTableauColIterator(fTableau, p);

		while (!varCoeffs.atEnd()) {
			if (fTableau.isRestricted(varCoeffs.fIndex)) {
				int cvi = fTableau.getBasicVar(varCoeffs.fIndex);
				QcFloat dualVar = (QcFloat)fDual.elementAt(cvi);

				// Register the new dual index
				if (!fSubVBiMap.identifierPresent(dualVar)) {
					int dualIndex = fSubTableau.addDual();
                         	// Should have linsimp method?
					fSubVBiMap.update(dualVar, dualIndex);
					QcFloat primeVar = fSubVBiMap.identifier(cvi);
					fDual.addElement(primeVar);
					fDualIndex.addElement(cvi);
					// This code is far from ideal for a number of reasons:
					// 1) It uses knowledge about how LinSimp is
					// implemented - specifically how the index for the new
					// dual is allocated on the fly.
					// 2) Looking up the handle for the corresponding
					// variable via the variable bimap seems a bit heavy
					// handed.
					//
					// This will be rethought when the interfaces are
					// revamped.
				}
				lpoly.plus(new QcLinPolyTerm(varCoeffs.fValue, dualVar));
			}
			varCoeffs.increment();
		}
	}

	private void cacheTableau()
	{
		fCacheCoeff.resize(0, 0);
		fCacheCoeff.resize(fTableau.getRows(), fTableau.getColumns());

		QcParamVarIndexIterator vIt = new QcParamVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			QcTableauColIterator colCoeffs =
            		new QcTableauColIterator(fTableau, vIt.fIndex);

			while (!colCoeffs.atEnd()) {
				fCacheCoeff.setValue(colCoeffs.fIndex, vIt.fIndex,
                 		colCoeffs.fValue);
				colCoeffs.increment();
			}
			vIt.increment();
		}
	}

	private void createDesValVars()
	{
		int cols = fTableau.getColumns();
		fDesValVar.setSize(cols);
		fDesValIndex.setSize(cols);

		for (int i = 0; i < cols; i++)
			fDesValIndex.fData[i] = QcTableau.fInvalidVariableIndex;

		for (int i = 0; i < fVarsByIndex.fCount; i++) {
			int vi = fVarsByIndex.fData[i];
			QcFloat v = new QcFloat("desval var",
						fSubVBiMap.identifier(vi).fDesireValue);
			v.setWeight(0.0f);
			v.suggestValue(v.fValue);
			fDesValIndex.fData[vi] = fSubTableau.addDesValVar();
			fDesValVar.setElementAt(v, vi);
			fSubVBiMap.update(v, fDesValIndex.fData[vi]);
			fDual.addElement(v);
			fDualIndex.addElement(fDesValIndex.fData[vi]);
		}
	}

	private void createDualVars1()
		// Create complementary variables for each variable in
		// the primal tableau.  The fDual array is used to locate these
		// variables.  N.B. Not all these variables need be used.
		// The rest of fDual is initialised in AddDualCoeffs.
	{
		int cols = fSubTableau.getColumns();

		for (int i = 0; i < cols; i++) {
			// Note that some of the following fDual names are changed later.
			fDual.addElement(new QcFloat("Dual Var"));
			fDualIndex.addElement(QcTableau.fInvalidVariableIndex);
		}
	}

	private void createDualVars2()
		// Create complementary variables for each variable in
		// the primal tableau.  The fDual array is used to locate these
		// variables.  N.B. Not all these variables need be used.
		// The rest of fDual is initialised in AddDualCoeffs.
	{
		int cols = fSubTableau.getColumns();

		for (int i = 0; i < cols; i++) {
			// Note that some of the following fDual names are changed later.
			fDual.addElement(new QcFloat("Dual Var"));
			fDualIndex.addElement(QcTableau.fInvalidVariableIndex);
		}
	}

	/** Generate the expression (lpoly,rhs): df'/dp + dual(p).  where f' =
	    least squares objective function with basic variables eliminated, and
	    dual(p) is the column coeffs of the original tableau for p applied
	    to new dual variables.  The variable coefficients are accumulated in
	    lpoly, the rhs value is returned.

	    @precondition <tt>p = fSubVBiMap.identifier(pIndex)</tt>
	    @precondition <tt>p</tt> not already in <tt>lpoly</tt>
	**/
	private double derivative1(QcFloat p, int pIndex, QcLinPoly lpoly)
	{
#if qcCheckInternalPre
		if (p != fSubVBiMap.identifier(pIndex))
			throw new InternalPreconditionException("p nequiv pIndex");
#endif
		double rhs = 0.0f;
		double pWeight2 = 2.0f * p.fWeight;
		double pAccumulator = pWeight2;

		QcBasicVarIndexIterator bvIt = new QcBasicVarIndexIterator(fTableau);

		while (!bvIt.atEnd()) {
			QcFloat bv = fSubVBiMap.identifier(bvIt.fIndex);
			double bvWeight = bv.fWeight;

			if (!QcUtility.isZero(bvWeight)) {
				int c = bvIt.fCurrent.fIsBasicIn;
				// bv is basic in the solved form constraint with index c
				double pCoeff = fCacheCoeff.getValue(c, pIndex);

				if (!QcUtility.isZero(pCoeff)) {
					double factor = 2.0f * bv.fWeight * pCoeff;
					pAccumulator += factor * pCoeff;

					// constraint c will be used to eliminate bv,
					// and pCoeff is the coefficient of p in the elimination.
	 				QcSparseMatrixRowIterator varCoeffs =
						new QcSparseMatrixRowIterator(fCacheCoeff, c);

					while (!varCoeffs.atEnd()) {
						if (varCoeffs.fIndex != pIndex) {
							QcFloat pv = fSubVBiMap.identifier(
								varCoeffs.fIndex);
							lpoly.plus(new QcLinPolyTerm(
								factor * varCoeffs.fValue, pv));
						}
						varCoeffs.increment();
					}
					rhs += factor * (fTableau.getRHS(c) - bv.fDesireValue);
				}
			}
			bvIt.increment();
		}

		lpoly.addTerm(new QcLinPolyTerm(pAccumulator, p));
		addDualCoeffs1(lpoly, pIndex);
		rhs += pWeight2 * p.fDesireValue;
		return rhs;
	}

	/** Generate the expression (lpoly,rhs): df'/dp + dual(p).  where f' =
	    least squares objective function with basic variables eliminated and
	    dual(p) is the column coeffs of the original tableau for p applied
	    to new dual variables.  The variable coefficients are accumulated in
	    lpoly, the rhs value is returned.

	    @precondition <tt>p = fSubVBiMap.identifier(pIndex)</tt>
	    @precondition <tt>p</tt> not already in <tt>lpoly</tt>
	**/
	private double derivative2(QcFloat p, int pIndex, QcLinPoly lpoly)
	{
#if qcCheckInternalPre
		if (p != fSubVBiMap.identifier(pIndex))
			throw new InternalPreconditionException("p nequiv pIndex");
#endif
		double rhs = 0.0f;
		double pWeight2 = 2.0f * p.fWeight;
		double pAccumulator = pWeight2;

		QcBasicVarIndexIterator bvIt = new QcBasicVarIndexIterator(fTableau);

		while (!bvIt.atEnd()) {
			QcFloat bv = fSubVBiMap.identifier(bvIt.fIndex);
			double bvWeight = bv.fWeight;

			if (!QcUtility.isZero(bvWeight)) {
				int c = bvIt.fCurrent.fIsBasicIn;
				double pCoeff = fCacheCoeff.getValue(c, pIndex);

				if (!QcUtility.isZero(pCoeff)) {
					// bv is basic in the solved form constraint with index c
					double factor = 2.0f * bv.fWeight * pCoeff;
					pAccumulator += factor * pCoeff;

					// constraint c will be used to eliminate bv,
					// and pCoeff is the coefficient of p in the elimination.
	 				QcSparseMatrixRowIterator varCoeffs =
						new QcSparseMatrixRowIterator(fCacheCoeff, c);

					while (!varCoeffs.atEnd()) {
						if (varCoeffs.fIndex != pIndex) {
							QcFloat pv = fSubVBiMap.identifier(
								varCoeffs.fIndex);
							lpoly.plus(new QcLinPolyTerm(
								factor * varCoeffs.fValue, pv));
						}
						varCoeffs.increment();
					}

					if (isDepBasicVar(bvIt.fIndex) && isEditVar(bv)) {
						rhs += factor * fTableau.getRHS(c);
						lpoly.plus(new QcLinPolyTerm(factor,
							(QcFloat)fDesValVar.elementAt(bvIt.fIndex)));
					} else
			         		rhs += factor * (fTableau.getRHS(c) - bv.fDesireValue);
				}
			}
			bvIt.increment();
		}

		lpoly.addTerm(new QcLinPolyTerm(pAccumulator, p));
		addDualCoeffs2(lpoly, pIndex);

		if (isEditVar(p)) {
			QcFloat pDV = (QcFloat)fDesValVar.elementAt(pIndex);
			lpoly.addTerm(new QcLinPolyTerm(-pWeight2, pDV));
		} else
			rhs += pWeight2 * p.fDesireValue;

		return rhs;
	}

	private void generateKT1()
		// Generate the constraint df'/dp + dual(p) == 0 for each unconstrained
		// primal parameter p.  Then generate the constraint
		// df'/dp + dual(p) >= 0 for each constrained (i.e. slack) primal
		// parameter.  Add the new constraint directly to the subsystem.
		// Where f' = least squares objective function with basic variables
		// eliminated, and dual(p) is the column coeffs of the original
		// tableau for p applied to new dual variables.
	{
		QcParamVarIndexIterator viIt = new QcParamVarIndexIterator(fTableau);

		while (!viIt.atEnd()) {
			int vi = viIt.fIndex;

			// The !IsFree test avoids entering a redundant constraint in
			// the subsystem (to save time).  There is no need to call
			// v.set when the variable is free since this will be
			// done when fSystem.solve() is called later.
			if (!fTableau.isFree(vi) && fTableau.isConstrained(vi)) {
				// Primal slack parameters were given a weight of 0 in Solve.
				QcLinPoly lpoly = new QcLinPoly();
 				QcFloat p = fSubVBiMap.identifier(vi);
				double rhs = derivative1(p, vi, lpoly);
				lpoly.plus(new QcLinPolyTerm(1.0f, fAV));
				// artificial variable coeff
				QcConstraint lc = new QcConstraint("", lpoly,
								   QcConstraint.fOpGreaterEqual, rhs);
				fSubSystem.copyInEq(lc);
				fDual.addElement(p);	// p corresponds to new slack
				fDualIndex.addElement(vi);
				fDualIndex.setElementAt(fSubTableau.getColumns() - 1, vi);
			}
			viIt.increment();
		}

		fSubTableau.setBeginEq();
		viIt.reset();

		while (!viIt.atEnd()) {
			int vi = viIt.fIndex;

			// The !IsFree test avoids entering a redundant constraint in
			// the subsystem (to save time).  There is no need to call
			// v.set when the variable is free since this will be
			// done when fSystem.Solve() is called later.
			if (!fTableau.isFree(vi) && !fTableau.isConstrained(vi)) {
				// Do the unrestricted primal parameters.
				QcLinPoly lpoly = new QcLinPoly();
 				QcFloat viVar = fSubVBiMap.identifier(vi);
				double rhs = derivative1(viVar, vi, lpoly);
				// No need to add the (-1) artificial variable coeff
				// here since the primal parameter is unrestricted.
				QcConstraint lc = new QcConstraint("", lpoly,
								   QcConstraint.fOpEqual, rhs);
				boolean consistent = fSubSystem.addConstraint(lc, viVar);
				qcAssertPost(consistent);
			}
			viIt.increment();
		}
	}

	private void generateKT2()
		// Generate the constraint df'/dp + dual(p) == 0 for each unconstrained
		// primal parameter p.  Then generate the constraint
		// df'/dp + dual(p) >= 0 for each constrained (i.e. slack) primal
		// parameter.  Add the new constraint directly to the subsystem.
		// Where f' = least squares objective function with basic variables
		// eliminated, and dual(p) is the column coeffs of the original
		// tableau for p applied to new dual variables.
	{
		for (int viIt = 0; viIt < fDepPars.fCount; viIt++) {
			int vi = fDepPars.fData[viIt];

			if (!fTableau.isFree(vi) && fTableau.isConstrained(vi)) {
				// Primal slack parameters were given a weight of 0 in Solve.
 				QcLinPoly lpoly = new QcLinPoly();
 				QcFloat viVar = fSubVBiMap.identifier(vi);
 				double rhs = derivative2(viVar, vi, lpoly);
				lpoly.plus(new QcLinPolyTerm(1.0f, fAV));
				// artificial variable coeff
 				QcConstraint lc = new QcConstraint("", lpoly,
                 		QcConstraint.fOpGreaterEqual, rhs);
 				fSubSystem.copyInEq(lc);	// new slack is made basic
 				fDual.addElement(viVar);	// viVar corresponds to new slack
 				fDualIndex.addElement(vi);
 				fDualIndex.fData[vi] = fSubTableau.getColumns() - 1;
			}
		}

		fSubTableau.setBeginEq();

		for (int viIt = 0; viIt < fDepPars.fCount; viIt++) {
			int vi = fDepPars.fData[viIt];

			if (!fTableau.isFree(vi) && !fTableau.isConstrained(vi)) {
				// First do the unrestricted primal parameters.
				QcLinPoly lpoly = new QcLinPoly();
				QcFloat viVar = fSubVBiMap.identifier(vi);
				double rhs = derivative2(viVar, vi, lpoly);
				// No need to add the (-1) artificial variable coeff
				// here since the primal parameter is unrestricted.
				QcConstraint lc = new QcConstraint("", lpoly,
								   QcConstraint.fOpEqual, rhs);
				fSubSystem.addConstraint(lc, viVar);
			}
		}
	}

	public void loadDesVarCache()
		// Required for desval variables.
	{
		int dvSize = fDesValVar.size();

		for (int i = 0; i < dvSize; i++) {
			if (fDesValIndex.fData[i] != QcTableau.fInvalidVariableIndex) {
		    		fSubTableau.setDesireValue(fDesValIndex.fData[i],
                 		fSubVBiMap.identifier(i).fDesireValue);
			}
		}
	}

	private void makeFeasible()
		// Make the subsystem tableau feasible
	{
		// find the restricted row with most negative RHS
		double mostNeg = 0.0f;
		int mostNegIndex = QcTableau.fInvalidConstraintIndex;
		int rows = fSubTableau.getRows();

		for (int ci = 0; ci < rows; ci++) {
			if (ci >= fSubTableau.getBeginEq())
				break;

			if (fSubTableau.isRestricted(ci)) {
				double coeff = fSubTableau.getRHS(ci);

				if (coeff < mostNeg) {
					mostNeg = coeff;
					mostNegIndex = ci;
				}
			}
		}

		if (!QcUtility.isNegative(mostNeg))
			return;	// tableau is already feasible

		// Initialize the coefficients of fAV in all the constraints with 1.0.
		fSubTableau.setArtificial(-1.0f);

		int justLeft = fSubTableau.getBasicVar(mostNegIndex);
		fSubTableau.restrictedPivot(mostNegIndex, fAI);

		// We now have a subsystem tableau which is basic feasible.
		// But we wish to eliminate the artificial variable from it.
		// To do so we keep pivoting on the variable complementary to the
		// variable which just left the basis.
		int ci;

		while (fSubTableau.isBasic(fAI)) {	// artificial variable is still basic
			int vi = fDualIndex.fData[justLeft];  // complementary variable
			// Select Pivot row
			ci = fSubSystem.selectExitVar(vi);

			if (ci < 0)
				break;	// Better to give up than continue now.

			justLeft = fSubTableau.getBasicVar(ci);
			fSubTableau.restrictedPivot(ci, vi);
		}
	}

	private void makeFeasibleComplex()
		// Make the subsystem tableau feasible
	{
		// find the restricted row with most negative RHS
		double mostNeg = 0.0f;
		int mostNegIndex = QcTableau.fInvalidConstraintIndex;
		int rows = fSubTableau.getRows();

		for (int ci = 0; ci < rows; ci++) {
			if (ci >= fSubTableau.getBeginEq())
				break;

			if (fSubTableau.isRestricted(ci)) {
				double coeff = fSubTableau.getComplexRHS(ci);
				if (coeff < mostNeg) {
					mostNeg = coeff;
					mostNegIndex = ci;
				}
			}
		}

		if (!QcUtility.isNegative(mostNeg))
			return;	// tableau is already feasible

		// Initialize the coefficients of fAV in all the constraints with 1.0.
		fSubTableau.setArtificial(-1.0f);

		int justLeft = fSubTableau.getBasicVar(mostNegIndex);
		fSubTableau.restrictedPivot(mostNegIndex, fAI);

		// We now have a subsystem tableau which is basic feasible.
		// But we wish to eliminate the artificial variable from it.
		// To do so we keep pivoting on the variable complementary to the
		// variable which just left the basis.
		int ci;

		while (fSubTableau.isBasic(fAI)) {	// artificial variable is still basic
			int vi = fDualIndex.fData[justLeft];  // complementary variable
			// Select Pivot row
			ci = fSubSystem.selectExitVarComplex(vi);

			if (ci < 0)
				break;	// Better to give up than continue now.

			justLeft = fSubTableau.getBasicVar(ci);
			fSubTableau.restrictedPivot(ci, vi);
		}
	}

	private void prepareVariables()
	{
		int cols = fTableau.getColumns();

		for (int vi = 0; vi < cols; vi++) {
			QcFloat v = fVBiMap.identifier(vi);

			if (fTableau.isSlack(vi)) {
				v.suggestValue(0.0f);
				fSubSystem.addSlack(v);
			} else
				fSubSystem.addVar(v);
		}
	}

	private void subSysEqSolve()
	{
		QcVariableIndexIterator vIt = new QcVariableIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			if (fTableau.isBasic(vIt.fIndex)) {
				QcFloat v = fVBiMap.identifier(vIt.fIndex);
				int ci = fTableau.isBasicIn(vIt.fIndex);
				double value = fTableau.getRHS(ci);
				QcTableauRowIterator varCoeffs =
					new QcTableauRowIterator(fTableau, ci);

				while (!varCoeffs.atEnd()) {
					if (varCoeffs.fIndex != vIt.fIndex) {
						QcFloat bv = fVBiMap.identifier(varCoeffs.fIndex);
						value -= varCoeffs.fValue * bv.fValue;
					}
					varCoeffs.increment();
				}

				v.setValue(QcUtility.zeroise(value));
			}
			vIt.increment();
		}
	}

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//
	public QcConstraint getConstraint(String n)
		{ return fSystem.getConstraint(n); }

	public QcFloat getVariable(String n)
		{ return fSystem.getVariable(n); }

	public void getVariableSet(QcVariableSet vars)
		{ fSystem.getVariableSet(vars); }

	public void loadFromBinary(DataInputStream s, QcVariableSet vars)
	{
		super.loadFromBinary(s, vars);
		fSystem.loadFromBinary(s, vars);

		fVBiMap = fSystem.fVBiMap;
		fNotifier = fSystem.fNotifier;
		fOCBiMap = fSystem.fOCBiMap;
 		fTableau = (QcDelLinInEqTableau)fSystem.fTableau;
	}

	public void loadFromString(StringTokenizer token, QcVariableSet vars)
	{
		super.loadFromString(token, vars);
		fSystem.loadFromString(token, vars);

		fVBiMap = fSystem.fVBiMap;
		fNotifier = fSystem.fNotifier;
		fOCBiMap = fSystem.fOCBiMap;
 		fTableau = (QcDelLinInEqTableau)fSystem.fTableau;
	}

	public void print()
	{
		fSystem.print();

 		if (fEditVarsSetup) {
			System.out.println("Complementary pivot tableau: ");
			fSubTableau.print();
			System.out.println();
			System.out.println("Sub-system notifier: ");
			fSubNotifier.print();
			System.out.println();
		}
	}

	public void restart()
	{
		fSystem.restart();
		fSubSystem.restart();
	}

	public void saveAsBinary(DataOutputStream s, QcVariableSet vars)
	{
		super.saveAsBinary(s, vars);
		fSystem.saveAsBinary(s, vars);
	}

	public void saveAsText(OutputStream s, QcVariableSet vars)
	{
		super.saveAsText(s, vars);
		fSystem.saveAsText(s, vars);
	}

	public void saveToString(StringBuffer s, QcVariableSet vars)
	{
		super.saveToString(s, vars);
		fSystem.saveToString(s, vars);
	}

	/** Calculate fDepBasicVars and fDepPars as transitive closure of
	    variables in constraints which mention EditVars. */
	public void transitiveClosure()
	{
		QcIntVector fDepVars = fVarsByIndex.duplicate(QcDupable.DEEP);
		fTableau.transitiveClosure(fDepVars);

		// Sift fDepVars into fDepPars, fDepBasicVars and fDepParVars.
		int cols = fTableau.getColumns();
		fDepBasicVars.setSize(cols);
		fDepParVars.setSize(cols);

		for (int i = 0; i < cols; i++) {
			fDepBasicVars.fData[i] = false;
			fDepParVars.fData[i] = false;
		}

		fDepPars.removeAllElements();

		for (int i = 0; i < fDepVars.size(); i++) {
			int vi = fDepVars.fData[i];

			if (fTableau.isBasic(vi))
				fDepBasicVars.fData[vi] = true;
			else {
				fDepPars.addElement(vi);
				fDepParVars.fData[vi] = true;
			}
		}
	}
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
