package qoca;
// $Id: QcLinInEqSystem.java.cpp,v 1.6 2000/11/21 05:35:16 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//----------------------------------------------------------------------------//
// This interface needs to be read in conjuction with that of Solver.         //
//============================================================================//

import java.util.*;
import java.io.*;

#include "QcDefines.jh"

class QcLinInEqSystem extends QcDelLinEqSystem
{
	QcLinInEqTableau fTableau;
	private QcIntVector fArtVars;

	//-----------------------------------------------------------------------//
	// Constructor.                                                          //
	//-----------------------------------------------------------------------//
	public QcLinInEqSystem()
	{
		super(0, 0, new QcLinInEqTableau(0, 0, new QcBiMapNotifier()));
		fArtVars = new QcIntVector();
		fTableau = (QcLinInEqTableau)super.fTableau;
	}

	public QcLinInEqSystem(int hintRows, int hintCols)
	{
		super(hintRows, hintCols,
			new QcLinInEqTableau(hintRows, hintCols,
			new QcBiMapNotifier()));
		fArtVars = new QcIntVector();
		fTableau = (QcLinInEqTableau)super.fTableau;
	}

	public QcLinInEqSystem(int hintRows, int hintCols, QcLinInEqTableau tab)
	{
		super(hintRows, hintCols, tab);
		fArtVars = new QcIntVector();
		fTableau = tab;
	}

	public QcLinInEqSystem(DataInputStream s, QcVariableSet vars)
		{ loadFromBinary(s, vars); }

	public QcLinInEqSystem(StringTokenizer token, QcVariableSet vars)
		{ loadFromString(token, vars); }

	//-----------------------------------------------------------------------//
	// Variable management methods (see QcSolver.java for descriptions)      //
	//-----------------------------------------------------------------------//
	public boolean removeVar(QcFloat v)
	{
		System.out.println("Error: Should not call QcLinInEqSystem::removeVar");
		return false;
	}

	public boolean removeVarSuper(QcFloat v)
		{ return super.removeVar(v); }

	public void restSolver()
	{
		QcStructVarIndexIterator vIt =
			new QcStructVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			fVBiMap.identifier(vIt.fIndex).restVariable();
			vIt.increment();
		}
	}

	//-----------------------------------------------------------------------//
	// High Level Edit Variable Interface for use by solver clients.         //
	//-----------------------------------------------------------------------//
	public void beginEdit()
		// This is the same as LinEqSystem::BeginEdit except where noted with
		// the comment ++.
	{
		// build a version of fEditVars by index.
		fVarsByIndex.removeAllElements();
		int evSize = fEditVars.size();

		for (int voiIt = 0; voiIt < evSize; voiIt++) {
			QcFloat f = (QcFloat)fEditVars.elementAt(voiIt);
			if (isRegistered(f))
				fVarsByIndex.addElement(fVBiMap.index(f));
		}

		// try to make all variables of interest into parameters.
		for (int viIt = 0; viIt < fVarsByIndex.fCount; viIt++) {
			int vi = fVarsByIndex.fData[viIt];
			if (fTableau.isBasic(vi)) {
				int ci = fTableau.isBasicIn(vi);
				// Cache tableau column coefficients since it is possible that
				// the iterator may be invalidated during iteration.
				QcTableauRowIterator varCoeffs =
					new QcTableauRowIterator(fTableau, ci);
				fCoeffCache.removeAllElements();

				while (!varCoeffs.atEnd()) {
					fCoeffCache.addElement(
            				new QcSparseCoeff(varCoeffs.fValue,
						varCoeffs.fIndex));
					varCoeffs.increment();
				}

				Enumeration cIt = fCoeffCache.elements();

				while (cIt.hasMoreElements()) {
					QcSparseCoeff coeff = (QcSparseCoeff)cIt.nextElement();
					boolean reject = fTableau.isConstrained(coeff.fIndex);
						// ++

					for (int viIt2 = 0; viIt2 < fVarsByIndex.fCount; viIt2++)
						if (coeff.fIndex == fVarsByIndex.fData[viIt2])
							reject = true;

					if (!reject) {
						fTableau.pivot(ci, coeff.fIndex);
						break;
					}
				}
			}
		}

		solve();							// Compute new solution here.
		fParsOfInterest.removeAllElements(); 	// setup fParsOfInterest

		for (int viIt = 0; viIt < fVarsByIndex.fCount; viIt++) {
			int vi = fVarsByIndex.fData[viIt];
			if (!fTableau.isBasic(vi) && !isParOfInterest(vi))
				fParsOfInterest.addElement(vi);
		}

		fDepVars.removeAllElements(); 		// setup fDepVars

		// setup fDepVarCoeffs.  Extend fDepVarCoeffs so 0..fTableau.Cols()-1 are
		// legal indices.
		fDepVarCoeffs = new Vector[fTableau.getColumns()];
		fDepVarBaseVals.setSize(fTableau.getColumns());
		QcBasicVarIndexIterator bvIt = new QcBasicVarIndexIterator(fTableau);

		for (int viIt = 0; viIt < fDepVarCoeffs.length; viIt++)
			fDepVarCoeffs[viIt] = new Vector(
				QcConstants.fVectorInitCapacity,
				QcConstants.fVectorCapacityIncrement);

		while (!bvIt.atEnd()) {
			int bvi = bvIt.fIndex;
			int ci = bvIt.fCurrent.fIsBasicIn;
			fDepVarBaseVals.fData[bvi] = fTableau.getRHS(ci);
			QcTableauRowIterator varCoeffs =
				new QcTableauRowIterator(fTableau, ci);

			while (!varCoeffs.atEnd()) {
				if (isParOfInterest(varCoeffs.fIndex)) {
					if (!isDepVar(bvi))
						fDepVars.addElement(bvi);
						fDepVarCoeffs[bvi].addElement(
							new QcSparseCoeff(varCoeffs.fValue,
                                	varCoeffs.fIndex));
				} else if (bvi != varCoeffs.fIndex) {
					if (fTableau.isStructural(varCoeffs.fIndex)) { // ++
						double val = fVBiMap.identifier(varCoeffs.fIndex).fValue;
						fDepVarBaseVals.fData[bvi] -= val * varCoeffs.fValue;
					}
				}
				varCoeffs.increment();
			}
			bvIt.increment();
		}

		fEditVarsSetup = true;
	}

	//-----------------------------------------------------------------------//
	// Constraint management methods                                         //
	//-----------------------------------------------------------------------//
	public boolean addConstraint(QcConstraint c)
		// Extracts the various components and enters them in the tableau
		// as well as updating the bimapnotifier for the tableau
		// if successful.  The return value indicates success or failure.
		// If AddConstraint fails the state of the tableau is unchanged.
	{
		return doAddConstraint(c, -1);
	}

	public boolean addConstraint(QcConstraint c, QcFloat hint)
	{
#if qcCheckPre
		qcAssertExternalPre(hint != null);
		int hintIx = fVBiMap.safeIndex(hint);
		qcAssertExternalPre(hintIx >= 0);
#else
		int hintIx = fVBiMap.index(hint);
#endif
		return doAddConstraint(c, hintIx);
	}

	private boolean doAddConstraint(QcConstraint c, int hintIx)
	{
		qcAssertExternalPre(c != null);

		// If in batch mode and the one of more of the preceeding add constraint
		// operation failed, then should not proceed to do anymore constraint
		// addition.
		if (fBatchAddConst && fBatchAddConstFail)
			return false;

		boolean success = true;
		int co = addToTableau(c);
		// co is the original constraint index

		int cs = fTableau.eliminate(co);
		// cs is solved form constraint index

		if (hintIx >= 0) {
			fTableau.fPivotHints.removeAllElements();
			fTableau.fPivotHints.addElement(hintIx);
		}

		if (cs == QcTableau.fInvalidConstraintIndex) {
			fInconsistant.addElement(c);
			fEditVarsSetup = false; // effic: This may be unnecessary.
			return false;
		}

		if (fBatchAddConst) {
			// Check if it has unrestricted variables.
			QcTableauRowIterator varCoeffs = new QcTableauRowIterator(fTableau, cs);
			while (!varCoeffs.atEnd()) {
				if (!fTableau.isSlack(varCoeffs.fIndex))
					break;
				varCoeffs.increment();
			}

			// If there is, then do the normal add constraint.
			if (!varCoeffs.atEnd()) {
				/* TODO: It's not clear that the "hintIx < 0 || " part
				   was deliberate. */
				if (hintIx < 0 || !fTableau.isRedundant(cs)) {
					if (fTableau.isRedundant(cs)) // pjm fixme
						System.err.println("see line " + __LINE__ + " of " + __FILE__);
					success = applyHints(cs);
					if (!success)
						fTableau.deleteRow(co);
				}
			} else {
				// If the constraint has a new slack variable.
				if (c.fOp != QcConstraint.fOpEqual) {
#if qcCheckInternalPre
					double slackCoeff = fTableau.getValue(cs, fTableau.getColumns() - 1);
					// slackCoeff is 1.0 for &le;, -1.0 for &ge;.
					qcAssertPre(   (c.fOp == QcConstraint.LE && slackCoeff == 1.0)
							       || (c.fOp == QcConstraint.GE && slackCoeff == -1.0));
#endif
					double rhs = fTableau.getRHS(cs);

					if ((c.fOp == QcConstraint.GE) // negative slack coefficient
					    ? !QcUtility.isPositive(rhs)
					    : !QcUtility.isNegative(rhs)) {
						fTableau.pivot(cs, fTableau.getColumns() - 1);
						// Should consider just setting the state of the
						// row and column instead of calling Pivot.  This
						// is because we know the variable to be make
						// basic is not in any other rows.
					} else {
						System.err.println("here!");
						qcAssertPre(!QcUtility.isZero(rhs));
						int vi = fTableau.addArtificial((rhs > 0.0)
										? 1.0
										: -1.0);
						fArtVars.addElement(vi);
						fTableau.pivot(cs, vi);
						// May scale row by -1 to make the newly added
						// artifical variable to +1.  This method may
						// require explicit management of column and row
						// state.
					}
				}
			}

			if (success) {
				fOCBiMap.update(c, co);
				fBatchConstraints.addElement(c);
			} else
				fBatchAddConstFail = true;
		} else {
			if (!fTableau.isRedundant(cs)) {
				success = applyHints(cs);
				if (!success)
					fTableau.deleteRow(co);
			}

			if (success)
				fOCBiMap.update(c, co);
		}

		fEditVarsSetup = false;
		return success;
	}

	public boolean addInEq(QcConstraint c)
		// Used for adding inequalities only.  Always pivots on the new slack
		// variable (hence making it basic).
	{
		boolean success = true;
		int co = addToTableau(c);

		if (co < 0)
			return false;

		int cs = fTableau.eliminate(co);	 // cs is solved form constraint index

		if (cs == QcTableau.fInvalidConstraintIndex) {
			fInconsistant.addElement(c);
			success = false;
		} else if (!fTableau.isRedundant(cs)) {	// Pivot on new slack variable
			success = fTableau.pivot(cs, fTableau.getColumns() - 1);

			// Should get the slack var index returned from AddToTableau
			// rather than just knowing how it is implemented.
			if (!success)
				fTableau.deleteRow(co);
		}

		if (success)
			fOCBiMap.update(c, co);

		fEditVarsSetup = false;
		return success;
	}

	/** Does first part of addConstraint, calling the correct tableau
	    function.

	    @return <tt>fTableau.add*</tt>
	    @precondition <tt>c</tt> is non-null, and has a valid relation (eq, le, ge).
	    @postcondition ret &ge; 0
	**/
	protected int addToTableau(QcConstraint c)
	{
		QcRowAdaptor ra = new QcRowAdaptor(c.fLinPoly, fVBiMap, fTableau);
		int co;

		switch (c.fOp) {
			case QcConstraint.fOpEqual:
				co = fTableau.addEq(ra, c.fRHS);
				break;

			case QcConstraint.fOpLesserEqual:
				co = fTableau.addLtEq(ra, c.fRHS);
				fVBiMap.update(
					new QcFloat("", 0.0f, 0.0f, 0.0f, true),
					fTableau.getColumns() - 1);
				break;

			case QcConstraint.fOpGreaterEqual:
				co = fTableau.addGtEq(ra, c.fRHS);
				fVBiMap.update(
					new QcFloat("", 0.0f, 0.0f, 0.0f, true),
					fTableau.getColumns() - 1);
				break;

			default:
				throw new InternalPreconditionException("constraint has invalid relation " + c.fOp);
		}
		qcAssertPost(co >= 0);
		return co;
	}

	public boolean applyHints(int cs)
		// This version of applyHints is similar to QcLinEqSystem::applyHints
		// only it is aware of slack variables and calls SimplexI if none of
		// the slack variables can be made basic with a single pivot.
		// This is a convenience function to pivot an independent new
		// solved form constraint according to the PivotHints.
	{
		boolean pivoted = false;

		for (int i = 0; i < fTableau.fPivotHints.fCount && !pivoted; i++) {
			int hint = fTableau.fPivotHints.fData[i];
			if (!fTableau.isConstrained(hint))
				pivoted = fTableau.pivot(cs, hint);
		}

		Vector slackCoeffs = new Vector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);

		// If PivotHints failed, this means all the pivot hints variables
		// have zero coeffs or are constrained (slack).
		// Now try all the rest of the parametric variables with non-zero
		// coeffs in the solved form constraint.  First try non-slack
		// variables, saving the indices of any slack parameters for later use.
		if (!pivoted) {
			// Cache tableau column coefficients since it is possible that
			// the iterator may be invalidated during iteration.
			QcTableauRowIterator varCoeffs =
				new QcTableauRowIterator(fTableau, cs);
			fCoeffCache.removeAllElements();

			while (!varCoeffs.atEnd()) {
				fCoeffCache.addElement(
            			new QcSparseCoeff(varCoeffs.fValue, varCoeffs.fIndex));
				varCoeffs.increment();
			}

			Enumeration cIt = fCoeffCache.elements();

			while (cIt.hasMoreElements() && !pivoted) {
				QcSparseCoeff coeff = (QcSparseCoeff)cIt.nextElement();

				if (!fTableau.isBasic(coeff.fIndex)) {
					if (fTableau.isConstrained(coeff.fIndex))
						slackCoeffs.addElement(
							new QcSparseCoeff(coeff.fValue,
                                	coeff.fIndex));
					else
						pivoted = fTableau.pivot(cs, coeff.fIndex);
				}
			}
		}

		if (!pivoted) {	// try the new slack variable
			double rhs = fTableau.getRHS(cs);
			QcSparseCoeff s = (QcSparseCoeff)slackCoeffs.elementAt(
            		slackCoeffs.size() - 1);
			slackCoeffs.setSize(slackCoeffs.size() - 1);

			if (s.fIndex == fTableau.getColumns() - 1) {	// new slack belongs to cs
				if (QcUtility.isNegative(s.fValue) ?
						!QcUtility.isPositive(rhs) :
						!QcUtility.isNegative(rhs)) {
					pivoted = fTableau.pivot(cs, s.fIndex);
				}
			}

			// Call SimplexI here as have rhs and new slack is popped off.
			if (!pivoted)
				pivoted = simplexI(cs);
		}

		return pivoted;
	}

	/* FIXME: These should be spelled beginAddConstraint and
	   endAddConstraint.  Why is it still working when these B* / E*
	   routines are never being called?  TODO: Add appropriate test case. */

	public void XbeginAddConstraint()
		// Marks the beginning of a batch of AddConstraint operation.
	{
		super.beginAddConstraint();
		fArtVars.removeAllElements();
	}
	
	public boolean XendAddConstraint()
		// Marks the end of a batch of AddConstraint operation.
	{
		super.endAddConstraint();

		if (fBatchAddConstFail)
			return false;

		if (fArtVars.size() == 0)
			return true;
			
		// Set objective function based on the added artifical
		// variables, and perform Simplex phase II.
		int cols = fTableau.getColumns();

		// First, initialize the objective function.
		for (int c = 0; c < cols; c++)
			fTableau.setObjValue(c, 0.0);

			// Then, set the objective function base on the artifical
			// variables.
		for (int n = fArtVars.size(), i = 0; i < n; i++) {
			int a = fArtVars.elementAt(i);
			QcFloat v = fVBiMap.identifier(a);
			fTableau.setObjValue(a, v.fWeight);
		}

		fTableau.eliminateObjective();
		simplexII();

		// Remove the added artifical variables from the tableau.
		for (int n = fArtVars.size(), i = 0; i < n; i++) {
			int vi = fArtVars.elementAt(i);

			if (!fTableau.isBasic(vi))
				fTableau.removeVar(vi);
			else {
				int maxcol = -1;
				double max = 0.0;
				int r = fTableau.isBasicIn(vi);
				QcTableauRowIterator iR =
					new QcTableauRowIterator(fTableau, r);

				while (!iR.atEnd()) {
					if (max < Math.abs(iR.fValue) && iR.fIndex != vi) {
						max = iR.fValue;
						maxcol = iR.fIndex;
					}

					iR.increment();
				}

				// If there is at least one non-zero coefficient in the
				// current column.
				if (maxcol != -1) {
					fTableau.pivot(r, maxcol);
					fTableau.removeVar(vi);
				} else {
					if (QcUtility.isZero(fTableau.getRHS(r))) {
						fTableau.setRowCondition(r, QcLinEqRowState.fRedundant);
						fTableau.removeVar(vi);
					} else {
						fTableau.setRowCondition(r, QcLinEqRowState.fInvalid);
					}
				}
			}
		}
		return true;
	}

	public boolean reset()
		// While this routine is in progress the tableau is slowly
		// rebuilt from row 0 forwards retaining the original constraint
		// matrix fA.
		// The implementation used here only relies on the rest
		// of the tableau methods being able to cope with multiple
		// undetermined constraints.
		// Reset starts by calling Restore to establish a state equivalent
		// to the original constraints having been all entered using AddEq
		// and nothing else apart from the deleted entries being left alone.
	{
		boolean result = true;
		int cols = fTableau.getColumns();
		int rows = fTableau.getRows();

		// Record the columns of the previous basic variables so to make them
		// basic after this Reset() method.
		boolean rbasic[] = new boolean[cols];
		boolean urbasic[] = new boolean[cols];

		for (int c = 0; c < cols; c++) {
			if (fTableau.isBasic(c)) {
				if (fTableau.isSlack(c)) {
					rbasic[c] = true;
					urbasic[c] = false;
				} else {
					rbasic[c] = false;
					urbasic[c] = true;
				}
			} else {
				rbasic[c] = false;
				urbasic[c] = false;
			}
		}

		fEditVarsSetup = false;
		fTableau.restore(); // Reset variable state information

		// Establish solved form and quasi-inverse from original constraints.
		// Step 1: Pivot using unrestricted variables.
		for (int c = 0; c < cols; c++) {
			if (urbasic[c]) {
	     		int maxrow = -1;
     			double max = 0.0;
     			QcTableauColIterator iC =
                    		new QcTableauColIterator(fTableau, c);

				while (!iC.atEnd()) {
					if (max < Math.abs(iC.fValue)) {
          	     		max = iC.fValue;
               	     	maxrow = iC.fIndex;
	               	}

          			iC.increment();
          		}

				// If there is at least one non-zero coefficient in the
				// current column.
				if (maxrow != -1)
					fTableau.pivot(maxrow, c);
			}
		}

		// Step 2: Make sure that all solved form constraints with
		// unrestricted variables are assigned a basic variable.
		for (int r = 0; r < rows; r++) {
			if (fTableau.getRowCondition(r) != QcLinEqRowState.fRegular) {
				int maxcol = -1;
				double max = 0.0;
				int count = 0;
				QcTableauRowIterator iR =
                    		new QcTableauRowIterator(fTableau, r);

				while (!iR.atEnd()) {
					if (!fTableau.isSlack(iR.fIndex)) {
						if (max < Math.abs(iR.fValue)) {
     	     	     		max = iR.fValue;
          	     	     	maxcol = iR.fIndex;
	          	     	}
                    	}

                    	count++;
          			iR.increment();
          		}

				// If there is at least one non-zero coefficient in the
				// current row.
				if (maxcol != -1)
					fTableau.pivot(r, maxcol);
				else if (count == 0) {
					// Verify that the Reset operation is successful.  It
					// fails if one of more of the solved form row is
					// marked as "invalid".
					if (QcUtility.isZero(fTableau.getRHS(r)))
               			fTableau.setRowCondition(r, QcLinEqRowState.fRedundant);
					else {
						fTableau.setRowCondition(r, QcLinEqRowState.fInvalid);
          	          	result = false;
	               	}
               	}
          	}
		}

		// Step 3: Pivot using restricted variables.
		for (int c = 0; c < cols; c++) {
			if (rbasic[c]) {
				int maxrow = -1;
				double max = 0.0;
				QcTableauColIterator iC =
                    		new QcTableauColIterator(fTableau, c);

				while (!iC.atEnd()) {
					if (max < Math.abs(iC.fValue)) {
						max = iC.fValue;
						maxrow = iC.fIndex;
					}

					iC.increment();
				}

				// If there is at least one non-zero coefficient in the
				// current column.
				if (maxrow != -1)
					fTableau.pivot(maxrow, c);
			}
		}

		// Step 4: Pivot on any variables to make all the solved form rows
		// with a basic variables.
		for (int r = 0; r < rows; r++) {
         		if (fTableau.getRowCondition(r) != QcLinEqRowState.fRegular) {
				int maxcol = -1;
				double max = 0.0; 
				QcTableauRowIterator iR =
                    		new QcTableauRowIterator(fTableau, r);

				while (!iR.atEnd()) {
					if (max < Math.abs(iR.fValue)) {
    	     	     		max = iR.fValue;
         	     	     	maxcol = iR.fIndex;
                    	}
          			iR.increment();
          		}

				// If there is at least one non-zero coefficient in the
				// current row.
				if (maxcol != -1)
					fTableau.pivot(r, maxcol);
				else {
					// Verify that the Reset operation is successful.  It
					// fails if one of more of the solved form row is
					// marked as "invalid".
					if (QcUtility.isZero(fTableau.getRHS(r)))
						fTableau.setRowCondition(r, QcLinEqRowState.fRedundant);
					else {
						fTableau.setRowCondition(r, QcLinEqRowState.fInvalid);
						result = false;
					}
				}
          	}
		}

		// Step 5: For each restricted rows with negative RHS, add an artifical 
		// variable with -1 coefficient.  This is followed by doing a pivot on it.
		boolean moreProcess = false;
		Vector artVars = new Vector();

		for (int r = 0; r < rows; r++) {
			if (fTableau.isRestricted(r)
					&& QcUtility.isNegative(fTableau.getRHS(r))) {
				int vi = fTableau.addArtificial(-1.0);
				artVars.addElement(new Integer(vi));
				fTableau.pivot(r, vi);
					// May scale row by -1 to make the newly added
					// artifical variable to +1.  This method may require
					// explicit management of column and row state.
				moreProcess = true;
			}
		}

		if (!moreProcess)
			return result;

		// Step 6: Set objective function based on the added artifical
		// variables, and perform Simplex phase II.

		// First, initialize the objective function.
		cols = fTableau.getColumns();	// since number of columns may have
								// changed.
	
		for (int c = 0; c < cols; c++)
			fTableau.setObjValue(c, 0.0);

		// Then, set the objective function base on the artifical variables.
		Enumeration aIt = artVars.elements();

		while (aIt.hasMoreElements()) {
			int vi = ((Integer)aIt.nextElement()).intValue();
			QcFloat v = fVBiMap.identifier(vi);
			fTableau.setObjValue(vi, v.fWeight);
		}

		fTableau.eliminateObjective();
		simplexII();

		// Step 7: Remove the added artifical variables from the tableau.
		aIt = artVars.elements();

		while (aIt.hasMoreElements()) {
			int vi = ((Integer)aIt.nextElement()).intValue();
			int r = fTableau.isBasicIn(vi);
			
			if (!fTableau.isBasic(vi))
				fTableau.removeVar(vi);
			else {
				int maxcol = -1;
				double max = 0.0;
				QcTableauRowIterator iR =
						new QcTableauRowIterator(fTableau, r);

				while (!iR.atEnd()) {
					if (max < Math.abs(iR.fValue) && iR.fIndex != vi) {
						max = iR.fValue;
						maxcol = iR.fIndex;
					}

					iR.increment();
				}

				// If there is at least one non-zero coefficient in the
				// current column.
				if (maxcol != -1) {
					fTableau.pivot(r, maxcol);
					fTableau.removeVar(vi);
				} else {
					if (QcUtility.isZero(fTableau.getRHS(r))) {
						fTableau.setRowCondition(r, QcLinEqRowState.fRedundant);
						fTableau.removeVar(vi);
					} else {
						fTableau.setRowCondition(r, QcLinEqRowState.fInvalid);
						result = false;
					}
				}
			}
		}

		artVars.removeAllElements();
		return result;
	}

	public boolean removeConstraint(QcConstraint c)
	{
		System.out.println(
			"Error: Should not call QcLinInEqSystem::removeConstraint");
		return false;
	}

	//-----------------------------------------------------------------------//
	// Constraint Solving methods                                            //
	//-----------------------------------------------------------------------//
	public void eqSolve()
		// Equivalent to QcLinEqSystem::Solve
	{
		// The following loop is equiavalent to QcLinEqSystem::loadDesValCache()
		// N.B. QcLinEqTableau::evalBasicVar needs caching of non-structural
		// variables hence QcLinInEqSystem::loadDesValCache() will not do.
		QcVariableIndexIterator vIt = new QcVariableIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			fTableau.setDesireValue(vIt.fIndex,
				fVBiMap.identifier(vIt.fIndex).fDesireValue);
			vIt.increment();
		}

		vIt.reset();

		while (!vIt.atEnd()) {
			QcFloat v = fVBiMap.identifier(vIt.fIndex);

			if (fTableau.isBasic(vIt.fIndex)) {
				double value = ((QcLinEqTableau)fTableau).evalBasicVar(vIt.fIndex);
				v.fValue = QcUtility.zeroise(value);
			} else
				v.setVariable();
			vIt.increment();
		}
	}

	protected void initObjective()
		// Build objective function SIGMA(x.weight * x) for structural
		// variables x.  Replace basic variables by their parametric
		// representation but don't include any non-slack parameters.
	{
		QcStructVarIndexIterator vIt = new QcStructVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			// TODO: The C++ version doesn't have the isStructural test.
			if (fTableau.isBasic(vIt.fIndex)
			    && fTableau.isStructural(vIt.fIndex)) {
				QcFloat v = fVBiMap.identifier(vIt.fIndex);

				fTableau.setObjValue(vIt.fIndex, v.fWeight);
			} else
				fTableau.setObjValue(vIt.fIndex, 0.0f);

			vIt.increment();
		}

		fTableau.eliminateObjective();
	}

	protected void rawSolve()
		// Evaluate basic Variables from parameters
	{
		loadDesValCache();

		QcStructVarIndexIterator vIt =
			new QcStructVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			QcFloat v = fVBiMap.identifier(vIt.fIndex);

			if (fTableau.isBasic(vIt.fIndex)) {
				double value = fTableau.evalBasicVar(vIt.fIndex);
				v.fValue = QcUtility.zeroise(value);
			} else
				v.setVariable();

			vIt.increment();
		}
	}

	public void resolve()
		// This is the same as QcLinEqSystem::resolve except where noted with
		// the comment ++.
	{
		if (!fEditVarsSetup)
			beginEdit();

		// This paragraph not in LinEqSystem::Resolve ++
		if (fTableau.getRestrictedRows() > 0) {
			// Perform second phase of revised simplex algorithm on restricted
			// portion of the tableau.
			initObjective();
      		simplexII();
		}

		// adjust the dependent basic variables
		for (int vIt = 0; vIt < fDepVars.fCount; vIt++) {
			QcFloat dv = fVBiMap.identifier(fDepVars.fData[vIt]);
			Vector coeffs = fDepVarCoeffs[fDepVars.elementAt(vIt)];
			double dvValue = fDepVarBaseVals.fData[fDepVars.fData[vIt]];
			int cfSize = coeffs.size();

			for (int pIt = 0; pIt < cfSize; pIt++) {
				QcSparseCoeff cv = (QcSparseCoeff)coeffs.elementAt(pIt);
				QcFloat p = fVBiMap.identifier(cv.fIndex);
				dvValue -= cv.fValue * p.fDesireValue;
			}

			dv.fValue = QcUtility.zeroise(dvValue);
		}

		// adjust the parameters of interest
		for (int vIt = 0; vIt < fParsOfInterest.fCount; vIt++)
			fVBiMap.identifier(fParsOfInterest.fData[vIt]).setVariable();
     }

	public void solve()
	{
		if (fTableau.getRestrictedRows() > 0) {
			// Perform second phase of revised simplex algorithm on restricted
			// portion of the tableau.
			initObjective();
			simplexII();
		}

		// Calculate the value for each variable depending upon
		// whether it is basic or parametric.
		rawSolve();
	}

	//-----------------------------------------------------------------------//
	// Advanced methods.                                                     //
	//-----------------------------------------------------------------------//
	protected boolean cannotIncrease(double coeff, double rhs)
		// Used by SimplexI and applied to rhs = QcLinTableau::getRHS.
		// The sign of coeff and rhs is critical hence the extra check.
	{
		if (QcUtility.isNearZero(coeff))
			return false;

		if (rhs < 0.0)
			return (coeff < 0.0);	// rhs is zeroised already,
            						// otherwise use if isNeg(rhs)
		else
			return (coeff > 0.0);
	}

	public boolean dualSimplexII()
		// As for SimplexII only this algorithm uses the dual algorithm and
		// needs an initially optimal (but infeasible) solution.
		// Maintains optimality while searching for a feasible solution.
	{
		int n = 1;

		// This loop contains 2 exits, a "return true" and a "return false".
		while (true) {
			// May be better than inf loop
      		if (++n == QcConstants.fMaxLoopCycles)
				return false;

			int ei = selectDualExitVar(); // Find pivot row.

			// Feasible
			if (ei < 0)
				return true;

			// Select parameter to make basic
			int vi = selectDualEntryVar(ei);

			if (vi < 0)
				return false;

			// Do a full pivot for the now.  Later we may choose to implement a
			// slack pivot which only eliminates the new basic slack variable
			// from the restricted portion of the tableau.  This would violate
			// some of the integrity constraints for lineqtableau - so more
			// thought is needed first (maybe use two separate tableaus).
			fEditVarsSetup = false;	// for when called by Resolve
			fTableau.simplexPivot(ei, vi);
		}
	}

	protected void loadDesValCache()
		// Used for Solve
	{
		QcStructVarIndexIterator vIt =
			new QcStructVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			fTableau.setDesireValue(vIt.fIndex,
           		fVBiMap.identifier(vIt.fIndex).fDesireValue);
			vIt.increment();
		}
	}

	public int selectEntryVar()
		// Returns parameter for SimplexII to make basic.
		// Or returns invalid index if there are no candidates.
		// Any constrained parameter with negative objective coefficient will
		// do.  But to implement Bland's anti-cycling rule, we choose the
		// parameter with the lowest index.
	{
		QcParamVarIndexIterator vIt =
			new QcParamVarIndexIterator(fTableau);

		while (!vIt.atEnd()) {
			if (fTableau.isConstrained(vIt.fIndex)) {
				if (QcUtility.isNegative(fTableau.getObjValue(vIt.fIndex)))
					return vIt.fIndex;
			}
			vIt.increment();
		}

		return QcTableau.fInvalidVariableIndex;
	}

	public int selectExitVar(int vi)
		// Select the pivot row for pivot variable vi (slack parameter)
		// maintaining Basic Feasible Solved Form.
		// Returns InvalidCIndex if no such pivot is possible.
		// This version implement's Bland's anticycling rule.
		// Used by SimplexII, BeginEdit and in derived classes.
	{
		int ei = QcTableau.fInvalidConstraintIndex;
		double minval = Double.MAX_VALUE;
		int bi = fTableau.getColumns();		// Bland Index

		QcTableauColIterator colCoeffs =
			new QcTableauColIterator(fTableau, vi);

		while (!colCoeffs.atEnd()) {
			if (fTableau.isRestricted(colCoeffs.fIndex)) {
				double rhs = fTableau.getRHS(colCoeffs.fIndex);
				if (cannotIncrease(colCoeffs.fValue, rhs)) {
					double coeff = rhs / colCoeffs.fValue;
					if (coeff < minval || (coeff == minval &&
                   				fTableau.getBasicVar(colCoeffs.fIndex) < bi)) {
						minval = coeff;
						ei = colCoeffs.fIndex;
						bi = fTableau.getBasicVar(ei);
					}
				}
			}
			colCoeffs.increment();
		}

		return ei;
	}

	public int selectDualEntryVar(int ei)
		// Select a slack parameter to enter the basis which will approach
		// feasibility. Used by DualSimplexII.
	{
		int vi = QcTableau.fInvalidVariableIndex;
		double maxval = -Double.MAX_VALUE;

		QcTableauRowIterator varCoeffs = new QcTableauRowIterator(fTableau, ei);

		while (!varCoeffs.atEnd()) {
			if (fTableau.isConstrained(varCoeffs.fIndex)) {
				if (QcUtility.isNegative(varCoeffs.fValue)) {
					double coeff = fTableau.getObjValue(varCoeffs.fIndex) /
						varCoeffs.fValue;

					if (coeff > maxval) {
						maxval = coeff;
						vi = varCoeffs.fIndex;
					}
				}
			}
			varCoeffs.increment();
		}

		return vi;
	}

	public int selectDualExitVar()
		// Select the basic variable to leave the basis so that optimality
		// is maintained (in the narrow sense - feasibility is not assumed).
		// And feasibility is approached.
		// Returns InvalidCIndex if no such pivot is possible (i.e. optimal
		// feasible).
	{
		double minval = 0.0f;
		int ei = QcTableau.fInvalidConstraintIndex;	// Result constraint index

		// Find the the constrainted basic with the most negative value.
		QcBasicVarIndexIterator bvIt = new QcBasicVarIndexIterator(fTableau);

		while (!bvIt.atEnd()) {
			if (fTableau.isConstrained(bvIt.fIndex)) {
				int ci = fTableau.isBasicIn(bvIt.fIndex);
				double rhs = fTableau.getRHS(ci);

				if (QcUtility.isNegative(rhs) && (rhs < minval)) {
					minval = rhs;
					ei = ci;
				}
			}
			bvIt.increment();
		}

		return ei;
	}

	public void setObjective(QcLinPoly obj)
		// Simplex objective minimise function
	{
		QcRowAdaptor objIt = new QcRowAdaptor(obj, fVBiMap, fTableau);
		fTableau.setObjective(objIt);
	}

	public boolean simplexI(int cs)
		// Uses the first phase of the revised simplex algorithm to obtain
		// a basic feasible solution for the new constraint with index cs.
		// Returns false only if the problem is not satisfiable.
		// SimplexI tries to find a basic feasible solution when adding
		// a new row to a tableau previously in basic feasible form.
		// It has not been possible to find an unconstrained variable to
		// make basic.  Hence a restricted row is being created.
		// The slackCoeffs vector parameter does not include the new slack
		// variable, this has already been tried in Applyhints.
		// A precondition for this call is that the tableau is basic
		// feasible apart from the new row cs which is being added.
	{
		int n = 1;
		int ei;
		fTableau.normalize(cs);

		do {
			double rhs = fTableau.getRHS(cs);

			// May be better than inf loop
			if (++n == QcConstants.fMaxLoopCycles)
				return false;

			// Find parameter to make basic (any with positive coefficient)
			int vi = QcTableau.fInvalidVariableIndex;
            	// Index of variable which is to become basic
			double val = 0.0f;
			boolean found = false;
			QcTableauRowIterator varCoeffs =
				new QcTableauRowIterator(fTableau, cs);

			while (!varCoeffs.atEnd()) {
				vi = varCoeffs.fIndex;
				val = varCoeffs.fValue;
				if (!fTableau.isBasic(vi) && fTableau.isConstrained(vi) &&
						!QcUtility.isZero(varCoeffs.fValue) &&
						varCoeffs.fValue > 0.0) {
					found = true;
					break;
				}
				varCoeffs.increment();
			}

			// New constraint is inconsistant.
			if (!found)
				return false;

			// Choose basic variable to make parametric (ei)
			ei = cs;		// The prefered row to choose from.
			double minval = rhs / val;
			QcTableauColIterator colCoeffs =
				new QcTableauColIterator(fTableau, vi);

			while (!colCoeffs.atEnd()) {
				double rhs2 = fTableau.getRHS(colCoeffs.fIndex);
				double minval2 = rhs2 / colCoeffs.fValue;

				if (QcUtility.isPositive(colCoeffs.fValue) &&
                 			fTableau.isRestricted(colCoeffs.fIndex)) {
					if (minval2 < minval || (minval2 == minval &&
							colCoeffs.fIndex != cs &&
							fTableau.getBasicVar(colCoeffs.fIndex) <
							fTableau.getBasicVar(ei))) {
						minval = minval2;
						ei = colCoeffs.fIndex;
					}
				}
				colCoeffs.increment();
			}

			fTableau.pivot(ei, vi);
		} while (ei != cs);

		return true;
	}

	public boolean simplexII()
		// Uses the second phase of the revised simplex algorithm to
		// find a solution minimising the objecive function.
		// Perform simplex optimisation on the restricted rows of the tableau.
		// Should only be called if there are some restricted rows.
		// The objective function should be initialised before this call.
		// Bland's anti-cycling rule is implemented and depends upon tableau
		// column iterators providing coeffs in order of increasing
		// variable index.
	{
		int n = 1;

		// This loop contains 2 exits, a "return true" and a "return false".
		while (true) {
			// may be better than inf loop
			if (++n == QcConstants.fMaxLoopCycles)
				return false;

			// Find parameter to make basic.
			int vi = selectEntryVar();
			// All objective coeffs are non-negative
			if (vi < 0)
				return true;

			// Choose row index of basic slack variable to make parametric (ei)
			int ei = selectExitVar(vi);

			if (ei < 0)
				return false;

			// Do a full pivot for the now.  Later we may choose to implement a
			// slack pivot which only eliminates the new basic slack variable
			// from the restricted portion of the tableau.  This would violate
			// some of the integrity constraints for lineqtableau - so more
			// thought is needed first (maybe use two separate tableaus).
			fEditVarsSetup = false;	// for when called by Resolve
			fTableau.simplexPivot(ei, vi);
		}
	}

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//
	public void loadFromBinary(DataInputStream s, QcVariableSet vars)
	{
		super.loadFromBinary(s, new QcLinInEqTableau(s,
			new QcBiMapNotifier(s, vars)), vars);
		fArtVars = new QcIntVector();
		fTableau = (QcLinInEqTableau)super.fTableau;
	}

	public void loadFromBinary(DataInputStream s, QcLinInEqTableau tab,
		QcVariableSet vars)
	{
		super.loadFromBinary(s, tab, vars);
		fArtVars = new QcIntVector();
		fTableau = tab;
	}

	public void loadFromString(StringTokenizer token, QcVariableSet vars)
	{
		super.loadFromString(token, new QcLinInEqTableau(token,
			new QcBiMapNotifier(token, vars)), vars);
		fArtVars = new QcIntVector();
		fTableau = (QcLinInEqTableau)super.fTableau;
	}

	public void loadFromString(StringTokenizer token, QcLinInEqTableau tab,
		QcVariableSet vars)
	{
		super.loadFromString(token, tab, vars);
		fArtVars = new QcIntVector();
		fTableau = tab;
	}
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
