package qoca;
// $Id: QcLinPoly.java.cpp,v 1.10 2000/11/25 13:56:17 pmoulder Exp $

#include "QcDefines.jh"

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//============================================================================//

import java.util.*;
import java.io.*;

/** Linear expression, a sum of <var>n</var> terms.  Each term is of class QcLinPolyTerm. */
public class QcLinPoly
	implements QcDupable, java.io.Serializable
{
	private final static boolean fCheckPre = QcConstants.fCheckPre;
	private final static int fDefaultInitCapacity = 4;
	private final static int fDefaultIncrement = 0;

	/** The <tt>QcLinPolyTerm</tt>s.  Order is not significant. */
	public QcUnsortedListSet fTerms;

	//-----------------------------------------------------------------------//
	// Constructors                                                          //
	//-----------------------------------------------------------------------//
	public QcLinPoly()
	{
		fTerms = new QcUnsortedListSet(fDefaultInitCapacity, fDefaultIncrement);
	}

	public QcLinPoly(int initialCapacity)
	{
		fTerms = new QcUnsortedListSet(initialCapacity, fDefaultIncrement);
	}

	public QcLinPoly(QcLinPoly other)
	{
		int n = other.fTerms.size();
		fTerms = new QcUnsortedListSet (n);
		fTerms.setSize(n);
		for (int i = n; --i >= 0;)
			fTerms.setElementAt(new QcLinPolyTerm((QcLinPolyTerm)other.fTerms.elementAt(i)), i);
	}

	public QcLinPoly(QcLinPolyTerm term)
 	{
		fTerms = new QcUnsortedListSet(fDefaultInitCapacity, fDefaultIncrement);
		addUniqueTerm(term);
 	}

//#if qcEnableLoadSave
	public QcLinPoly(DataInputStream s, QcVariableSet vars)
		{ loadFromBinary(s, vars); }

	/** @deprecated */
	public QcLinPoly(DataInputStream s, QcVariableBiMap bimap)
		{ loadFromBinary(s, bimap); }

	public QcLinPoly(StringTokenizer token, QcVariableBiMap bimap)
		{ loadFromString(token, bimap); }
//#endif

#ifndef NDEBUG
	public void assertInvar()
	{
		int n = getNTerms();
		qcAssertInvar(n >= 0);
		for(int i = 0; i < n; i++) {
			QcLinPolyTerm ti = fTerms.elementAt (i);
			qcAssertInvar (ti != null);
			ti.assertInvar();
			qcAssertInvar (ti.getCoeff() != 0.0);
			/* effic: If we expect largish (n >= 6) termsets, then we could
			   use a marking technique (using a flag in each QcFloatRep). */
			QcFloat tiv = ti.getVariable();
			for(int j = i; --j >= 0;)
				qcAssertInvar (fTerms.elementAt (j).getVariable() != tiv);
		}
	}
#endif

	//-----------------------------------------------------------------------//
	// Query functions                                                       //
	//-----------------------------------------------------------------------//
	/** Synonym of <tt>getNTerms</tt>. */
	public int getSize()
		{ return fTerms.size(); }

	/** Returns the number of terms in this <tt>QcLinPoly</tt>. */
	public int getNTerms()
		{ return fTerms.size(); }

	public QcFloat getVar(int ix)
	{
		/* This is used by KpSolver, and it needs to have the actual
                   variable (or at least something that suggestValue is
                   effective for). */
		return ((QcLinPolyTerm) fTerms.elementAt(ix)).fVariable;
	}

	public boolean hasVar (QcFloat v)
	{
		for (int i = fTerms.size(); --i >= 0;)
			if (((QcLinPolyTerm) fTerms.elementAt (i)).fVariable == v)
				return true;
		return false;
	}

//#ifdef qcCheckPost
	double getValue()
	{
		double v = 0.0;
		for (int i = fTerms.size(); --i >= 0;)
			v += ((QcLinPolyTerm) fTerms.elementAt (i)).getValue();
		return v;
	}
//#endif

	//-----------------------------------------------------------------------//
	// Manipulation functions                                                //
	//-----------------------------------------------------------------------//

	/** Add <tt>term</tt> to this <tt>QcLinPoly</tt>.  Does nothing if
	    <tt>isZero(term.fCoeff)</tt>.

	    @precondition <tt>term</tt>'s variable is not already present in
	    this <tt>QcLinPoly</tt>.
	**/
	public QcLinPoly addUniqueTerm(QcLinPolyTerm term)
	{
		return addUniqueTerm(term.fCoeff, term.fVariable);
	}

	public QcLinPoly addUniqueTerm(double coeff, QcFloat var)
	{
		if (fCheckPre) { //#if qcCheckPre
			for (int i = fTerms.size(); --i >= 0;) {
				QcLinPolyTerm t = (QcLinPolyTerm) fTerms.elementAt(i);
				if (var == t.fVariable)
					throw new PreconditionException("variable already present");
			}
		} //#endif

		if (!QcUtility.isZero(coeff))
			fTerms.addElement(new QcLinPolyTerm(coeff, var));

		dbg(assertInvar());
		return this;
	}

	/** @deprecated Renamed to <tt>addUniqueTerm</tt>. */
	public void addTerm(QcLinPolyTerm term)
	{
		addUniqueTerm(term);
	}

	public void plus(QcLinPoly poly)
	{
		int size = poly.fTerms.size();

		for (int i = 0; i < size; i++) {
			QcLinPolyTerm t = (QcLinPolyTerm)poly.fTerms.elementAt(i);
			doPlus(t.fCoeff, t.fVariable);
		}
	}

	public void plus(QcLinPolyTerm other)
	{
		if (QcUtility.isZero(other.fCoeff))
			return;
		doPlus(other.fCoeff, other.fVariable);
	}

	public void minus(QcLinPolyTerm other)
	{
		if (QcUtility.isZero(other.fCoeff))
			return;
		doPlus(-other.fCoeff, other.fVariable);
	}

	private void doPlus(double coeff, QcFloat var)
	{
		for (int i = fTerms.size(); --i >= 0;) {
			QcLinPolyTerm ti = (QcLinPolyTerm)fTerms.elementAt(i);
			if (ti.fVariable != var)
				continue;
			ti.fCoeff += coeff;
			if (QcUtility.isZero(ti.fCoeff))
				fTerms.removeElementAt(i);
			return;
		}
		// not found; append to list.
		fTerms.addElement(new QcLinPolyTerm(coeff, var));
	}

	/** Unary minus. */
	public void negate()
	{
		int size = fTerms.size();

		for (int i = 0; i < size; i++)
			((QcLinPolyTerm)fTerms.elementAt(i)).negate();
	}

	public void multiply(double scalar)
	{
		if (scalar == 0.0) {
			fTerms.removeAllElements();
			return;
		}
		int newIx = 0;
		int n = fTerms.size();
		for (int oldIx = 0; oldIx < n; oldIx++) {
			QcLinPolyTerm toi = (QcLinPolyTerm)fTerms.elementAt(oldIx);
			double newCoeff = toi.fCoeff *= scalar;
			if (!QcUtility.isZero(newCoeff)) {
				if (newIx < oldIx)
					fTerms.setElementAt(toi, newIx);
				newIx++;
			}
		}
		if (newIx != n)
			fTerms.setSize(newIx);
	}

	//-----------------------------------------------------------------------//
	// Utility functions                                                     //
	//-----------------------------------------------------------------------//

	// Note: this method doesn't exist in C++ version.
	public void copy(QcLinPoly rhs)
	{
		if (this != rhs) {
			fTerms.removeAllElements();
			int size = rhs.fTerms.size();
			fTerms.setSize(size);

			for (int i = 0; i < size; i++)
				fTerms.addElement(
					((QcLinPolyTerm)rhs.fTerms.elementAt(i)).duplicate(NOT_QCFLOAT));
		}
	}

	public QcLinPoly duplicate(int depth)
	{
		int size = fTerms.size();
		QcLinPoly dup = new QcLinPoly(size);

		for (int i = 0; i < size; i++)
			dup.fTerms.setElementAt((((QcLinPolyTerm)fTerms.elementAt(i))
						   .duplicate(depth)),
						  i);

		return dup;
	}

	public QcLinPoly duplicate()
		{ return duplicate(NOT_QCFLOAT); }

	public String toString()
	{
		StringBuffer ret = new StringBuffer();
		ret.append('(');
		int n = fTerms.size();
		if (n != 0) {
			ret.append(fTerms.elementAt(0));
			for (int i = 1; i < n; i++) {
				ret.append(", ");
				ret.append(fTerms.elementAt(i));
			}
		}
		ret.append(')');
		return ret.toString();
	}

	public void print()
	{
		System.out.print(toString());
	}

//#if qcEnableLoadSave
	// pjm 2000-07-25: I'm not sure that this version is actually used.
	public void loadFromBinary(DataInputStream s, QcVariableBiMap bimap)
	{
		int size = QcUtility.readBinaryInt(s);

		// This will create the underlying data structures with the
		// enough capacity.
		fTerms = new QcUnsortedListSet(size, fDefaultIncrement);
		fTerms.setSize(size);

		for (int i = 0; i < size; i++)
			fTerms.setElementAt(new QcLinPolyTerm(s, bimap), i);
	}

	public void loadFromBinary(DataInputStream s, QcVariableSet vars)
	{
		int size = QcUtility.readBinaryInt(s);

		// This will create the underlying data structures with the
		// enough capacity.
		fTerms = new QcUnsortedListSet(size, fDefaultIncrement);
		fTerms.setSize(size);

		for (int i = 0; i < size; i++)
			fTerms.setElementAt(new QcLinPolyTerm(s, vars), i);
	}

	public void loadFromString(StringTokenizer token, QcVariableBiMap bimap)
	{
		int size = QcUtility.integerValue(token.nextToken());

		// This will create the underlying data structures with the
		// enough capacity.
		fTerms = new QcUnsortedListSet(size, fDefaultIncrement);
		fTerms.setSize(size);

		for (int i = 0; i < size; i++)
			fTerms.setElementAt(new QcLinPolyTerm(token, bimap), i);
	}

	public void saveAsBinary(DataOutputStream s)
	{
		int size = fTerms.size();

		if (size > 0) {
			QcUtility.writeBinaryInt(s, size);
			for (int i = 0; i < size; i++)
				((QcLinPolyTerm)fTerms.elementAt(i)).saveAsBinary(s);
		}
	}

	public void saveAsText(OutputStream s)
	{
		int size = fTerms.size();

		if (size > 0) {
			String str = "" + size + " ";
			QcUtility.writeTextString(s, str);
			((QcLinPolyTerm)fTerms.elementAt(0)).saveAsText(s);

			for (int i = 1; i < size; i++) {
				QcUtility.writeTextString(s, " ");
				((QcLinPolyTerm)fTerms.elementAt(i)).saveAsText(s);
			}
		}
	}

	public void saveToString(StringBuffer s)
	{
		int size = fTerms.size();

		if (size > 0) {
			s.append(size);
			s.append(" ");
			((QcLinPolyTerm)fTerms.elementAt(0)).saveToString(s);

			for (int i = 1; i < size; i++) {
				s.append(" ");
				((QcLinPolyTerm)fTerms.elementAt(i)).saveToString(s);
			}
		}
	}
//#endif
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
