package qoca;
// $Id: QcSolver.java.cpp,v 1.9 2000/11/25 13:56:17 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//============================================================================//

import java.util.*;
import java.io.*;

#ifdef JDK_GE_1_1  // JDK version >= 1.1.
import java.util.zip.*;
#endif

/** Encapsulates a constrained optimisation problem.
   
 <p>Currently QOCA provides three different solvers. All are based on keeping
 the current constraints in a tableau and using pivoting to keep them in a
 solved form.

 <p>The first, <a href="qoca.QcLinEqSolver.html" ><tt>QcLinEqSolver</tt></a>,
 provides linear equalities and uses the square of the (weighted) Euclidean
 distance to compare solutions.

 <p>The second, <a href="qoca.QcLinIneqSolver.html" ><tt
 >QcLinIneqSolver</tt></a>, also uses the square of the Euclidean distance but
 also allows linear inequalities.  It is based on linear complementary pivoting.

 <p>The third, <a href="qoca.QcCassSolver.html" ><tt>QcCassSolver</tt></a>, is
 based on the Cassowary algorithm.  It also provides linear equalities and
 inequalities but instead uses the (weighted) Manhattan distance to compare
 solutions.
**/
public abstract class QcSolver
	implements java.io.Serializable
{
	protected final static boolean fCheckPost = QcConstants.fCheckPost;
	protected final static boolean fCheckInternalPre = QcConstants.fCheckInternalPre;

	/** Records inconsistant constraints in the order they were
	    encountered. */
	protected Vector fInconsistant;
	protected Vector fEditVars;			// Used for Edit/Resolve
	protected boolean fEditVarsSetup;		// Used for Edit/Resolve
	protected boolean fAutoSolve;
	protected boolean fBatchAddConst;
	protected boolean fBatchAddConstFail;
	protected Vector fBatchConstraints;

	//-----------------------------------------------------------------------//
	// Constructor.                                                          //
	//-----------------------------------------------------------------------//
	public QcSolver()
	{
		fEditVarsSetup = false;
		fAutoSolve = false;
		fBatchAddConst = false;
		fInconsistant = new Vector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fEditVars = new Vector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
		fBatchConstraints = new Vector(QcConstants.fVectorInitCapacity,
			QcConstants.fVectorCapacityIncrement);
#ifdef qcCheckPost
		checkedConstraints = new Vector();
#endif
	}

	public void assertInvar() { } // fixme

	//-----------------------------------------------------------------------//
	// Variable management methods                                           //
	//-----------------------------------------------------------------------//

	/** Register a variable with a solver.  It is usually not necessary to
	    do this, as variables are registered automatically when they are
	    first encountered during an <tt>addConstraint</tt> operation.
	    <tt>addVar</tt> is useful to indicate variables which are to be
	    included in <tt>solve</tt> and <tt>restSolver</tt> operations when
	    not referred to in any of the constraints added to the solver.

	    @precondition <tt>isInConstraintMode()</tt>  (not confirmed)
	**/
	abstract public void addVar(QcFloat v);

	/** Removes variable from solver and returns true provided <tt>v</tt> is
	    free.  Does nothing and returns false if <tt>v</tt> is not free.

	    @precondition <tt>this.isRegistered(v)</tt>  (denied by doc)
	    @precondition <tt>isInConstraintMode()</tt>  (not confirmed)
	**/
	public boolean removeVar(QcFloat v)
	{
		System.out.println("Error: Should not call QcSolver::removeVar");
		return false;
	}

	/** Set the desired value of <tt>v</tt> to <tt>desval</tt>, iff
	    <tt>v</tt> is registered with this solver.  Equivalent to <tt>if
	    (this.isRegistered(v)) v.suggestValue(desval)</tt>.

	    <p>In the current implementation, <tt>v</tt> does not need to be an
	    edit var, nor does this solver need to be currently in goal-edit
	    mode.
	**/
	abstract public void suggestValue(QcFloat v, double desval);

	/** Call <tt>restVariable</tt> on each variable registered with this solver.
	    Useful in some situations after a solve.

	    @see QcFloat#restVariable
	**/
	abstract public void restSolver();

	//-----------------------------------------------------------------------//
	// Enquiry functions for variables                                       //
	//-----------------------------------------------------------------------//

	/** Indicates whether a variable has been introduced to the solver
	    either by virtue of a constraint being added, or via <tt>addVar</tt>.
	**/
	abstract public boolean isRegistered(QcFloat v);

	abstract public Enumeration getVariables();

	/** Indicates a parametric variable which does not depend on any basic
	    variables (i.e. has zero coefficients in all added constraints).

	    @precondition <tt>isRegistered(v)</tt>
	**/
	public boolean isFree(QcFloat v)
	{
		System.out.println("Error: Should not call QcSolver::isFree");
		return false;
	}


	/** Indicates whether a variable used by the tableau is basic.

	    @precondition <tt>isRegistered(v)</tt>
	**/
	abstract public boolean isBasic(QcFloat v);

	/** Returns true iff <tt>v</tt> is an edit var. */
	public boolean isEditVar(QcFloat v)
		{ return fEditVars.contains(v); }

	/** Add <tt>v</tt> to edit vars.  Its weight becomes its edit weight.

	    @precondition <tt>v</tt> &ne; <tt>null</tt>
	    @precondition <tt>isRegistered(v)</tt> (in practice; but earlier comments
	    suggested that <tt>addEditVar</tt> was supposed to do
	    <tt>addVar</tt> if necessary).
	**/
	public void addEditVar(QcFloat v)
	{
		v.setToEditWeight();
		registerEditVar(v);
	}

	/** Prepare for <tt>resolve</tt> call.  The first call to
	    <tt>resolve</tt> after changing the edit vars or the constraints
	    will call <tt>beginEdit</tt> and incur a slight delay.  An explicit
	    <tt>beginEdit</tt> call avoids that delay by allowing the
	    preprocessing to be done at a more convenient time. */
	abstract public void beginEdit();

	/** For every variable: make it non-edit (i.e.&nbsp;undoes
	    <tt>addEditVar</tt>), set it to stay weight, and (if not a
	    QcFixedFloat) set its desired value to the current value.

	    <p>If both fixed and nomadic variables are in this solver, and not
	    (each fixed variable weight is well outside of each nomadic edit
	    variable's stay--edit weight range), then <tt>endEdit</tt> may cause
	    the solution to be no longer optimal (i.e. with respect to the stay
	    weights).
	**/
	public void endEdit()
	{
		// Subclasses do the restSolver bit for themselves.
		int size = fEditVars.size();

		for (int i = 0; i < size; i++)
			((QcFloat)fEditVars.elementAt(i)).setToStayWeight();

		clearEditVars();
	}

	//-----------------------------------------------------------------------//
	// Constraint management methods                                         //
	//-----------------------------------------------------------------------//
	public void clearInconsistant()
     	{ fInconsistant.removeAllElements(); }

	/** Add to this solver all the constraints of <tt>other</tt>.  If each
	    of the addConstraint operations is successful, then return
	    <tt>true</tt>; otherwise, <tt>false</tt> is returned, and this
	    solver is left in an undefined state.

	    @precondition <tt>other != null
	**/
	public boolean swallow (QcSolver other)
	{
#ifdef qcCheckPre
		if (other == null)
			throw new PreconditionException("other == null");
		if (other == this)
			throw new PreconditionException("other == this");
#endif
#ifndef qcCheckPost
		throw new RuntimeException("qoca.QcSolver.swallow nyi unless compiled with qcCheckPost defined");
#else
		/* TODO: Consider transferring the variables too.  E.g. for v in
                   other.getVariables; do addVar(v); done.  (Except that
                   getVariables currently returns "internal" variables.) */
		for (int i = other.checkedConstraints.size(); --i >= 0;)
			if (!addConstraint ((QcConstraint) other.checkedConstraints.elementAt (i)))
				return false;
		return true;
#endif
	}

	/** Add <tt>c</tt> to the solver iff the result would be consistent.
	    (If the constraint is consisitent, then it is added to the solver,
	    and <tt>addConstraint</tt> returns true.  If the constraint is
	    inconsistant then it is not added and <tt>addConstraint</tt> returns
	    false.)

	    @return True iff <tt>c</tt> is consistent with the existing constraints.
	**/
	abstract public boolean addConstraint(QcConstraint c);

	/** Version of <tt>addConstraint</tt> with hint of which variable to
            make basic.

	    @precondition <tt>v != null</tt>
	    @precondition <tt>isRegistered(hint)</tt>
	**/
	abstract public boolean addConstraint(QcConstraint c, QcFloat hint);

#ifdef qcCheckPost
	/** Set of <tt>QcConstraint</tt>s in this solver. */
	private Vector checkedConstraints;
	protected void addCheckedConstraint(QcConstraint c)
	{
		checkedConstraints.addElement(c);
	}
#endif

	/** Marks the beginning of a batch of <tt>addConstraint</tt> operations. */
	public void beginAddConstraint()
	{
		fBatchConstraints.removeAllElements();
		fBatchAddConst = true;
		fBatchAddConstFail = false;
	}

	/** Change the RHS of a constraint.

	    @return <tt>true</tt> iff successful.
	**/
	abstract public boolean changeConstraint(QcConstraint oldc, double rhs);

#ifdef qcCheckPost
	protected void changeCheckedConstraint(QcConstraint oldc, double rhs)
	{
		oldc.fRHS = rhs;
	}
#endif

	/** Marks the end of a batch of <tt>addConstraint</tt> operations.

	    <p>In the current implementation, <tt>addConstraint</tt> always
	    returns false as soon as an inconsistent constraint is passed, but
	    in future we may postpone some checks until <tt>endAddConstraint</tt>.

	    @return <tt>true</tt> iff the constraints in this batch are
	    consistent.
	**/
	public boolean endAddConstraint()
	{
		fBatchAddConst = false;

		if (fBatchAddConstFail) {
			int size = fBatchConstraints.size();

			for (int i = 0; i < size; i++)
				removeConstraint((QcConstraint)fBatchConstraints.elementAt(i));
		}

		fBatchConstraints.removeAllElements();
		return !fBatchAddConstFail;
	}

	/** Remove <tt>c</tt> if present.  If the specified constraint is
	    present, then remove it without disturbing the values of any
	    variables, and return <tt>true</tt>.  Returns <tt>false</tt> if the
	    constraint is not present. */
	public boolean removeConstraint(QcConstraint c)
	{
		System.out.println(
          	"Error: Should not call QcSolver::removeConstraint");
		return false;
	}

#ifdef qcCheckPost
	protected void removeCheckedConstraint(QcConstraint c)
	{
		int i = checkedConstraints.indexOf(c);
		if (i < 0)
			throw new PreconditionException("removing non-present constraint");
		int n1 = checkedConstraints.size() - 1;
		if (i < n1) {
			Object last = checkedConstraints.elementAt (n1);
			checkedConstraints.setElementAt (last, i);
		}
		checkedConstraints.setSize(n1);
	}

	public void checkSatisfied()
	{
		for (int i = checkedConstraints.size(); --i >= 0;) {
			QcConstraint c = (QcConstraint) checkedConstraints.elementAt (i);
			if (!c.isSatisfied())
				throw new PostconditionException("constraint not satisfied: " + c);
		}
	}

	/** Returns true iff <tt>c</tt> is in this solver.  Note that the
            equality used is the actual <tt>QcConstraint</tt> reference pointers
            (i.e. <tt>Object.equals</tt>): it is not sufficient to have the same
            LHS and same RHS. */
	public boolean hasConstraint (QcConstraint c)
	{
		return checkedConstraints.contains (c);
	}
#endif

	/** Solve all the original constraints again from scratch.  Returns
	    false if any of the original constraints are found to be
	    inconsistant. */
	abstract public boolean reset();

	//-----------------------------------------------------------------------//
	// Constraint Solving methods                                            //
	//-----------------------------------------------------------------------//
	public boolean autoSolve()
		{ return fAutoSolve; }

	/** If set to <tt>true</tt>, the system is re-solved after certain
            (currently unspecified) operations.  Completely undocumented until
            now; don't rely on exactly when auto-solve implies a solve. */
	public void setAutoSolve(boolean a)
		{ fAutoSolve = a; }

	/** <tt>solve</tt> uses the desired values of the parametric variables
	    to calculate values for all the variables. */
	abstract public void solve();

	/** <tt>resolve</tt> (as in re-solve) is similar to solve but only
	    values for the variables of interest and dependent (basic) variables
	    are recalculated.  If exact arithmetic were used, this would be
	    equivalent to calling <tt>solve</tt>, but in practice <tt>solve</tt>
	    is more accurate. */
	abstract public void resolve();

	//-----------------------------------------------------------------------//
	// Low Level Edit Variable Interface for use by solvers. These functions //
	// manipulate the set of edit variables without any adjustment of the    //
	// variable weights or automatic registration on addition or restSolver()//
	// on removal.
	//-----------------------------------------------------------------------//

	/** Add <tt>v</tt> to <tt>fEditVars</tt> without calling <tt>addVar</tt> or
            <tt>setToEditWeight</tt>.

	    @precondition <tt>v != null</tt>
	    @precondition <tt>isRegistered(v)</tt>  (see addEditVar comment)
	**/
	void registerEditVar(QcFloat v)
	{
#if qcCheckPre
		if (v == null)
			throw new PreconditionException("v==null");
#endif
		if (!isEditVar(v))
			fEditVars.addElement(v);
		fEditVarsSetup = false;
	}

	/** Remove <tt>v</tt> from <tt>fEditVars</tt> without calling
	    <tt>restVariable</tt>.  (<tt>v</tt> need not already be an edit var.)
	 **/
	public void removeEditVar(QcFloat v)
	{
		fEditVars.removeElement(v);
		// TODO: [pjm]: Why isn't fEditVarsSetup set to false here?
	}

	/** Make all variables non-edit. */
	public void clearEditVars()
	{
		fEditVars.removeAllElements();
		fEditVarsSetup = false;
	}

	//-----------------------------------------------------------------------//
	// Utility functions.                                                    //
	//-----------------------------------------------------------------------//

	/** Return an existing variable registered with this server whose name
	    equals <tt>n</tt>.  If many variables in this solver match the name,
	    then the behavior is unspecified.

	    @postcondition ret = null &or; ret.getName().equals(n).
	 **/
	abstract public QcFloat getVariable(String n);

	/** Analagous to <tt>getVariable</tt>. */
	abstract public QcConstraint getConstraint(String n);

	abstract public void getVariableSet(QcVariableSet vars);

	public void print()
	{
		if (fEditVarsSetup) {
			System.out.println("Edit Variables");
			int size = fEditVars.size();
			for (int i = 0; i < size; i++)
				((QcFloat)fEditVars.elementAt(i)).print();
		}
	}

	/** Erase everything ready to start afresh. */
	public void restart()
	{
		clearEditVars();
		clearInconsistant();
		fEditVarsSetup = false;
	}

//#if qcEnableLoadSave
	public void loadFromStream(InputStream s)
		{ loadFromStream(s, null); }

	public void loadFromStream(InputStream s, QcVariableSet vars)
		{ loadFromBinary(new DataInputStream(s), vars); }

# ifdef JDK_GE_1_1  // JDK version >= 1.1.
	public void loadFromZip(String filename)
	{
		try {
			ZipInputStream zips = new ZipInputStream(
       			new BufferedInputStream(
                 	new FileInputStream(filename + ".zip")));
			zips.getNextEntry();
			loadFromString(QcUtility.makeTokenizer(zips), null);
			zips.close();
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}
	}
# endif

	public void loadFromBinary(DataInputStream s, QcVariableSet vars)
		{ fAutoSolve = QcUtility.readBinaryBoolean(s); }

	public void loadFromString(StringTokenizer token, QcVariableSet vars)
		{ fAutoSolve = (token.nextToken().equals("t")) ? true : false; }

# ifdef JDK_GE_1_1  // JDK version >= 1.1.
	public void saveToZip(String filename)
	{
		try {
			ZipOutputStream zips = new ZipOutputStream(
       			new FileOutputStream(filename + ".zip"));
			zips.setMethod(ZipOutputStream.DEFLATED);
			zips.putNextEntry(new ZipEntry(filename));
 			saveAsText(zips);
			zips.flush();
			zips.closeEntry();
			zips.close();
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}
	}
# endif

	public void saveAsText(OutputStream s)
		{ saveAsText(s, null); }

	public void saveAsText(OutputStream s, QcVariableSet vars)
	{
		String str = "" + ((fAutoSolve) ? "t" : "f") + "\n";
		QcUtility.writeTextString(s, str);
	}

	public void saveAsBinary(DataOutputStream s)
		{ saveAsBinary(s, null); }

	public void saveAsBinary(DataOutputStream s, QcVariableSet vars)
	{
		QcUtility.writeBinaryBoolean(s, fAutoSolve);
# ifndef qcCheckPost
		QcUtility.writeBinaryInt(s, 0);
# else
		int n = checkedConstraints.size();
		QcUtility.writeBinaryInt(s, n);
		for (int i = 0; i < n; i++) {
			QcConstraint c = (QcConstraint) checkedConstraints.elementAt (i);
			c.saveAsBinary(s);
		}
# endif
	}

	public void saveToStream(OutputStream s)
		{ saveToStream(s, null); }

	public void saveToStream(OutputStream s, QcVariableSet vars)
		{ saveAsBinary(new DataOutputStream(s), vars); }

	public void saveToString(StringBuffer s)
		{ saveToString(s, null); }

	public void saveToString(StringBuffer s, QcVariableSet vars)
	{
		s.append((fAutoSolve) ? "t" : "f");
		s.append("\n");
	}
//#endif /* qcEnableLoadSave */
}

/*
  Local Variables:
  mode:java
  c-file-style:"java8"
  fill-column:80
  End:
  vim: set filetype=java :
*/
