package qoca;
// $Id: QcUtility.java.cpp,v 1.2 2000/07/19 07:30:16 pmoulder Exp $

//============================================================================//
// Written by Sitt Sen Chok                                                   //
//----------------------------------------------------------------------------//
// The QOCA implementation is free software, but it is Copyright (C)          //
// 1994-1999 Monash University.  It is distributed under the terms of the GNU //
// General Public License.  See the file COPYING for copying permission.      //
//                                                                            //
// The QOCA toolkit and runtime are distributed under the terms of the GNU    //
// Library General Public License.  See the file COPYING.LIB for copying      //
// permissions for those files.                                               //
//                                                                            //
// If those licencing arrangements are not satisfactory, please contact us!   //
// We are willing to offer alternative arrangements, if the need should arise.//
//                                                                            //
// THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED OR  //
// IMPLIED.  ANY USE IS AT YOUR OWN RISK.                                     //
//                                                                            //
// Permission is hereby granted to use or copy this program for any purpose,  //
// provided the above notices are retained on all copies.  Permission to      //
// modify the code and to distribute modified code is granted, provided the   //
// above notices are retained, and a notice that the code was modified is     //
// included with the above copyright notice.                                  //
//============================================================================//

import java.io.*;
import java.util.*;

class QcUtility
{
	private static byte fBuffer[] = new byte[QcConstants.fOneMegaBytes];
	private static Integer fInteger = new Integer(0);

	//-----------------------------------------------------------------------//
	// We need some conventions about when small values are converted        //
	// to 0.0 or we will be performing unnecessary isZero tests.             //
	// Generally we don't want to test and "zeroise" values every time       //
	// a calculation is done (e.g. matrix row operations) since the values   //
	// may never need to be compared with zero anyway.  Better to take       //
	// care where efficiency may be impacted to ensure the test is only      //
	// done once, and where there is no efficiency impact, be cautious       //
	// and do isZero tests which are not strictly necessary - with           //
	// a comment to this effect.  Note that sparse iterators which by        //
	// definition do not return zero coefficients should do the isZero       //
	// test, clients do not need to repeat this test.                        //
	//-----------------------------------------------------------------------//
	public static boolean isNearZero(double val)
		{ return (val < QcConstants.fNearEps && val > -QcConstants.fNearEps); }

	public static boolean isNegative(double val)
		{ return (val <= -QcConstants.fEps); }

	public static boolean isPositive(double val)
		{ return (val >= QcConstants.fEps); }

	public static boolean isZero(double val)
		{ return (val < QcConstants.fEps && val > -QcConstants.fEps); }

	public static double zeroise(double val)
	{
		if (val > QcConstants.fEps)
			return val;
		else if (val < -QcConstants.fEps)
			return val;
		return 0.0f;
	}

	//-----------------------------------------------------------------------//
	// Convertion methods.                                                   //
	//-----------------------------------------------------------------------//
	public static String toShortestString(double val)
	{
		String str = String.valueOf(val);
		int p = str.indexOf('.');
		int e = str.indexOf('E');
		String left = str.substring(0, p);

		if (p != -1) {
			if (e == -1) {
				String right = str.substring(p + 1);

				if (right.equals("0"))
					return left;
			} else {
				String right = str.substring(p + 1, e);

				if (right.equals("0"))
					return left + str.substring(e);
			}
		}

		return str;
	}

	public static double doubleValue(String str)
		{ return Double.valueOf(str).doubleValue(); }

	public static int integerValue(String str)
		{ return fInteger.parseInt(str); }

	//-----------------------------------------------------------------------//
	// File management functions.                                            //
	//-----------------------------------------------------------------------//
	public static String readTextStream(InputStream s)
	{
		int total = 0;
		int size = 0;

		try {
			do {
				size = s.read(fBuffer, total, fBuffer.length - total);

				if (size == -1)
					break;

				total += size;

				if (total == fBuffer.length) {
					byte temp[] = new byte[fBuffer.length * 2];
					System.arraycopy(fBuffer, 0, temp, 0, total);
					fBuffer = temp;
				}
			} while (size != -1);
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}

#ifdef JDK_GE_1_1
		return new String(fBuffer);
#else
		// JDK version <= 1.1.
		return new String(fBuffer, 0);
#endif
	}

	public static StringTokenizer makeTokenizer(InputStream s)
		{ return new StringTokenizer(readTextStream(s), " \r\n", false); }

	public static boolean readBinaryBoolean(DataInputStream s)
	{
		try {
			return s.readBoolean();
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}

		return false;
	}

	public static double readBinaryDouble(DataInputStream s)
	{
		try {
			return s.readDouble();
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}

		return 0.0f;
	}

	public static int readBinaryInt(DataInputStream s)
	{
		try {
			return s.readInt();
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}

		return 0;
	}

	public static String readBinaryString(DataInputStream s)
	{
		try {
			return s.readUTF();
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}

		return null;
	}

	public static void writeBinaryBoolean(DataOutputStream s, boolean v)
	{
		try {
			s.writeBoolean(v);
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}
	}

	public static void writeBinaryDouble(DataOutputStream s, double v)
	{
		try {
			s.writeDouble(v);
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}
	}

	public static void writeBinaryInt(DataOutputStream s, int v)
	{
		try {
			s.writeInt(v);
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}
	}

	public static void writeBinaryString(DataOutputStream s, String str)
	{
		try {
			s.writeUTF(str);
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}
	}

	public static void writeTextString(OutputStream s, String str)
	{
		int strlen = str.length();

#ifdef JDK_GE_1_1
		try {
			s.write(str.getBytes(), 0, strlen);
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}
#else
		// JDK version <= 1.1.
		str.getBytes(0, strlen, fBuffer, 0);
#endif
		try {
			s.write(fBuffer, 0, strlen);
		} catch (IOException ioe) {
			System.out.println(ioe.toString());
		}
	}
}
